/* ======================================================================
 * 
 *  CCCCC    ++      ++   pfex3_dynamic.c++  -- example 3 of my Performer VR examples
 * CC   CC   ++      ++   Author(s): Bill Sherman
 * CC      ++++++  ++++++ Created: March 1, 2000
 * CC   CC   ++      ++   Last Modified: April 23, 2002
 *  CCCCC    ++      ++
 * 
 * =================================================================== */
/***********************************************************************/
/* This example code demonstrates a pfCAVE application program that    */
/*   renders a dynamic, albeit simple, world.  The blue cube and       */
/*   red pyramid fly in a circle about a stationary green pyramid.     */
/*                                                                     */
/*                                                                     */
/* The new features of this example include:                           */
/*   - The basic Performer Library functions/classes:                  */
/*      * pfMalloc() -- function to allocate shared memory             */
/*          [NOTE: it is important to use this instead of CAVEMalloc()]*/
/*      * pfGetFrameTimeStamp() -- used instead of *CAVETime           */
/*      * pfDCS -- dynamic coordinate system node class                */
/*                                                                     */
/*   - The basic pfCAVE Library functions:                             */
/*      [none introduced in this example]                              */
/*                                                                     */
/*   - Some pfCAVE functions/variables used for interaction:           */
/*      [none introduced in this example]                              */
/*                                                                     */
/* Considerations:                                                     */
/*   - World simulation dynamics are done in the APP process (ie. the  */
/*       main process).                                                */
/*   - Information about the objects is passed via the scene-graph     */
/*                                                                     */
/* Copyright 2002, William R. Sherman, All Rights Reserved.            */
/***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <math.h>

#define	PF_C_API 1

#include <Performer/pf/pfChannel.h>
#include <Performer/pf/pfDCS.h>		/* also handles pfSCS */
#include <Performer/pf/pfGeode.h>
#include <Performer/pf/pfLightSource.h>
#include <Performer/pfdu.h>
#include <Performer/prmath.h>		/* for pfVec4 */

#if defined(CAVE)
#include <pfcave.h>
#else /* CAVE vs. FreeVR */
#define GFX_PERFORMER
#include "freevr.h"
#endif /* FreeVR */


typedef struct {
		pfDCS	*obj1_dcs;	/* pointer to DCS node of object 1 */
		pfDCS	*obj2_dcs;	/* pointer to DCS node of object 2 */
	} WorldDataType;

/* This is very important -- this function will be used as arguments. */
void	exit_application(int);

void		update_world(WorldDataType *wd);
static pfScene	*createScene(WorldDataType *wd);

#define PERSONAL_SIMULATOR	/* define this to override the default simulator */
#ifdef PERSONAL_SIMULATOR
void	my_sim(vrRenderInfo *renderinfo, int mask);
#endif


int main(int argc, char **argv)
{
	WorldDataType	*wd;

	/********************************/
	/*** initialize CAVE routines ***/
	/********************************/
	signal(SIGINT, SIG_IGN);

	/* initialize Performer */
printf("about to call pfInit()\n");
	pfInit();

	signal(SIGINT, exit_application);
#if defined(CAVE)
	/* in CAVE library, this is where the tracker process is forked */
	pfCAVEConfig(&argc, argv, NULL);	/* allow CLAs to affect config */
#else /* CAVE vs. FreeVR */
	vrConfigure(&argc, argv, NULL);		/* allow CLAs to affect config */

#ifdef PERSONAL_SIMULATOR
	vrFunctionSetCallback(VRFUNC_DISPLAY_SIM, vrCallbackCreate(my_sim, 0));
#endif
#endif /* FreeVR */

	/* configure Performer */
printf("about to call pfConfig()\n");
	pfConfig();

#if defined(CAVE)
	/* is this necessary for this example? */
	CAVENear = 1.0;
	CAVEFar = 10000.0;
#endif /* !FreeVR */

	/* create the channels for viewing the world */
#if defined(CAVE)
	pfCAVEInitChannels();
#else /* CAVE vs. FreeVR */
	vrStart();
#endif /* FreeVR */

	/***************************************************************/
	/* Level of detail of performance information is higher for    */
	/*   simulator display.  This allows greater debugging ability */
	/*   at the expense of some performance.                       */
#if defined(CAVE)
	if (CAVEConfig->Simulator)
		pfCAVEMasterChan()->getFStats()->setClass(PFSTATS_ALL ^ PFSTATSHW_ENGFXPIPE_FILL, PFSTATS_ON);
	else	pfCAVEMasterChan()->getFStats()->setClass(PFSTATS_ALL, PFSTATS_OFF);
#endif /* !FreeVR */


	/***************************/
	/*** do world simulation ***/
	/***************************/
	wd = (WorldDataType *)pfMalloc(sizeof(WorldDataType), pfGetSharedArena());

	/* Create a scene and assign it to the channels */
#if defined(CAVE)
	pfCAVEMasterChan()->setScene(createScene(wd));
#else /* CAVE vs. FreeVR */
# if 1 /* for the moment we'll go with the default scene which has objects on all sides */
	vrPfMasterChannel()->setScene(createScene(wd));
# endif
#endif /* FreeVR */

	/* run until 'Esc' key is pressed */
	while (
#if defined(CAVE)
		!CAVEgetbutton(CAVE_ESCKEY)
#else /* CAVE vs. FreeVR */
		!vrGet2switchValue(0)
#endif /* FreeVR */
		) {

		/* In the non-Peformer model, most normal world operations */
		/* would go here -- in update_world().  With Performer,    */
		/* applications can also use callback routines within the  */
		/* scene-graph to modify the world.                        */
		update_world(wd);


		pfSync();
		/* latency critical operations go here */
#if defined(CAVE)
		pfCAVEPreFrame();
#else /* CAVE vs. FreeVR */
		vrPfPreFrame();
#endif /* FreeVR */
		pfFrame();
#if defined(CAVE)
		pfCAVEPostFrame();
#else /* CAVE vs. FreeVR */
		vrPfPostFrame();
#endif /* FreeVR */
	}

	/*********************/
	/*** close up shop ***/
	/*********************/
	exit_application(0);
}


/********************************************************************/
/* exit_application(): clean up anything started by the application */
/*   (forked processes, open files, open sockets, etc.)             */
/********************************************************************/
void exit_application(int)
{
#if defined(CAVE)
	pfCAVEHalt();
#else /* CAVE vs. FreeVR */
	vrExit();
#endif /* FreeVR */
	pfExit();	/* must come after vrExit() because FreeVR uses pfMemory */

	exit(0);
}


#ifdef PERSONAL_SIMULATOR
void my_sim(vrRenderInfo *renderinfo, int mask)
{
#if 0
	vrPrintf("Yo, doing my own simulator!\n");
#endif
	vrRenderDefaultSimulator(renderinfo, mask);
}
#endif



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**********************************/
		/**********************************/
		/** The World Simulation section **/
		/**********************************/
		/**********************************/

/**********************************************************************/
/* update_world(): The World Simulation routine.                      */
void update_world(WorldDataType *wd)
{
	float	obj_x, obj_y, obj_z;
	float	sim_time = pfGetFrameTimeStamp();

	obj_z = 5.5;
	obj_x =  3*sin(sim_time * 2.0);
	obj_y = -3*cos(sim_time * 2.0) + 4.0;
	wd->obj1_dcs->setTrans(obj_x, obj_y, obj_z);

	obj_y = 4.0;
	obj_x = 3*cos(sim_time * 2.0);
	obj_z = 3*sin(sim_time * 2.0) + 5.5;
	wd->obj2_dcs->setTrans(obj_x, obj_y, obj_z);

	/********************************************/
	/* allow other processes to get in edgewise */
	sginap(0);
}



/* ----------------8<-----------------8<-----------------8<----------------*/
/* In a non-example applicatation, the following would be a separate file. */

		/**************************/
		/**************************/
		/** The Graphics section **/
		/**************************/
		/**************************/

pfNode *make_cube(float r, float g, float b);
pfNode *make_pyramid(float r, float g, float b);
pfNode *make_axis();

/*****************************************************************************/
/* createScene() - creates the scene.  A geostate is assigned to the scene,  */
/* in order to control global rendering options such as lighting, texturing, */
/* and backface culling.  A light source is added; the light has its class's */
/* default settings (white, pointed straight down, etc.).  Then a simple     */
/* world consisting of a cube and pyramid are created.                       */
/*****************************************************************************/
static pfScene *createScene(WorldDataType *wd)
{
	pfScene		*scene;
	pfGeoState	*gstate;
	pfDCS		*cs_node;
	pfSCS		*scs_node;
	pfMatrix	cs_matrix;

printf("HEY: in pfex3_dynamic.c++::createScene()\n\n\n");
	scene = new pfScene;

	/* Set up a base geostate for the scene */
	gstate = new pfGeoState;
	gstate->setMode(PFSTATE_ENLIGHTING, PF_ON);
	gstate->setMode(PFSTATE_CULLFACE, PFCF_OFF);
	scene->setGState(gstate);

	/* Add a default light */
	scene->addChild(new pfLightSource);

	/**********************************************************/
	/*** Create a DCS (w/ matrix) for each *dynamic* object ***/

	/* an axis set */
	cs_node = new pfDCS();
	cs_node->setTrans(0.0, 3.5, 0.5);
	cs_node->addChild(make_axis());
	scene->addChild(cs_node);

	/* a blue cube */
	cs_node = new pfDCS();
	wd->obj1_dcs = cs_node;
	cs_node->setTrans(-2.0, 4.0, 5.5);
	cs_node->addChild(make_cube(100.0/256.0, 100.0/256.0, 255.0/256.0));
	scene->addChild(cs_node);

	/* a red pyramid */
	cs_node = new pfDCS();
	wd->obj2_dcs = cs_node;
	cs_node->setTrans( 2.0, 4.0, 5.5);
	cs_node->addChild(make_pyramid(255.0/256.0, 100.0/256.0, 100.0/256.0));
	scene->addChild(cs_node);

	/* a green upside down pyramid */
	cs_node = new pfDCS();
	cs_node->setTrans( 0.0, 4.0, 6.0);
	cs_node->setScale(1.0, 1.0, -1.0);	/* NOTE: flip in Z, not Y */
	cs_node->addChild(make_pyramid(100.0/256.0, 255.0/256.0, 100.0/256.0));
	scene->addChild(cs_node);

	/* extra stuff for testing multiple screens */
	/**********************************************************/
	/***  Create an SCS (w/ matrix) for each static object  ***/

	/* a green cube (front) */
	pfMakeTransMat(cs_matrix, -2.0,  5.5, 0.5);
	scs_node = pfNewSCS(cs_matrix);
	pfAddChild(scene, scs_node);
	pfAddChild(scs_node, make_cube(100.0/256.0, 255.0/256.0, 100.0/256.0));

	/* a red pyramid (front) */
	pfMakeTransMat(cs_matrix,  2.0,  5.5, 0.0);
	scs_node = pfNewSCS(cs_matrix);
	pfAddChild(scene, scs_node);
	pfAddChild(scs_node, make_pyramid(255.0/256.0, 100.0/256.0, 100.0/256.0));

	/* a blue cube (back) */
	pfMakeTransMat(cs_matrix, 0.0,  -5.5, 0.5);
	scs_node = pfNewSCS(cs_matrix);
	pfAddChild(scene, scs_node);
	pfAddChild(scs_node, make_cube(100.0/256.0, 100.0/256.0, 255.0/256.0));

	/* a blue cube (left) */
	pfMakeTransMat(cs_matrix, -6.0,  0.0, 0.5);
	scs_node = pfNewSCS(cs_matrix);
	pfAddChild(scene, scs_node);
	pfAddChild(scs_node, make_cube(100.0/256.0, 100.0/256.0, 255.0/256.0));

	/* a blue cube (right) */
	pfMakeTransMat(cs_matrix, 6.0,  0.0, 0.5);
	scs_node = pfNewSCS(cs_matrix);
	pfAddChild(scene, scs_node);
	pfAddChild(scs_node, make_cube(100.0/256.0, 100.0/256.0, 255.0/256.0));

	/* a blue cube (below) */
	pfMakeTransMat(cs_matrix, 0.0,  0.0, -2.5);
	scs_node = pfNewSCS(cs_matrix);
	pfAddChild(scene, scs_node);
	pfAddChild(scs_node, make_cube(100.0/256.0, 100.0/256.0, 255.0/256.0));

	/* a blue cube (above) */
	pfMakeTransMat(cs_matrix, 0.0,  0.0, 10.5);
	scs_node = pfNewSCS(cs_matrix);
	pfAddChild(scene, scs_node);
	pfAddChild(scs_node, make_cube(100.0/256.0, 100.0/256.0, 255.0/256.0));


	return (scene);
}


/****************************************/
/* make a unit cube of the given color. */
/*   Note that this cube is smaller than the shapes.c cube. */
pfNode *make_cube(float r, float g, float b)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;
	pfVec4		*color_vec;

	/* NOTE, classes/structures that are specificed by the pr library     */
	/*   do not have "new" overloaded to use shared memory for allocating */
	/*   the memory, so we need to be specifically allocat them via the   */
	/*   CAVE's shared memory allocation API.                             */
	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(r, g, b, 1.0);

	geoset = pfdNewCube(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode = new pfGeode;
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}


/*******************************************/
/* make a unit pyramid of the given color. */
/*   Note that this pyramid is smaller than the shapes.c cube. */
pfNode *make_pyramid(float r, float g, float b)
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;
	pfVec4		*color_vec;

	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(r, g, b, 1.0);

	geoset = pfdNewPyramid(pfGetSharedArena());
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode = new pfGeode;
	new_geode->addGSet(geoset);

	return (pfNode *)new_geode;
}


/****************************************/
/* make a unit cube of the given color. */
/*   Note that this cube is smaller than the shapes.c cube. */
pfNode *make_axis()
{
	pfGeoSet	*geoset;
	pfGeode		*new_geode;
	pfVec4		*color_vec;
	pfMatrix	ztox_mat;
	pfMatrix	ztoy_mat;
	pfMatrix	ztoz_mat;

	/* make the Geode node that we will return */
	new_geode = new pfGeode;

	/* setup some other values needed to create the axes */
#if 1 /* for C++ */
	ztox_mat.makeRot( 90.0, 0.0, 1.0, 0.0);
	ztoy_mat.makeRot(-90.0, 1.0, 0.0, 0.0);
	ztoz_mat.makeRot(  0.0, 1.0, 0.0, 0.0);

	ztox_mat.preScale(2.0, 2.0, 2.0, ztox_mat);
	ztoy_mat.preScale(2.0, 2.0, 2.0, ztoy_mat);
	ztoz_mat.preScale(2.0, 2.0, 2.0, ztoz_mat);
#else /* for C */
	pfMakeRotMat(ztox_mat,  90.0, 0.0, 1.0, 0.0);
	pfMakeRotMat(ztoy_mat, -90.0, 1.0, 0.0, 0.0);
#endif

	/* X-axis */
	geoset = pfdNewArrow(20, pfGetSharedArena());
	pfdXformGSet(geoset, ztox_mat);

	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(1.0, 0.2, 0.2, 1.0);
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode->addGSet(geoset);

	/* Y-axis */
	geoset = pfdNewArrow(20, pfGetSharedArena());
	pfdXformGSet(geoset, ztoy_mat);

	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(0.2, 1.0, 0.2, 1.0);
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode->addGSet(geoset);

	/* Z-axis */
	geoset = pfdNewArrow(20, pfGetSharedArena());
	pfdXformGSet(geoset, ztoz_mat);

	color_vec = (pfVec4 *)pfMalloc(sizeof(pfVec4), pfGetSharedArena());
	color_vec->set(0.2, 0.2, 1.0, 1.0);
	geoset->setAttr(PFGS_COLOR4, PFGS_OVERALL, color_vec, 0);
	new_geode->addGSet(geoset);


	return (pfNode *)new_geode;
}

