/* ======================================================================
 *
 *  CCCCC          vr_input.win32.c
 * CC   CC         Author(s): Sukru Tikves
 * CC              Created: August 5, 2002
 * CC   CC         Last Modified: June 7, 2003 (minor updates 09/30/21)
 *  CCCCC
 *
 * Code file for FreeVR inputs from MS-Windows events.
 *
 * Copyright 2021, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/*************************************************************************

FreeVR USAGE:
	...

HISTORY:
	8/5/02 -- initial version by Sukru

	9/9/02 -- formatted by Bill to match FreeVR style

	9/11/02 -- functions renamed by Bill to match FreeVR style (plus)
		  some other formatting fixes.
		Moved the control callback array into the _Win32Create() callback.

	21-23 April 2003 (Bill Sherman)
		Updated to new input matching function code style.  Changed "opaque"
		  field to "aux_data".

	3 June 2003 (Bill Sherman)
		Added the "system_pause_toggle" control callback.
		Added comments classifying the controls.

	----------

	5 May 2021 (Bill Sherman)
		Minor clean-up to make sizeof() safer.

		[corresponds to Freevr Version 0.7b]


TODO:
	...

*************************************************************************/
#include <windows.h>

#include <ctype.h>
#include <string.h>

#include "vr_input.h"
#include "vr_input.opts.h"
#include "vr_debug.h"
#include "vr_visren.h"
#include "vr_visren.wgl.h"
#include "vr_shmem.h"
#include "vr_config.h"
#include "vr_objects.h"

#define MAX_VKEYS         256
#define MAX_BUTTONS       16
#define MAX_POINTERS      4

#define MAX_INPUT_DEVICES 16

#define X_WINDOW 0
#define Y_WINDOW 1
#define X_SCREEN 2
#define Y_SCREEN 3


/****************************************************************************/
typedef struct {
		vrWindowInfo   *window;
		HWND            hWnd;

		vrGenericInput *key_inputs[MAX_VKEYS];
		vrGenericInput *button_inputs[MAX_BUTTONS];
		vrGenericInput *pointer_inputs[MAX_POINTERS];

		int             pointer_sign[MAX_POINTERS];
		int             width, height;

		int             silent;
	} _Win32PrivateInfo;


/****************************************************************************/
static	int	w32_keys[MAX_VKEYS];
static	int	w32_buttons[MAX_BUTTONS];
static	int	w32_x, w32_y;

static	_Win32PrivateInfo	*input_devices[MAX_INPUT_DEVICES];
static	vrLock			input_device_lock;


	/***********************************/
	/*** General NON public routines ***/
	/***********************************/


/****************************************************************************/
static void _Win32SetInputValue(vrGenericInput *input, int value)
{
	if (!input)
		return;

	switch (input->input_type) {

	case VRINPUT_BINARY:
		vrAssign2switchValue((vr2switch *)input, value);
		break;
	case VRINPUT_VALUATOR:
		vrAssignValuatorValue((vrValuator *)input, (float)value);
		break;
	case VRINPUT_CONTROL:
		vrCallbackInvokeDynamic(((vrControl *)input)->callback, 1, value);
		break;
	default:
		/* TODO: ... */
		break;
	}
}


/****************************************************************************/
static void _Win32SetInputFloat(vrGenericInput *input, float value)
{
	if (!input)
		return;

	switch (input->input_type) {

	case VRINPUT_BINARY:
		vrAssign2switchValue((vr2switch *)input, (int)value);
		break;
	case VRINPUT_VALUATOR:
		vrAssignValuatorValue((vrValuator *)input, value);
		break;
	default:
		/* TODO: ... */
		break;
	}
}


/****************************************************************************/
static _Win32PrivateInfo *_Win32FindInputDevice(HWND hWnd)
{
	int	count;

	if (!input_device_lock || !hWnd)
		return NULL;

	vrSetReadLock(input_device_lock);

	for (count = 0; count < MAX_INPUT_DEVICES; count++)
		if (input_devices[count] && input_devices[count]->hWnd == hWnd) {
			vrReleaseReadLock(input_device_lock);
			return input_devices[count];
		}

	vrReleaseReadLock(input_device_lock);

	return NULL;
}


/****************************************************************************/
static void _Win32UpdateKey(HWND hWnd, WPARAM wParam, int status)
{
	_Win32PrivateInfo *device;

	if (wParam > 255)
		return;

	w32_keys[wParam] = status;

	if ((device = _Win32FindInputDevice(hWnd)) == NULL)
		return;

	if (device->key_inputs[wParam])
		_Win32SetInputValue(device->key_inputs[wParam], status);
}


/****************************************************************************/
static void _Win32UpdateMouseButton(HWND hWnd, int button, int status)
{
	_Win32PrivateInfo *device;

	if (button < 0 || button >= MAX_BUTTONS)
		return;

	w32_buttons[button] = status;

	if ((device = _Win32FindInputDevice(hWnd)) == NULL)
		return;

	if (!device->button_inputs[button])
		_Win32SetInputValue(device->button_inputs[button], status);
}


/****************************************************************************/
static void _Win32UpdateSize(HWND hWnd, int width, int height)
{
	_Win32PrivateInfo *device;

	if ((device = _Win32FindInputDevice(hWnd)) == NULL)
		return;

	if (height < 1)
		height = 1;
	if (width < 1)
		width = 1;

	device->width = width;
	device->height = height;
}


/****************************************************************************/
static void _Win32UpdatePointerValue(_Win32PrivateInfo *info, int id, int v, int max)
{
	_Win32SetInputFloat(info->pointer_inputs[id], info->pointer_sign[id] *
			(((float)v/(float)max) / 2.0 - 1.0));
}


/****************************************************************************/
static void _Win32UpdatePointer(HWND hWnd, int x, int y)
{
	_Win32PrivateInfo	*device;
	POINT			screen_pos;
	HDC			hdcPrimaryMonitor;

	if ((device = _Win32FindInputDevice(hWnd)) == NULL)
		return;

	hdcPrimaryMonitor = GetDC(NULL);

	_Win32UpdatePointerValue(device, X_WINDOW, x, device->width);
	_Win32UpdatePointerValue(device, X_WINDOW, y, device->height);

	GetCursorPos(&screen_pos);

	x = GetDeviceCaps(hdcPrimaryMonitor, HORZRES);
	y = GetDeviceCaps(hdcPrimaryMonitor, VERTRES);

	_Win32UpdatePointerValue(device, X_SCREEN, screen_pos.x, x);
	_Win32UpdatePointerValue(device, Y_SCREEN, screen_pos.y, y);

	ReleaseDC(NULL, hdcPrimaryMonitor);
}


/****************************************************************************/
static LRESULT CALLBACK _Win32WindowProcess(HWND window, UINT message, WPARAM wParam, LPARAM lParam)
{
	POINT	screen_pos;

	switch (message) {

	case WM_SIZE:
		_Win32UpdateSize(window, LOWORD(lParam), HIWORD(lParam));
		break;

	case WM_KEYDOWN:
		_Win32UpdateKey(window, wParam, 1);
		return 0;

	case WM_KEYUP:
		_Win32UpdateKey(window, wParam, 0);
		return 0;

	case WM_LBUTTONDOWN:
		_Win32UpdateMouseButton(window, 1, 1);
		return 0;

	case WM_LBUTTONUP:
		_Win32UpdateMouseButton(window, 1, 0);
		return 0;

	case WM_MBUTTONDOWN:
		_Win32UpdateMouseButton(window, 3, 1);
		return 0;

	case WM_MBUTTONUP:
		_Win32UpdateMouseButton(window, 3, 0);
		return 0;

	case WM_RBUTTONDOWN:
		_Win32UpdateMouseButton(window, 2, 1);
		return 0;

	case WM_RBUTTONUP:
		_Win32UpdateMouseButton(window, 2, 0);
		return 0;

	case WM_MOUSEMOVE:
		w32_x = LOWORD(lParam);
		w32_y = HIWORD(lParam);
		GetCursorPos(&screen_pos);

		_Win32UpdatePointer(window, w32_x, w32_y);

		_Win32UpdatePointer(window, X_WINDOW, w32_x);
		_Win32UpdatePointer(window, Y_WINDOW, w32_y);
		_Win32UpdatePointer(window, X_SCREEN, screen_pos.x);
		_Win32UpdatePointer(window, Y_SCREEN, screen_pos.y);

		return 0;
	}

	return DefWindowProc(window, message, wParam, lParam);
}


/****************************************************************************/
static HWND _Win32NewHiddenWindow(void)
{
	HINSTANCE	hInstance;
	WNDCLASS	wc;
	HWND		hWnd;

	hInstance = GetModuleHandle(NULL);
	wc.style         = 0;
	wc.lpfnWndProc   = (WNDPROC)_Win32WindowProcess;
	wc.cbClsExtra    = 0;
	wc.cbWndExtra    = 0;
	wc.hInstance     = hInstance;
	wc.hIcon         = LoadIcon(NULL, IDI_WINLOGO);
	wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = NULL;
	wc.lpszMenuName  = NULL;
	wc.lpszClassName = "HiddenWindow";

	if (!RegisterClass(&wc)) {
		vrErr("WIN32: Failed to register window class.\n");
		return NULL;
	}

	if (!(hWnd = CreateWindowEx(
					0,
					"HiddenWindow",
					"Input Window",
					WS_POPUP,
					0, 0, 10, 10,
					NULL,
					NULL,
					hInstance,
					NULL
				  ))) {
		vrErr("WIN32: Could not allocate a window handle.\n");
		return NULL;
	}

	ShowWindow(hWnd, SW_HIDE);

	return hWnd;
}


/****************************************************************************/
static int _Win32KeyValue(const char *key)
{
	if (!key || !key[0])
		return 0;

	if (!key[1])
		switch (tolower(key[0])) {

		case '0': case ')': return 0x30;
		case '1': case '!': return 0x31;
		case '2': case '@': return 0x32;
		case '3': case '#': return 0x33;
		case '4': case '$': return 0x34;
		case '5': case '%': return 0x35;
		case '6': case '^': return 0x36;
		case '7': case '&': return 0x37;
		case '8': case '*': return 0x38;
		case '9': case '(': return 0x39;

		case 'a': return 0x41;
		case 'b': return 0x42;
		case 'c': return 0x43;
		case 'd': return 0x44;
		case 'e': return 0x45;
		case 'f': return 0x46;
		case 'g': return 0x47;
		case 'h': return 0x48;
		case 'i': return 0x49;
		case 'j': return 0x4A;
		case 'k': return 0x4B;
		case 'l': return 0x4C;
		case 'm': return 0x4D;
		case 'n': return 0x4E;
		case 'o': return 0x4F;
		case 'p': return 0x50;
		case 'q': return 0x51;
		case 'r': return 0x52;
		case 's': return 0x53;
		case 't': return 0x54;
		case 'u': return 0x55;
		case 'v': return 0x56;
		case 'w': return 0x57;
		case 'x': return 0x58;
		case 'y': return 0x59;
		case 'z': return 0x5A;

		default:
			  return 0;
		}

	if (!strcasecmp(key, "space"))
		return VK_SPACE;
	if (!strcasecmp(key, "escape") || !strcasecmp(key, "esc"))
		return VK_ESCAPE;
	if (!strcasecmp(key, "backspace"))
		return VK_BACK;
	if (!strcasecmp(key, "tab"))
		return VK_TAB;
	if (!strcasecmp(key, "enter") || !strcasecmp(key, "return"))
		return VK_RETURN;

	return 0;
}


/****************************************************************************/
static int _Win32ButtonValue(const char *button)
{
	if (!strcasecmp(button, "left"))
		return 1;
	if (!strcasecmp(button, "middle"))
		return 3;
	if (!strcasecmp(button, "right"))
		return 2;

	return 0;
}


/****************************************************************************/
static int _Win32PointerValue(const char *pointer)
{
	if (!strcasecmp(pointer, "x"))
		return X_WINDOW;
	if (!strcasecmp(pointer, "y"))
		return Y_WINDOW;
	if (!strcasecmp(pointer, "xroot"))
		return X_SCREEN;
	if (!strcasecmp(pointer, "yroot"))
		return Y_SCREEN;

	return -1;
}


	/*****************************************************************/
	/*   Function(s) for parsing Win32 "input" declarations.         */
	/*                                                               */
	/*  These _Win32<type>Input() functions are called during the    */
	/*  CREATE phase of the input interface.                         */

/****************************************************************************/
static vrInputMatch _Win32KeyInput(vrInputDevice *device, vrGenericInput *input, vrInputDTI *dti)
{
	_Win32PrivateInfo	*info = (_Win32PrivateInfo *)device->aux_data;
	int			keyvalue;

	if (!strcasecmp(device_name, "keyboard")) {
		keyvalue = _Win32KeyValue(dti->instance);

		if (keyvalue <= 0 || keyvalue > MAX_VKEYS)
			return VRINPUT_MATCH_UNABLE;

		info->key_inputs[keyvalue] = input;
		return VRINPUT_MATCH_ABLE;
	}
	return VRINPUT_NOMATCH;
}


/****************************************************************************/
static vrInputMatch _Win32ButtonInput(vrInputDevice *device, vrGenericInput *input, vrInputDTI *dti)
{
	_Win32PrivateInfo	*info = (_Win32PrivateInfo *)device->aux_data;
	int			buttonvalue;

	if (!strcasecmp(device_name, "mouse"))
	{
		buttonvalue = _Win32ButtonValue(dti->instance);

		if (buttonvalue <= 0 || buttonvalue > MAX_BUTTONS)
			return VRINPUT_MATCH_UNABLE;

		info->button_inputs[buttonvalue] = input;
		return VRINPUT_MATCH_ABLE;
	}

	return VRINPUT_NOMATCH;
}


/****************************************************************************/
static vrInputMatch _Win32PointerInput(vrInputDevice *device, vrGenericInput *input, vrInputDTI *dti)
{
	_Win32PrivateInfo	*info = (_Win32PrivateInfo *)device->aux_data;
	int			pointervalue;
	int			sign = 1;

	if (dti->instance[0] == '-') {
		pointervalue = _Win32PointerValue(dti->instance + 1);
		sign = -1;
	} else {
		pointervalue = _Win32PointerValue(dti->instance);
	}

	if (pointervalue < 0 || pointervalue > MAX_POINTERS)
		return VRINPUT_MATCH_UNABLE;

	info->pointer_inputs[pointervalue] = input;
	info->pointer_sign[pointervalue] = sign;

	return VRINPUT_MATCH_ABLE;
}


	/************************************************************/
	/***************** FreeVR Callback routines *****************/
	/************************************************************/

	/************************************************************/
	/*    Callbacks for controlling the MS-Windows features.    */
	/*                                                          */

#if 0 /* delete me */
/************************************************************/
static void _Win32SystemPauseToggleCallback(vrInputDevice *devinfo, int value)
{
	_Win32PrivateInfo	*aux = (_Win32PrivateInfo *)devinfo->aux_data;

        if (value == 0)
                return;

	if (!aux->silent)
		_Win32BellCallback(devinfo, 1);

	devinfo->context->paused ^= 1;
	vrDbgPrintfN(SELFCTRL_DBGLVL, "Win32 Control: system_pause = %d.\n",
		devinfo->context->paused);
}

/****************************************************************************/
static void _Win32PrintContextCallback(vrInputDevice *device, int value)
{
	_Win32PrivateInfo	*info = (_Win32PrivateInfo *)device->aux_data;

	if (value == 0)
		return;

	if (!info->silent)
		_Win32BellCallback(device, 1);

	vrFprintContext(stdout, verbose);
}
#endif

/****************************************************************************/
static void _Win32BellCallback(vrInputDevice *device, int value)
{
	if (value == 0)
		return;

	MessageBeep(-1);
}

/****************************************************************************/
static void _Win32SoftBellCallback(vrInputDevice *device, int value)
{
	if (value == 0)
		return;

	MessageBeep(MB_ICONASTERISK);
}



	/*******************************************************/
	/*   Callbacks for interfacing with the Win32 device.  */
	/*                                                     */


/****************************************************************************/
static void _Win32Create(vrInputDevice *device)
{
	/*** List of possible inputs ***/
static	vrInputFunction	_Win32Inputs[] = {
				{ "key",     VRINPUT_2WAY,     _Win32KeyInput },
				{ "button",  VRINPUT_2WAY,     _Win32ButtonInput },
				{ "pointer", VRINPUT_VALUATOR, _Win32PointerInput },
				{ NULL,      VRINPUT_UNKNOWN,  NULL } };
	/*** List of control functions ***/
static	vrControlFunc	_Win32ControlList[] = {
				/* overall system controls */
				{ "system_pause_toggle", vrInputSystemPauseToggleCallback },

				/* informational output controls */
				{ "print_context", vrInputPrintContextStructCallback },
				{ "print_config", vrInputPrintConfigStructCallback },
				{ "print_input", vrInputPrintInputStructCallback },

		/** TODO: other callback control functions go here **/
				/* simulated 6-sensor selection controls */
				/* simulated 6-sensor manipulation controls */

				/* other controls */
				{ "bell", _Win32BellCallback },
				{ "softbell", _Win32SoftBellCallback },

				/* end of the list */
				{ NULL, NULL } };

	_Win32PrivateInfo	*info;

	/******************************************/
	/* allocate and initialize auxiliary data */
	info = (_Win32PrivateInfo *)vrShmemAlloc0(sizeof(*info));
	device->aux_data = info;

	info->hWnd = _Win32NewHiddenWindow();

	vrInputCreateDataContainers(device, _Win32Inputs);
	vrInputCreateSelfControlContainers(device, _Win32Inputs, _Win32ControlList);
	/* TODO: anything to do for NON self-controls?  Implement for Magellan first. */

	if (input_device_lock) {
		int	count;

		vrSetWriteLock(input_device_lock);

		for (count = 0; count < MAX_INPUT_DEVICES; count++)
			if (input_devices[count] == NULL) {
				input_devices[count] = info;
				break;
			}

		vrReleaseWriteLock(input_device_lock);

		if (count == MAX_INPUT_DEVICES)
			vrErr("WIN32: Too many input devices registered.\n");
	} else {
		vrErr("WIN32: Input device list not initialized.\n");
	}

	device->created = 1;
}


/****************************************************************************/
static void _Win32Open(vrInputDevice *device)
{
	_Win32PrivateInfo	*info = (_Win32PrivateInfo *)device->aux_data;
	vrConfigInfo		*vrConfig;

	vrConfig = device->context->config;

	if (vrConfig->num_windows == 0) {
		vrErr("WIN32: Input device could not find any windows.\n");
		device->operating = 1;
		return;
	}

	info->window = vrObjectArraySearch(VROBJECT_WINDOW,
						(vrObjectInfo **)(vrConfig->windows),
						vrConfig->num_windows,
						"default");

	if (!info->window || !vrIsWGLWindow(info->window)) {
		vrErr("WIN32: Input window does not exist or is not WGL.\n");
		return;
	}

	vrDbgPrintfN(ALWAYS_DBGLVL, "WIN32: Attaching to WGL window "
			"'%s' [%p]\n", info->window->name, info->window);

	if (!vrWGLRegisterInputHandler(info->window, info->hWnd)) {
		vrErr("WIN32: Cannot register input handler with WGL.\n");
		return;
	}

	vrDbgPrintfN(ALWAYS_DBGLVL, "WIN32: Input handler registered.\n");

	device->operating = 1;
}


/****************************************************************************/
static void _Win32Close(vrInputDevice *device)
{
	_Win32PrivateInfo *info = (_Win32PrivateInfo *)device->aux_data;

	if (input_device_lock) {
		int	count;

		vrSetWriteLock(input_device_lock);

		for (count = 0; count < MAX_INPUT_DEVICES; count++)
			if (input_devices[count] == info)
				input_devices[count] = NULL;

		vrReleaseWriteLock(input_device_lock);
	}

	/* TODO: ... */
}


/****************************************************************************/
static void _Win32Reset(vrInputDevice *device)
{
	_Win32PrivateInfo *info = (_Win32PrivateInfo *)device->aux_data;

	/* TODO: ... */
}


/****************************************************************************/
static void _Win32Poll(vrInputDevice *device)
{
	_Win32PrivateInfo *info = (_Win32PrivateInfo *)device->aux_data;
	MSG msg;


	while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
		if (msg.message == WM_QUIT) {
			return;
		} else {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
}


/****************************************************************************/
void vrWin32InitInfo(vrInputDevice *device)
{
	vrDbgPrintfN(DEFAULT_DBGLVL,
			"WIN32: Initializing Callback and Version info\n");

	if (!input_device_lock)
#if 0
		input_device_lock = vrCreateTracedLock(vrContext, "win32 input device");
#else
		input_device_lock = vrLockCreateName(vrContext, "win32 input device");
#endif

	device->version = (char *)vrShmemStrDup("Win32 Input Device, v0.1");

	device->Create = vrCallbackCreateNamed("Win32Inp:Create-Def", _Win32Create, 1, device);
	device->Open = vrCallbackCreateNamed("Win32Inp:Open-Def", _Win32Open, 1, device);
	device->Close = vrCallbackCreateNamed("Win32Inp:Close-Def", _Win32Close, 1, device);
	device->Reset = vrCallbackCreateNamed("Win32Inp:Reset-Def", _Win32Reset, 1, device);
	device->PollData = vrCallbackCreateNamed("Win32Inp:PollData-Def", _Win32Poll, 1, device);
	device->PrintAux = vrCallbackCreateNamed("Win32Inp:PrintAux-Def", vrDoNothing, 0);
}

