/* ======================================================================
 *
 * HH   HH         vr_math.h
 * HH   HH         Author(s): Ed Peters, Stuart Levy, Bill Sherman, John Stone
 * HHHHHHH         Created: June 4, 1998
 * HH   HH         Last Modified: May 5, 2021 (added vrQuatf 04/21/22)
 * HH   HH
 *
 * Header file for FreeVR math routines.
 *
 * Copyright 2022, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
/* NOTE: FreeVR operates in the OpenGL Y-UP coordinate system. */
#ifndef __VRMATH_H__
#define __VRMATH_H__

#include <stdio.h>		/* for FILE type definition */

#ifdef __cplusplus
extern "C" {
#endif


/* translational / homogeneous matrix elements (and standard quaternions) */
#define VR_X 0
#define VR_Y 1
#define VR_Z 2
#define VR_W 3

/* Euler ego-centric rotational elements */
#define	VR_AZIM 0	/* rotation about Y (upward) axis     */
#define	VR_ELEV 1	/* rotation about X (horizontal) axis */
#define	VR_ROLL 2	/* rotation about Z (inward) axis     */

/* quaternion elements -- when following translational elements in a sequence of 7 values */
#define VR_QX (VR_X + 3)
#define VR_QY (VR_Y + 3)
#define VR_QZ (VR_Z + 3)
#define VR_QW (VR_W + 3)

#if !defined(PI)
#define PI	3.14159265358979323846
#endif

#define DTOR(x)	((x) * (PI / 180.0))
#define RTOD(x)	((x) * (180.0 / PI))

#define VR_EPSILON 0.00000000001


/****************************************/
typedef enum {
		VRDIR_UNKNOWN = -1,
		VRDIR_FORE,
		VRDIR_BACK,
		VRDIR_UP,
		VRDIR_DOWN,
		VRDIR_LEFT,
		VRDIR_RIGHT
	} vrDirection;


/***************************************************************************/
/* vrMatrix: a 4x4 matrix stored as a simple 16 element array of doubles.  */
/*   The VRMAT_ROWCOL (primarily) and VRMAT_COLROW macros can be used to   */
/*   access specific elements based on the column and row coordinates.     */
/* NOTE: The layout of the array in matrix form is:                        */
/*                                                                         */
/*   v[0]   v[4]   v[8]  v[12]         rc(x,x)  rc(x,y)  rc(x,z)  rc(x,w)  */
/*   v[1]   v[5]   v[9]  v[13]    =    rc(y,x)  rc(y,y)  rc(y,z)  rc(y,w)  */
/*   v[2]   v[6]  v[10]  v[14]         rc(z,x)  rc(z,y)  rc(z,z)  rc(z,w)  */
/*   v[3]   v[7]  v[11]  v[15]         rc(w,x)  rc(w,y)  rc(w,z)  rc(w,w)  */
/*                                                                         */
/* which is determined by the VRMAT_ROWCOL and VRMAT_COLROW macros.        */
typedef struct {
		double	v[16];
	} vrMatrix;
#define VRMAT_MAXDIM 4
#define VRMAT_ROWCOL(m,r,c) ((m)->v[(r)+VRMAT_MAXDIM*(c)])
#define VRMAT_COLROW(m,c,r) ((m)->v[(r)+VRMAT_MAXDIM*(c)])


/*********************************************************/
/* This structure is used for breaking down a 4x4 matrix */
/*   into its component operations                       */
typedef struct {
		double	scale[3];
		double	shear[3];	/* XY, XZ, YZ */
		double	rotate[3];
		double	quaternion[4];
		double	translate[3];
		double	perspective[4];	/* X, Y, Z, W */
	} vrMatrixComposition;


/****************************************/
/* data for storing a rendering frustum */
typedef struct {
		double	left;
		double	right;
		double	bottom;
		double	top;
		double	near_clip;	/* NOTE: I prefer "near" but Cygwin compiler can't handle it */
		double	far_clip;	/* NOTE: I prefer "far" but Cygwin compiler can't handle it */
	} vrFrustum;


/*************************************************************/
/* vrPoint & vrPointf : ...   */
typedef struct {
		double	v[3];
	} vrPoint;
typedef struct {
		float	v[3];
	} vrPointf;

#define vrPointSet3d(pnt,x,y,z) \
	if ((pnt)!=NULL) { (pnt)->v[VR_X]=(x); (pnt)->v[VR_Y]=(y); (pnt)->v[VR_Z]=(z); }
#define vrPointSetAd(pnt,a) \
	if ((pnt)!=NULL) { (pnt)->v[VR_X]=(a)[VR_X]; (pnt)->v[VR_Y]=(a)[VR_Y]; (pnt)->v[VR_Z]=(a)[VR_Z]; }


/*************************************************************/
/* vrVector, vrVectorf & vrVector4: ...   */
typedef struct {
		double	v[3];
	} vrVector;
typedef struct {
		float	v[3];
	} vrVectorf;

#define vrVectorSet3d(vec,x,y,z) \
	if ((vec)!=NULL) { (vec)->v[VR_X]=(x); (vec)->v[VR_Y]=(y); (vec)->v[VR_Z]=(z); }
#define vrVectorSetAd(vec,a) \
	if ((vec)!=NULL) { (vec)->v[VR_X]=(a)[VR_X]; (vec)->v[VR_Y]=(a)[VR_Y]; (vec)->v[VR_Z]=(a)[VR_Z]; }


typedef struct {
		double	v[4];
	} vrVector4;

#define vrVector4Set4d(vec,x,y,z,w) \
	if ((vec)!=NULL) { (vec)->v[VR_X]=(x); (vec)->v[VR_Y]=(y); (vec)->v[VR_Z]=(z); (vec)->v[VR_W]=(w); }
#define vrVector4SetAd(vec,a) \
	if ((vec)!=NULL) { (vec)->v[VR_X]=(a)[VR_X]; (vec)->v[VR_Y]=(a)[VR_Y]; (vec)->v[VR_Z]=(a)[VR_Z]; (vec)->v[VR_W]=(a)[VR_W]; }


/*************************************************************/
/* vrQuat: ...   */
typedef struct {
		double v[4];
	} vrQuat;

typedef struct {
		float v[4];
	} vrQuatf;

/* vrQuatLengthSq() -- return the length-squared of the quaternion.  Useful for normalizing a quaternion */
#define vrQuatLengthSq(q) \
	( ((q)==NULL) ? -1 : \
	(((q)->v[VR_X]*(q)->v[VR_X])+((q)->v[VR_Y]*(q)->v[VR_Y])+((q)->v[VR_Z]*(q)->v[VR_Z])+((q)->v[VR_W]*(q)->v[VR_W])))


/*********************************************************************/
/* vrVQSTransform: a structure that uses UNC's VQS method of holding */
/*   transform information.  A VQS transform combines the operations */
/*   of translation, rotation and scaling.                           */
/* NOTE: at the moment FreeVR does not use this structure, but if    */
/*   feasible, it might be beneficial.                               */
/*********************************************************************/
typedef struct {
		vrVector	v;
		vrQuat		q;
		double		s;
	} vrVQSTransform;


/**************************************************************/
/* vrEuler:  a type containing 6 Euler values that give       */
/*   a particular positional value.  In FreeVR, their use     */
/*   is intentionally limited to only be used when necessary, */
/*   which is generally when some input device reports its    */
/*   values in Euler angles.                                  */
/*                                                            */
/* NOTE: vrEuler uses floats in order to match how information*/
/*   is stored in shared memory by the CAVE tracker daemon.   */
/**************************************************************/
typedef struct {
		float	t[3];	/* translational component */
		float	r[3];	/* rotational component    */
	} vrEuler;


/*************************/
/*** Matrix operations ***/
/* Defined are a series of operations to work on 4x4 matrices,      */
/* laid out as a 16-element float vector (defined as column-major   */
/* for compatibility * with GL).  The API includes functions to     */
/* create and destroy matrices, initialize their values and compose */
/* them with multiplication.                                        */
vrMatrix	*vrMatrixCreate(void);
void		 vrMatrixDelete(vrMatrix *mat);
int		 vrMatrixEqual(vrMatrix *mat_l, vrMatrix *mat_r);
vrMatrix	*vrMatrixCopy(vrMatrix *mat_dst, vrMatrix *mat_src);
vrMatrix	*vrMatrixSetIdentity(vrMatrix *mat_dst);
vrMatrix	*vrMatrixCreateIdentity(void);
vrMatrix	*vrMatrixSetIdentityRotationOnly(vrMatrix *mat_dst);
vrMatrix	*vrMatrixSetAd(vrMatrix *mat_dst, double *array);
vrMatrix	*vrMatrixSetXYZAERAf(vrMatrix *mat_dst, float *array);
vrMatrix	*vrMatrixSetXYZAERAd(vrMatrix *mat_dst, double *array);
vrMatrix	*vrMatrixSetXYZAERE(vrMatrix *mat_dst, vrEuler *euler);
vrMatrix	*vrMatrixSetEulerAzimaxis(vrMatrix *mat_dst, vrEuler *euler, int axis);
vrEuler		*vrEulerSetFromMatrix(vrEuler *euler_dst, vrMatrix *mat_src);
vrMatrix	*vrMatrixSetFromCoords(vrMatrix *mat_dst, double *coords);
vrMatrix	*vrMatrixSetTranslation3d(vrMatrix *mat_dst, double x, double y, double z);
vrMatrix	*vrMatrixSetTranslationAd(vrMatrix *mat_dst, double *array);
vrMatrix	*vrMatrixSetTranslationOnlyAd(vrMatrix *mat_dst, double *array);
vrMatrix	*vrMatrixGetResetTranslationAd(vrMatrix *mat_dst, double *array);
int		 vrMatrixCompositionFromMatrix(vrMatrixComposition *comp, vrMatrix *mat_src);
vrMatrix	*vrMatrixSetRotationId(vrMatrix *mat_dst, int axis, double theta);
vrMatrix	*vrMatrixSetRotation4d(vrMatrix *mat_dst, double x, double y, double z, double theta);
vrMatrix	*vrMatrixSetRotationAd(vrMatrix *mat_dst, double *array);
vrMatrix	*vrMatrixSetScale3d(vrMatrix *mat_dst, double sx, double sy, double sz);

vrMatrix	*vrMatrixProduct(vrMatrix *mat_dst, vrMatrix *mat_l, vrMatrix *mat_r);

vrMatrix	*vrMatrixPreMult(vrMatrix *mat_l, vrMatrix *mat_r);
vrMatrix	*vrMatrixPreTranslate3d(vrMatrix *mat_l, double x, double y, double z);
vrMatrix	*vrMatrixPreTranslateAd(vrMatrix *mat_l, double *array);
vrMatrix	*vrMatrixPreRotateId(vrMatrix *mat_l, int axis, double theta);
vrMatrix	*vrMatrixPreRotateAboutOriginId(vrMatrix *mat_l, int axis, double theta);
vrMatrix	*vrMatrixPreRotateAboutPointId(vrMatrix *mat_l, vrPoint *point, int axis, double theta);
vrMatrix	*vrMatrixPreRotate4d(vrMatrix *mat_l, double x, double y, double z, double theta);
vrMatrix	*vrMatrixPreScale3d(vrMatrix *mat_l, double sx, double sy, double sz);

vrMatrix	*vrMatrixPostMult(vrMatrix *mat_l, vrMatrix *mat_r);
vrMatrix	*vrMatrixPostTranslate3d(vrMatrix *mat_l, double x, double y, double z);
vrMatrix	*vrMatrixPostTranslateAd(vrMatrix *mat_l, double *array);
vrMatrix	*vrMatrixPostRotateId(vrMatrix *mat_l, int axis, double theta);
vrMatrix	*vrMatrixPostRotateAboutOriginId(vrMatrix *mat_l, int axis, double theta);
vrMatrix	*vrMatrixPostRotateAboutPointId(vrMatrix *mat_l, vrPoint *point, int axis, double theta);
vrMatrix	*vrMatrixPostRotate4d(vrMatrix *mat_l, double x, double y, double z, double theta);
vrMatrix	*vrMatrixPostScale3d(vrMatrix *mat_l, double sx, double sy, double sz);

vrMatrix	*vrMatrixTranspose(vrMatrix *mat_dst, vrMatrix *mat_src);
vrMatrix	*vrMatrixInvert(vrMatrix *mat_dst, vrMatrix *mat_src);
vrMatrix	*vrMatrixInvertInPlace(vrMatrix *mat_dst);
int		 vrMatrixIsEuclidean(vrMatrix *mat_src);
vrMatrix	*vrMatrixMakeEuclidean(vrMatrix *mat_dst, vrMatrix *mat_src);
vrMatrix	*vrMatrixInvertEuclidean(vrMatrix *mat_dst, vrMatrix *mat_src);
double		 vrMatrixGetScale(vrMatrix *mat_src);
vrMatrix	*vrMatrixSetScale(vrMatrix *mat_dst, vrMatrix *mat_src, double scale);
double		 vrMatrixDeterminant(vrMatrix *mat_src);
double		 vr3x3Determinant(double a1, double a2, double a3, double b1, double b2, double b3, double c1, double c2, double c3);
double		 vr2x2Determinant(double a, double b, double c, double d);


/*****************************/
/*** Quaternion operations ***/
/* Defined are routines for creating and manipulating quaternions,  */
/* laid out as an array of 4 doubles [X,Y,Z,W].  Quaternions are an */
/* especially succinct way of representing 3d rotations, with many  */
/* advantages over traditional matrices and Euler angles.  A full   */
/* description of quaternions is beyond the scope of this comment.  */
/* NOTE: Quaternion components are scalar-last: [X*i, Y*j, Z*k, W]. */
vrQuat		*vrQuatCreate(void);
vrQuat		*vrQuatSet4dThetaAxis(vrQuat *quat_dst, double theta, double x, double y, double z);
vrQuat		*vrQuatSet4d(vrQuat *quat_dst, double x, double y, double z, double w);
vrQuat		*vrQuatProduct(vrQuat *quat_dst, vrQuat *ql, vrQuat *qr);
vrQuat		*vrQuatProductEquals(vrQuat *quat_dst, vrQuat *ql, vrQuat *qr);
vrQuat		*vrQuatCopy(vrQuat *quat_dst, vrQuat *quat_src);
void		 vrQuatDelete(vrQuat *quat);


/*************************/
/*** Vector operations ***/
vrVector	*vrVectorCopy(vrVector *vec_dst, vrVector *vec_src);
vrVector	*vrVectorFromTwoPoints(vrVector *vec_dst, vrPoint *pnt1, vrPoint *pnt2);
vrVector	*vrVectorNormalize(vrVector *vec_dst, vrVector *vec_src);
double		 vrVectorLength(vrVector *vec_src);
double		 vrVectorDotProduct(vrVector *vec1, vrVector *vec2);
vrVector	*vrVectorCrossProduct(vrVector *vec_dst, vrVector *vec1, vrVector *vec2);
vrVector	*vrVectorScale(vrVector *vec_dst, double scale);
vrVector	*vrVectorSubtract(vrVector *vec_dst, vrVector *vec1, vrVector *vec2);
vrVector	*vrVectorAddScaledVectors(vrVector *vec_dst, vrVector *vec1, vrVector *vec2, double scale1, double scale2);


/************************/
/*** Point operations ***/
/* These are really defined for the sake of completeness; */
/* it's probably more of a pain to use them than not.     */
vrPoint 	*vrPointCreate(void);
vrPoint		*vrPointCreate3d(double x, double y, double z);
vrPoint		*vrPointCopy(vrPoint *pnt_dst, vrPoint *pnt_src);
void		 vrPointDelete(vrPoint *pnt);
vrPoint		*vrPointAddScaledVector(vrPoint *pnt_dst, vrPoint *pnt_src, vrVector *vec_src, double scale);
void		 vrCoordsSetIdentity(double *);


/************************************************/
/*** TRANSFORMATION and CONVERSION operations ***/
/* Converting quats to mats, vice versa, and applying these */
/* transformations to points.                               */
#if 1 /* it's time to get rid of this mistake (changed my mind back) [04/09/21: perhaps delete the above instead] */
vrMatrix	*vrMatrixSetRotationFromQuat(vrMatrix *mat_dst, vrQuat *quat_src);
#endif
//vrMatrix	*vrMatrixSetFromQuatCR(vrMatrix *mat_dst, vrQuat *quat_src);
vrMatrix	*vrMatrixSetFromQuat(vrMatrix *mat_dst, vrQuat *quat_src);
vrQuat		*vrQuatSetFromMatrix(vrQuat *quat_dst, vrMatrix *mat_src);
vrEuler		*vrEulerSetFromQuatPoint(vrEuler *euler_dst, vrQuat *quat_src, vrPoint *point_src);

vrPoint		*vrPointTransformByQuat(vrPoint *dst_pnt, vrPoint *src_pnt, vrQuat *quat_src);
vrPoint		*vrPointTransformByMatrix(vrPoint *dst_pnt, vrPoint *pnt_src, vrMatrix *mat_src);
vrPoint		*vrPointGetTransFromMatrix(vrPoint *dst_pnt, vrMatrix *src_mat);
vrMatrix	*vrMatrixSetTransFromVector(vrMatrix *mat_dst, vrVector *vec_src);
vrVector	*vrVectorTransformByMatrix(vrVector *vec_dst, vrVector *vec_src, vrMatrix *mat_src);
vrVector4	*vrVector4TransformByMatrix(vrVector4 *dst_vec, vrVector4 *src_vec, vrMatrix *src_mat);
vrVector	*vrVectorGetTransFromMatrix(vrVector *dst_vec, vrMatrix *src_mat);


/**************************************/
/*** DEBUGGING AND UTILITY ROUTINES ***/
void		vrFprintMatrix(FILE *file, char *before, vrMatrix *mat_src);
void 		vrPrintMatrix(char *before, vrMatrix *mat_src);
void 		vrPrintMatrixT(char *before, vrMatrix *mat_src);
void		vrFprintMatrixT(FILE *file, char *before, vrMatrix *mat_src);
void 		vrPrintQuat(char *before, vrQuat *quat_src);
void 		vrPrintPoint(char *before, vrPoint *pnt_src);
void		vrPrintVector(char *before, vrVector *vec_src);
void		vrPrintVector4(char *before, vrVector4 *vec_src);
void		vrPrintEuler(char *before, vrEuler *euler_src);
void		vrPrintMatrixComposition(char *before, vrMatrix *mat_src);
void		vrFprintMatrixCompositionOL(FILE *fp, char *before, vrMatrix *mat_src);

double		vrDegreesToRadians(double deg);
double		vrRadiansToDegrees(double rad);

#ifdef __cplusplus
}
#endif

#endif /* __VRMATH_H__ */
