/* ======================================================================
 *
 *  CCCCC          vr_utils.c
 * CC   CC         Author(s): John Stone, Bill Sherman
 * CC              Created: August 10, 2000
 * CC   CC         Last Modified: May 24, 2013 (minor/comment mods on 05/05/21)
 *  CCCCC
 *
 * Code file for miscellaneous machine and operating system dependent
 * utility routines.
 *
 * Copyright 2022, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */

#include <stdio.h> 		/* for snprintf() */
#include <string.h>		/* for myriad string functions */
#include <stdlib.h>		/* for system() and strtol() */
#include <ctype.h>		/* for isprint() */
#include <unistd.h>		/* for usleep() */

#include "vr_shmem.h"		/* for one call to vrShmemStrDup() */
#include "vr_utils.h"		/* for function declarations of this file */
#include "vr_debug.h"		/* Standard FreeVR debugging features */

#if defined(TEST_APP)
#  define	vrShmemStrDup	strdup
#endif


/***************************************************************************/
/* vrShellCmd(): this function makes a call to the system shell to execute */
/*   a standard command line command.  This is typically used to run       */
/*   commands that might start a sound server, or put a monitor into 3D    */
/*   stereomode when starting, and then normal mode when exiting.          */
/*                                                                         */
/* Three format substitutions are available:                               */
/*   * '#P' is substituted with the current process ID (pid)               */
/*   * '#N' is substituted with the current process name                   */
/*   * '#E' is substituted with an error code when terminating abnormally  */
/*                                                                         */
/* TOOD:                                                                   */
/*   * make sure the percentage (%) substitution works, looks suspect      */
/***************************************************************************/
void vrShellCmd(char *cmdstr, char *name, int pid, int error_code)
{
	if (cmdstr == NULL)
		return;

	if (strlen(cmdstr) > 1) {
		int	no_subst;		/* true when no substitution done thus far in a loop */
		char	syscmd[1024];
		char	syscmd2[1024];
		char	*subst = NULL;
		char	*percent_loc = NULL;

		strcpy(syscmd, cmdstr);

		/* loop until no substitutions were done */
		/*   NOTE: we do this to make sure multiple        */
		/*   instances of the same substitution will work. */
		do {
			no_subst = 1;

			/* replace "#P" with the process pid */
			subst = strstr(syscmd, "#P");
			if (subst != NULL) {
				/* replace all occurrences of '%' w/ "%%" first */
				while ((percent_loc = strchr(syscmd, '%')) != NULL) {
					strncpy(syscmd2, syscmd, (size_t)(percent_loc - syscmd));
					strcat(syscmd2, percent_loc);
				}

				/* now do the replacement */
				strncpy(syscmd2, syscmd, (subst - syscmd));
				syscmd2[subst - syscmd] = '\0';
				strcat(syscmd2, "%d");
				strcat(syscmd2, &(subst[2]));
				snprintf(syscmd, sizeof(syscmd), syscmd2, pid);

				no_subst = 0;
			}

			/* replace "#N" with the process name */
			subst = strstr(syscmd, "#N");
			if (subst != NULL) {
				/* replace all occurrences of '%' w/ "%%" first */
				while ((percent_loc = strchr(syscmd, '%')) != NULL) {
					strncpy(syscmd2, syscmd, (size_t)(percent_loc - syscmd));
					strcat(syscmd2, percent_loc);
				}

				/* now do the replacement */
				strncpy(syscmd2, syscmd, (subst - syscmd));
				syscmd2[subst - syscmd] = '\0';
				strcat(syscmd2, "%s");
				strcat(syscmd2, &(subst[2]));
				snprintf(syscmd, sizeof(syscmd), syscmd2, name);

				no_subst = 0;
			}

			/* replace "#E" with the error code */
			subst = strstr(syscmd, "#E");
			if (subst != NULL) {
				/* replace all occurrences of '%' w/ "%%" first */
				while ((percent_loc = strchr(syscmd, '%')) != NULL) {
					strncpy(syscmd2, syscmd, (size_t)(percent_loc - syscmd));
					strcat(syscmd2, percent_loc);
				}

				/* now do the replacement */
				strncpy(syscmd2, syscmd, (subst - syscmd));
				syscmd2[subst - syscmd] = '\0';
				strcat(syscmd2, "%d");
				strcat(syscmd2, &(subst[2]));
				snprintf(syscmd, sizeof(syscmd), syscmd2, error_code);

				no_subst = 0;
			}

		} while (no_subst == 0);

		system(syscmd);
	}
}


/*****************************************************************/
/*                       TIME routines                           */
/*****************************************************************/

/*************************************************************************/
/* vrSleep() takes a single argument, a time in micro-seconds, and calls */
/*    the system's sleep utility.  Unlike some system sleep utilities,   */
/*    vrSleep() can handle delays of longer than 1.0 second.             */
void vrSleep(long useconds) {

	if (useconds <= 0) {
#if defined(SGINAP)
		/* only use sginap() on IRIX 5.x where usleep() is not available */
		sginap(0);
#else
		usleep(0);	/* often just calling usleep() is important to release the CPU */
#endif
		return;
	}

#if defined(SGINAP)
	/* only use sginap() on IRIX 5.x where usleep is not available */
	sginap(useconds / 10000);
#else
	/* handles delays longer than one second */
	while (useconds >= 1000000) {
		usleep(999999);
		useconds -= 999999;
	}
#  if !defined(__linux) /* on systems that produce higher-res waits with calls less than 10000ms, this can improve resolution */
	while (useconds >= 5000) {
		usleep(5000);
		useconds -= 5000;
	}
#  endif
	usleep(useconds);
#endif
}


/*****************************************************************/
/*                  String Conversion routines                   */
/*****************************************************************/

/*********************************************************************/
/* NOTE: skips leading whitespace to function like atoi(), and   */
/*   make the test for other bases (eg. "0x") work properly.     */
long vrAtoI(char *str)
{
static	char	*whitespace = " \t\r\b\n";
	int	sign = 1;

	/* check that we got a string */
	if (str == NULL) {
		vrErrPrintf("vrAtoI(): argument is NULL\n");
		return 0;
	}

	str += strspn(str, whitespace);		/* skip leading whitespace */

	/* make sure we got an actual number */
	if ((!isdigit(str[0])) && (str[0] != '-')) {
		vrDbgPrintfN(DEFAULT_DBGLVL, "vrAtoI(\"%s\"): argument is non-numeric\n", str);
		return 0;
	}

	/* check for minus sign */
	if (strlen(str) > 1) {
		if (str[0] == '-') {
			sign = -1;		/* set the sign to negative */
			str++;			/* now skip the minus sign */
		}
	}

	/* check for radix */
	if (strlen(str) > 2) {
		if (str[0] == '0') {
			switch(str[1]) {
			case 'b':
				return (sign * strtol(&str[2], (char **)NULL, 2));
			case 'o':
				return (sign * strtol(&str[2], (char **)NULL, 8));
			case 'd':
				return (sign * strtol(&str[2], (char **)NULL, 10));
			case 'x':
				return (sign * strtol(&str[2], (char **)NULL, 16));
			default:
				return (sign * strtol(str, (char **)NULL, 0));
			}
		}
	}

	/* otherwise return with the default strtol() behavior */
	return (sign * strtol(str, (char **)NULL, 0));
}


/*********************************************************************/
/* This version returns an integer to be used as a pointer -- which */
/*   may require larger integers.                                   */
/* NOTE: on Linux, strtol() works on both 32 and 64 bit architectures */
/*   the compiler does the right thing.  I'm not sure whether this is */
/*   true for other machines -- I was using strtoll(), and I suspect  */
/*   that was for a good reason.                                      */
/*   And my notes from 03/13/03 indicate that it was tests on a big   */
/*   Sun machine that required strtoll(), or it would truncate the    */
/*   number -- now, this stuff only matters when entering addresses   */
/*   via the telnet interface.                                        */
void *vrAtoP(char *str)
{
static	char	*whitespace = " \t\r\b\n";

	str += strspn(str, whitespace);		/* skip leading whitespace */

	if (strlen(str) > 2) {
		if (str[0] == '0') {
			switch(str[1]) {
			case 'b':
				return ((void *)strtol(&str[2], (char **)NULL, 2));
			case 'o':
				return ((void *)strtol(&str[2], (char **)NULL, 8));
			case 'd':
				return ((void *)strtol(&str[2], (char **)NULL, 10));
			case 'x':
				return ((void *)strtol(&str[2], (char **)NULL, 16));
			default:
				return ((void *)strtol(str, (char **)NULL, 0));
			}
		} else {
			return ((void *)strtol(str, (char **)NULL, 0));
		}
	} else {
		return ((void *)strtol(str, (char **)NULL, 0));
	}
}


/*************************************************************************/
/* vrBinaryToString(): takes an integer, a number of bits, and a pointer */
/*   to memory in which to store the result.  The resultant string       */
/*   is placed in the memory, and returned.  The string is made up       */
/*   of the least significant N bits, based on the value of len.         */
/*************************************************************************/
char *vrBinaryToString(unsigned int val, int len, char *str)
{
	int     count;

	str[len] = '\0';
	for (count = 0; count < len; count++) {
		str[len-count-1] = (((1 << count) & val) ? '1' : '0');
	}
	return (str);
}


/**********************************************************************/
/* Convert special character sequences into new forms.  E.g. a "\n"   */
/*   sequence into a newline character.                               */
/* TODO: probably shouldn't have the must-be-shorter-than-4096-byte   */
/*   restriction.                                                     */
char *vrProcessString(char *string)
{
	if (string == NULL)
		return NULL;

	if (strlen(string) > 4096) {
		vrDbgPrintfN(1, "vrProcessString(): can't handle strings longer than 4096 bytes.\n");
		return (string);
	}

	if (strlen(string) > 1) {
		int	no_subst;		/* true when no substitution done thus far in a loop */
		char	str1[4096];
		char	str2[4096];
		char	*subst = NULL;
		char	*percent_loc = NULL;

		strcpy(str1, string);

		/* loop until no substitutions were done */
		/*   NOTE: we do this to make sure multiple        */
		/*   instances of the same substitution will work. */
		do {
			no_subst = 1;

			/* replace "\n" with the newline character */
			subst = strstr(str1, "\\n");
			if (subst != NULL) {
				/* now do the replacement */
				strncpy(str2, str1, (subst - str1));
				str2[subst - str1] = '\0';
				strcat(str2, "\n");
				strcat(str2, &(subst[2]));
				strcpy(str1, str2);

				no_subst = 0;
			}

		} while (no_subst == 0);

		return (vrShmemStrDup(str1));		/* TODO: consider whether this really needs to be in shared memory */
	}

	/* TODO: test whether we can get here.  (Though need this line for pickier compilers anyway) */
	return string;
}


/**********************************************************************/
/* Convert non-printable strings to print a "." instead.              */
/* TODO: probably shouldn't have the must-be-shorter-than-4096-byte   */
/*   restriction.                                                     */
char *vrPrintableString(char *string)
{
	int	count;
	char	str1[4096];

	if (string == NULL)
		return NULL;

	if (strlen(string) > 4096) {
		vrDbgPrintfN(1, "vrPrintableString(): can't handle strings longer than 4096 bytes.\n");
		return (string);
	}

	for (count = 0; count < strlen(string); count++) {
		if (isprint(string[count]))
			str1[count] = string[count];
		else	str1[count] = '_';
	}
	str1[count] = string[count];	/* now copy the terminating '\0' */
	
	return (vrShmemStrDup(str1));		/* TODO: consider whether this really needs to be in shared memory */
}


/***************************************************************/
/* vrStringCharWidth(): loops through all the lines of text in */
/*   a string (ie. each newline), and reports the widest line  */
/*   (which can be interpreted as the character-width of the   */
/*   string).                                                  */
int vrStringCharWidth(char *str)
{
	int	count;
	int	max_width = 0;
	int	curr_width = 0;
	int	str_len = strlen(str);

	for (count = 0; count < str_len; count++) {
		if (str[count] == '\n') {
			if (curr_width > max_width)
				max_width = curr_width;
			curr_width = 0;
		} else {
			curr_width++;
		}
	}

	return max_width;
}


/**************************************************************/
/* vrStringCharHeight(): loops through a string, counting the */
/*   number of newline characters (which can be interpreted   */
/*   as the character-height of the string).                  */
int vrStringCharHeight(char *str)
{
	int	count;
	int	str_len = strlen(str);
	int	result = 1;			/* there is at least one line */

	for (count = 0; count < str_len; count++) {
		if (str[count] == '\n')
			result++;
	}

	return result;
}


/******************************************************************/
/* Okay, this was intended to search a list of arbitrary length   */
/*   strings, but of course that was a ridiculous expectation of  */
/*   C.  So for now, don't use this function.  There must be a    */
/*   reasonable solution, but for the moment we'll have a version */
/*   that requires 256 length strings.                            */
int vrSearchStringList(char *item, int list_length, char *list[])
{
	int	count;

	vrDbgPrintfN(1, "vrSearchStringList(%s, %d, %#p)\n", item, list_length, list);
	for (count = 0; count < list_length; count++) {
		vrDbgPrintfN(1, "...vrSearchStringList() %d: comparing item with '%s'(%#p)\n", count, list[count], list[count]);
		if (!strcmp(item, list[count])) {
			vrDbgPrintfN(1, "...vrSearchStringList() returning %d\n", count);
			return count;
		}
	}

	vrDbgPrintfN(1, "...vrSearchStringList() returning -1\n");
	return -1;
}


/************************************************************/
/* a version of the above that assumes (ie. requires) that the */
/*   strings all have a length of 256 bytes.                   */
int vrSearch256StringList(char *item, int list_length, char list[][256])
{
	int	count;

	vrDbgPrintfN(1, "vrSearch256StringList(%s, %d, %#p)\n", item, list_length, list);
	for (count = 0; count < list_length; count++) {
		vrDbgPrintfN(1, "...vrSearchStringList() %d: comparing item with '%s'(%#p)\n", count, list[count], list[count]);
		if (!strcmp(item, list[count])) {
			vrDbgPrintfN(1, "...vrSearch256StringList() returning %d\n", count);
			return count;
		}
	}

	vrDbgPrintfN(1, "...vrSearch256StringList() returning -1\n");
	return -1;
}


/*****************************************************************/
/*                    Data Conversion routines                   */
/*****************************************************************/

/****************************************************************************/
/* vrEndianSwap4(): does an in-place Endian swap of a 4-byte integer number */
void vrEndianSwap4(void *bytes)
{
	char	temp0, temp1;
	char	*cbytes = (char *)bytes;

	temp0 = cbytes[0];
	temp1 = cbytes[1];
	cbytes[0] = cbytes[3];
	cbytes[1] = cbytes[2];
	cbytes[2] = temp1;
	cbytes[3] = temp0;
}

/*****************************************************************************/
/* vrEndianSwap8(): does an in-place Endian swap of an 8-byte integer number */
void vrEndianSwap8(void *bytes)
{
	char	temp;
	char	*cbytes = (char *)bytes;

	temp = cbytes[0];
	cbytes[0] = cbytes[7];
	cbytes[7] = temp;

	temp = cbytes[1];
	cbytes[1] = cbytes[6];
	cbytes[6] = temp;

	temp = cbytes[2];
	cbytes[2] = cbytes[5];
	cbytes[5] = temp;

	temp = cbytes[3];
	cbytes[3] = cbytes[4];
	cbytes[4] = temp;
}

