/* ======================================================================
 *
 *  CCCCC          vr_config.c
 * CC   CC         Author(s): Ed Peters, Bill Sherman
 * CC              Created: June 4, 1998
 * CC   CC         Last Modified: September 29, 2021
 *  CCCCC
 *
 * Code file for FreeVR configuration and context information.
 *
 * Copyright 2022, Bill Sherman, All rights reserved.
 * With the intent to provide an open-source license to be named later.
 * ====================================================================== */
#include <stdlib.h>		/* used by getenv() */
#define __USE_GNU		/* needed for string.h to include strchrnul() -- NOTE: this must come after <stdlib.h>, I don't know why. */
#include <string.h>		/* needed for strchrnul() and lots of other string functions */

#include "vr_input.h"
#include "vr_config.h"
#include "vr_entity.h"
#include "vr_objects.h"
#include "vr_parse.h"
#include "vr_system.h"
#include "vr_utils.h"		/* for vrSleep() */
#include "vr_debug.h"


/* eventually delete me -- but not this round */
#undef EXTRA_DEBUG
#ifdef EXTRA_DEBUG
#  define CONDITIONAL_DEBUG 23+0*
#else
#  define CONDITIONAL_DEBUG 0+
#endif

/*****************************************/
/*** function-as-argument declarations ***/
typedef vrTokenInfo (*_ConfigParseOptionList)(vrConfigInfo *, vrParseInfo *, vrTokenInfo, vrObjectInfo *, void *);

/************************************************************************/
/* function declarations of local-only functions called before defined. */
static vrTokenInfo	_ConfigParseGlobalOption(vrConfigInfo *config, vrParseInfo *parse_data, vrTokenInfo token, vrObjectInfo *object, _ConfigParseOptionList objparser);
static vrTokenInfo	_ConfigParseStatement(vrConfigInfo *config, vrParseInfo *config_parse, vrObjectInfo *object, _ConfigParseOptionList objparser);
static vrTokenInfo	_ConfigParseInputDescription(vrConfigInfo *config, vrParseInfo *parse, vrInputDevice *device);
static vrTokenInfo	_ConfigParseControlDescription(vrConfigInfo *config, vrParseInfo *parse, vrInputDevice *device);
static int		_ConfigReadFile(vrConfigInfo *config, char *path, char *file);
static int		_ConfigReadString(vrConfigInfo *config, char *info_string, char *conf_string);
static void		_ConfigReadStringList(vrConfigInfo *config, int *argc, char **argv);
static void		_ConfigReadEnvVariable(vrConfigInfo *config, char *name);


/*******************************************************************/
/* _ConfigTSMap[]: array that maps vrTokens to their corresponding */
/*   string/ASCII representations.                                 */
static vrTokenInfo _ConfigTSMap[] = {
	/* angle brackets are parsed as math punctuation, so these ten are for output only */
		{ VRTOKEN_PARSEERROR,	"<parse error>" },
		{ VRTOKEN_EOF,		"<eof>" },
		{ VRTOKEN_UNKNOWN,	"<unk>" },
		{ VRTOKEN_NIL,		"<nil>" },
		{ VRTOKEN_DEFAULT,	"<default>" },
		{ VRTOKEN_FALSE,	"<false/no>" },
		{ VRTOKEN_TRUE,		"<true/yes>" },
		{ VRTOKEN_STRING,	"<string>" },
		{ VRTOKEN_NUMBER,	"<number>" },
		{ VRTOKEN_VARIABLE,	"<variable>" },

	/* normal punctuation */
		{ VRTOKEN_HASH,		"#" },
		{ VRTOKEN_OPENCURL,	"{" },
		{ VRTOKEN_CLOSECURL,	"}" },
		{ VRTOKEN_COMMA,	"," },
		{ VRTOKEN_SEMICOLON,	";" },
		{ VRTOKEN_QUOTE,	"\"" },

	/* math punctuation */
		{ VRTOKEN_OPENPAREN,	"(" },			/* NOTE: like normal punctuation, the open   */
		{ VRTOKEN_CLOSEPAREN,	")" },			/*   and close parens are single char tokens */

		{ VRTOKEN_PLUS,		"+" },
		{ VRTOKEN_TIMES,	"*" },
		{ VRTOKEN_DIVIDE,	"/" },
		{ VRTOKEN_BIT_AND,	"&" },
		{ VRTOKEN_BIT_OR,	"|" },
		{ VRTOKEN_LOGICAL_AND,	"&&" },
		{ VRTOKEN_LOGICAL_OR,	"||" },

		{ VRTOKEN_SETEQUALS,	"=" },
		{ VRTOKEN_PLUSEQUALS,	"+=" },
		{ VRTOKEN_TIMESEQUALS,	"*=" },

		{ VRTOKEN_EQUALITY,	"==" },
		{ VRTOKEN_INEQUALITY,	"!=" },
		{ VRTOKEN_GREATER,	">" },
		{ VRTOKEN_GREATEROREQ,	">=" },
		{ VRTOKEN_LESS,		"<" },
		{ VRTOKEN_LESSOREQ,	"<=" },
		/* NOTE: it might be possible to have a MINUS operand that coexists with    */
		/*   the "-" as part of the number -- if we're expecting a math expression, */
		/*   then the "-" can be assumed to be that, or if we're expecting a number */
		/*   then it's a number -- hmmm, maybe the code isn't quite formulated that */
		/*   way.                                                                   */

	/* keywords (the rest) */

	   /** system commands **/
		{ VRTOKEN_READCONFIG,	"readConfig" },
		{ VRTOKEN_READCONFIGSTR,"readConfigString" },
		{ VRTOKEN_NOP,		"nop" },
		{ VRTOKEN_EXIT,		"exit" },
		{ VRTOKEN_ABORT,	"abort" },
		{ VRTOKEN_EXEC,		"exec" },
		{ VRTOKEN_ECHO,		"echo" },
		{ VRTOKEN_SETENV,	"setenv" },
		{ VRTOKEN_UNSETENV,	"unSetenv" },
		{ VRTOKEN_PRINTOBJ,	"printObject" },
		{ VRTOKEN_PRINTCONFIG,	"printConfig" },
		{ VRTOKEN_GDBPAUSE,	"gdbPause" },

	   /** conditionals **/
		{ VRTOKEN_IF,		"if" },
		{ VRTOKEN_IFEXEC,	"ifExec" },
		{ VRTOKEN_ELSE,		"else" },

	   /** global options **/
		{ VRTOKEN_SET,		"set" },
		{ VRTOKEN_SETDEFAULT,	"setDefault" },
		{ VRTOKEN_USESYSTEM,	"useSystem" },

	   /** inheritable settings **/
		{ VRTOKEN_DEBUG_LEVEL,		"debugLevel" },
		{ VRTOKEN_DEBUG_EXACT,		"debugExact" },
		{ VRTOKEN_DEBUG_EXACT,		"debugThisToo" },

		{ VRTOKEN_EXECSTART,		"execAtStart" },
		{ VRTOKEN_EXECSTOP,		"execAtStop" },
		{ VRTOKEN_EXECUPONERROR,	"execUponError" },
		{ VRTOKEN_EXITUPONERROR,	"exitUponError" },

		{ VRTOKEN_SYSTEM_PRECONTEXT,	"PreContext" },
		{ VRTOKEN_SYSTEM_PRECONTEXT,	"PreContextPrint" },
		{ VRTOKEN_SYSTEM_PRECONFIG,	"PreConfig" },
		{ VRTOKEN_SYSTEM_PRECONFIG,	"PreConfigPrint" },
		{ VRTOKEN_SYSTEM_PREINPUT,	"PreInput" },
		{ VRTOKEN_SYSTEM_PREINPUT,	"PreInputPrint" },
		{ VRTOKEN_SYSTEM_POSTCONTEXT,	"PostContext" },
		{ VRTOKEN_SYSTEM_POSTCONTEXT,	"PostContextPrint" },
		{ VRTOKEN_SYSTEM_POSTCONFIG,	"PostConfig" },
		{ VRTOKEN_SYSTEM_POSTCONFIG,	"PostConfigPrint" },
		{ VRTOKEN_SYSTEM_POSTINPUT,	"PostInput" },
		{ VRTOKEN_SYSTEM_POSTINPUT,	"PostInputPrint" },

		{ VRTOKEN_SYSTEM_LOCKCPU,	"LockCPU" },
		{ VRTOKEN_SYSTEM_LOCKCMD,	"LockCmd" },
		{ VRTOKEN_SYSTEM_LOCKCMD,	"LockCommand" },
		{ VRTOKEN_SYSTEM_UNLOCKCMD,	"UnlockCmd" },
		{ VRTOKEN_SYSTEM_UNLOCKCMD,	"UnlockCommand" },

		{ VRTOKEN_SYSTEM_VISMODE,	"VisrenMode" },
		{ VRTOKEN_NEARCLIP,		"near_clip" },
		{ VRTOKEN_NEARCLIP,		"nearClip" },
		{ VRTOKEN_NEARCLIP,		"near" },
		{ VRTOKEN_FARCLIP,		"far_clip" },
		{ VRTOKEN_FARCLIP,		"farClip" },
		{ VRTOKEN_FARCLIP,		"far" },

	   /** object specifiers **/
		{ VRTOKEN_SYSTEM,	"system" },
		{ VRTOKEN_PROCESS,	"process" },
		{ VRTOKEN_WINDOW,	"window" },
		{ VRTOKEN_EYELIST,	"eyeList" },
		{ VRTOKEN_USER,		"user" },
		{ VRTOKEN_PROP,		"prop" },
		{ VRTOKEN_INPUTDEV,	"inputDevice" },
		{ VRTOKEN_INPUTDEV,	"indev" },
		{ VRTOKEN_INPUTMAP,	"inputMap" },

	   /** object option specifiers **/
		{ VRTOKEN_NAME,		"name" },
		{ VRTOKEN_TYPE,		"type" },			/* for system objects, this is the UI-style */
		{ VRTOKEN_DSOFILE,	"dsoFile" },
		{ VRTOKEN_DSOFUNC,	"dsoFunc" },
		{ VRTOKEN_ARGS,		"args" },
		{ VRTOKEN_MALLEABLE,	"malleable" },
		{ VRTOKEN_MALLEABLE,	"changeable" },

		{ VRTOKEN_COPYOBJECT,	"copyOf" },

	   /** system options **/
		{ VRTOKEN_SYSTEM_CAVECONFIG,	"caveConfig" },
		{ VRTOKEN_SYSTEM_PROCS,		"procs" },
		{ VRTOKEN_SYSTEM_INPUTMAP,	"inputMap" },

		{ VRTOKEN_SYSTEM_MASTER,	"master" },
		{ VRTOKEN_SYSTEM_SLAVES,	"slaves" },

		{ VRTOKEN_WORLD_TRANSFORM,	"World_Transform" },	/* I may deprecate these, not sure they're valuable */
		{ VRTOKEN_WORLD_TRANSLATE,	"World_Translate" },
		{ VRTOKEN_WORLD_ROTATE,		"World_Rotate" },

	   /** process options **/
		{ VRTOKEN_PROCESS_MAIN,		"main" },
		{ VRTOKEN_PROCESS_COMPUTE,	"compute" },
		{ VRTOKEN_INPUT,		"input" },		/* doubles as an inputdevice option and a process type */
		{ VRTOKEN_PROCESS_TELNET,	"telnet" },
		{ VRTOKEN_PROCESS_VISREN,	"visRen" },
		{ VRTOKEN_PROCESS_AUDREN,	"audRen" },
		{ VRTOKEN_PROCESS_HAPREN,	"hapRen" },
		{ VRTOKEN_PROCESS_OLFREN,	"olfRen" },
		{ VRTOKEN_PROCESS_TASREN,	"tasRen" },
		{ VRTOKEN_PROCESS_VESREN,	"vesRen" },

		{ VRTOKEN_PROCESS_MACHINE,	"machine" },
		{ VRTOKEN_PROCESS_SYNC,		"sync" },
		{ VRTOKEN_PROCESS_USEC,		"usecMin" },
		{ VRTOKEN_PROCESS_COLOR,	"printColor" },
		{ VRTOKEN_PROCESS_STRING,	"printString" },
		{ VRTOKEN_PROCESS_DBGFILE,	"printFile" },
		{ VRTOKEN_PROCESS_OBJECT,	"object" },
		{ VRTOKEN_PROCESS_OBJECTS,	"objects" },
		{ VRTOKEN_PROCESS_STATSARGS,	"stats" },

	   /** window options **/
		{ VRTOKEN_WINDOW_MOUNT,		"mount" },
		{ VRTOKEN_WINDOW_MOUNT,		"mountType" },
		{ VRTOKEN_WINDOW_GRAPHICS,	"graphics" },
		{ VRTOKEN_WINDOW_GRAPHICS,	"graphicsType" },
		{ VRTOKEN_WINDOW_EYE_COORDS,	"e2w_coords" },		/* deprecating: changing to a dot-notation */
		{ VRTOKEN_WINDOW_EYE_COORDS,	"e2w.coords" },
		{ VRTOKEN_WINDOW_COORDS,	"rw2w_coords" },	/* deprecating: changing to a dot-notation */
		{ VRTOKEN_WINDOW_COORDS,	"rw2w.coords" },
		{ VRTOKEN_WINDOW_TRANSFORM,	"rw2w_transform" },	/* deprecating: changing to a dot-notation */
		{ VRTOKEN_WINDOW_TRANSFORM,	"rw2w.transform" },
		{ VRTOKEN_WINDOW_TRANSLATE,	"rw2w_translate" },	/* deprecating: changing to a dot-notation */
		{ VRTOKEN_WINDOW_TRANSLATE,	"rw2w.translate" },
		{ VRTOKEN_WINDOW_ROTATE,	"rw2w_rotate" },	/* deprecating: changing to a dot-notation */
		{ VRTOKEN_WINDOW_ROTATE,	"rw2w.rotate" },
		{ VRTOKEN_WINDOW_COORDS,	"coords" },
		{ VRTOKEN_WINDOW_VIEWPORT,	"viewport" },
		{ VRTOKEN_WINDOW_VIEWPORT,	"viewport:both" },
		{ VRTOKEN_WINDOW_VIEWPORT_LEFT,	"viewport:left" },
		{ VRTOKEN_WINDOW_VIEWPORT_RIGHT,"viewport:right" },
		{ VRTOKEN_WINDOW_FVIEWPORT,	"viewportf" },
		{ VRTOKEN_WINDOW_FVIEWPORT,	"viewportf:both" },
		{ VRTOKEN_WINDOW_FVIEWPORT_LEFT,"viewportf:left" },
		{ VRTOKEN_WINDOW_FVIEWPORT_RIGHT,"viewportf:right" },
		{ VRTOKEN_WINDOW_VIEWMASK,	"viewMask" },
		{ VRTOKEN_WINDOW_COLORMASK,	"colorMask" },
		{ VRTOKEN_WINDOW_UISHOW,	"showUI" },
		{ VRTOKEN_WINDOW_UILOC,		"placeUI" },
		{ VRTOKEN_WINDOW_UICOL,		"colorUI" },
		{ VRTOKEN_WINDOW_FPSSHOW,	"showFPS" },
		{ VRTOKEN_WINDOW_FPSLOC,	"placeFPS" },
		{ VRTOKEN_WINDOW_FPSCOL,	"colorFPS" },
		{ VRTOKEN_WINDOW_STATSSHOW,	"showStats" },
		{ VRTOKEN_WINDOW_STATSPROCS,	"statsProcs" },
		{ VRTOKEN_WINDOW_FRAMESHOW,	"showFrame" },
		{ VRTOKEN_WINDOW_SIMMASK,	"simMask" },

	   /** eyelist options **/
		{ VRTOKEN_EYELIST_MONOFB,	"monoFB" },		/* or "single" */
		{ VRTOKEN_EYELIST_LEFTFB,	"leftFB" },		/* or "dual_left" */
		{ VRTOKEN_EYELIST_RIGHTFB,	"rightFB" },		/* or "dual_right" */
		{ VRTOKEN_EYELIST_LEFTVP,	"leftVP" },		/* or "view port left" */
		{ VRTOKEN_EYELIST_RIGHTVP,	"rightVP" },		/* or "view port right" */
		{ VRTOKEN_EYELIST_ANAGLFB,	"anaglFB" },		/* or "anaglyph" */

	   /** user options **/
		{ VRTOKEN_USER_IOD,		"iod" },
		{ VRTOKEN_USER_COLOR,		"color" },
		{ VRTOKEN_USER_HEADSENSOR,	"headSensor" },
		{ VRTOKEN_USER_BODYSENSOR,	"bodySensor" },
		{ VRTOKEN_TRAVEL_TRANSFORM,	"travel_transform" },	/* deprecating: changing to a dot-notation */
		{ VRTOKEN_TRAVEL_TRANSFORM,	"travel.transform" },
		{ VRTOKEN_TRAVEL_TRANSLATE,	"travel_translate" },	/* deprecating: changing to a dot-notation */
		{ VRTOKEN_TRAVEL_TRANSLATE,	"travel.translate" },
		{ VRTOKEN_TRAVEL_ROTATE,	"travel_rotate" },	/* deprecating: changing to a dot-notation */
		{ VRTOKEN_TRAVEL_ROTATE,	"travel.rotate" },

	   /** input (and output) device options **/
		{ VRTOKEN_INDEV_FUNCTION,	"function" },
		{ VRTOKEN_INDEV_DOF,		"dof" },
		{ VRTOKEN_CONTROL,		"control" },
		{ VRTOKEN_INDEV_REFFRAME_TRANSFORM,	"t2rw_transform" },
		{ VRTOKEN_INDEV_REFFRAME_TRANSFORM,	"t2rw.transform" },
		{ VRTOKEN_INDEV_REFFRAME_TRANSLATE,	"t2rw_translate" },
		{ VRTOKEN_INDEV_REFFRAME_TRANSLATE,	"t2rw.translate" },
		{ VRTOKEN_INDEV_REFFRAME_ROTATE,	"t2rw_rotate" },
		{ VRTOKEN_INDEV_REFFRAME_ROTATE,	"t2rw.rotate" },
		{ VRTOKEN_INDEV_REFFRAME_TRANSFORM,	"for_transform" },	/* deprecating: changing to a dot-notation */
		{ VRTOKEN_INDEV_REFFRAME_TRANSFORM,	"for.transform" },	/* frame-of-reference -- same as t2rw, though */
		{ VRTOKEN_INDEV_REFFRAME_TRANSLATE,	"for_translate" },
		{ VRTOKEN_INDEV_REFFRAME_TRANSLATE,	"for.translate" },	/*   more meaningful in the case of shifting  */
		{ VRTOKEN_INDEV_REFFRAME_ROTATE,	"for_rotate" },
		{ VRTOKEN_INDEV_REFFRAME_ROTATE,	"for.rotate" },		/*   the coordinate-system for output devices.*/
		{ VRTOKEN_INDEV_RECEIVER_TRANSFORM,	"r2e_transform" },	/* deprecating: changing to a dot-notation */
		{ VRTOKEN_INDEV_RECEIVER_TRANSFORM,	"r2e.transform" },
		{ VRTOKEN_INDEV_RECEIVER_TRANSLATE,	"r2e_translate" },	/* deprecating: changing to a dot-notation */
		{ VRTOKEN_INDEV_RECEIVER_TRANSLATE,	"r2e.translate" },
		{ VRTOKEN_INDEV_RECEIVER_ROTATE,	"r2e_rotate" },		/* deprecating: changing to a dot-notation */
		{ VRTOKEN_INDEV_RECEIVER_ROTATE,	"r2e.rotate" },
		{ VRTOKEN_INDEV_INITPOS_TRANSFORM,	"ipos_transform" },	/* deprecating: changing to a dot-notation */
		{ VRTOKEN_INDEV_INITPOS_TRANSFORM,	"ipos.transform" },	/* The initial position of a 6-sensor */
		{ VRTOKEN_INDEV_INITPOS_TRANSLATE,	"ipos_translate" },	/* deprecating: changing to a dot-notation */
		{ VRTOKEN_INDEV_INITPOS_TRANSLATE,	"ipos.translate" },
		{ VRTOKEN_INDEV_INITPOS_ROTATE,		"ipos_rotate" },	/* deprecating: changing to a dot-notation */
		{ VRTOKEN_INDEV_INITPOS_ROTATE,		"ipos.rotate" },
};


	/**************************************************/
	/*** Configuration function evaluator functions ***/
	/**************************************************/

/*********************************************************************/
/* _ConfigFunctionExecPipe(): is an internal function that forks a   */
/*   process to execute an external function (ala a "shell command"),*/
/*   and returns as a string the output of that command.  The string */
/*   is passed into the configuration parsing for use in dynamic     */
/*   configuration settings.                                         */
/* This might be used in situations where the VR hardware can be     */
/*   reconfigured, and there is a command that perhaps can get the   */
/*   angle of a flexible CAVE and calculate the screen coordinates   */
/*   from that angle.                                                */
/*                                                                   */
/* Presently the configuration function name is:                     */
/*   * execPipe()                                                    */
char *_ConfigFunctionExecPipe(char *args)
{
	FILE	*pipe;
	char	buffer[1024];
	char	*result = NULL;
	char	*tofree;		/* pointer to memory that can be freed */

	/* make a pipe input from the command */
	pipe = popen(args, "r");

	if (pipe == NULL) {
		/* TODO: error message & return the empty string */
	}

	while (fgets(buffer, sizeof(buffer)-1, pipe)) {
		if (result == NULL) {
			result = vrShmemStrDup(buffer);
		} else {
			tofree = result;
			result = vrShmemStrCat(result, buffer);
			vrShmemFree(tofree);
		}
	}

	pclose(pipe);

#if 0
vrPrintf(RED_TEXT "Returning from function 'ExecPipe' with value of '%s'\n" NORM_TEXT, args);
#endif
	return (result);
}


/**********************************************************************/
/* _ConfigFunctionExecStatus(): is an internal function that forks a  */
/*   process to execute an external function (ala a "shell command"), */
/*   and returns as a string the output status of that command.       */
/*   Most "shell commands" will return a status number, where zero    */
/*   (0) is often consider a success (no error) return, and all other */
/*   values represent some form of failure.  The return value is      */
/*   passed into the configuration parsing as a string for use in     */
/*   dynamic configuration settings.                                  */
/* This might be used in situations where the VR hardware can be      */
/*   reconfigured, and there is a command that perhaps determines     */
/*   whether a particular tracking system is active or not, which     */
/*   affects how a CAVE, perhaps, does it's tracking.                 */
/*                                                                    */
/* Presently the configuration function name is:                      */
/*   * execStatus()                                                   */
char *_ConfigFunctionExecStatus(char *args)
{
	char	result[128];

	snprintf(result, sizeof(result), "%d", system(args));

#if 0
vrPrintf(RED_TEXT "Returning from function 'ExecStatus' with value of '%s'\n" NORM_TEXT, args);
#endif
	return (vrShmemStrDup(result));
}


/**************************************************************************/
/* The _ConfigFunctionExistsInputDeviceType() function is an internal     */
/*   function that can be accessed by the configuration process to enable */
/*   a system configuration to be affected by the current state of the    */
/*   library.                                                             */
/* In this case, the function returns the value of "0" or "1" (in string  */
/*   form, indicating whether or not a particular input device type has   */
/*   been compiled into the library -- for example "evio", "vruidd",      */
/*   "vruiddout, "vrpn", etc.                                             */
/* The list comes from the file: "vr_input.opts.h".                       */
/*                                                                        */
/* Presently the configuration function name is:                          */
/*   * existsInputDeviceType()                                            */
char *_ConfigFunctionExistsInputDeviceType(char *args)
{
	if (vrInputDeviceInList(args))
		return (vrShmemStrDup("1"));
	else	return (vrShmemStrDup("0"));
}


/**************************************************************************/
/* The _ConfigFunctionExistsVisrenDeviceType() function is an internal    */
/*   function that can be accessed by the configuration process to enable */
/*   a system configuration to be affected by the current state of the    */
/*   library.                                                             */
/* In this case, the function returns the value of "0" or "1" (in string  */
/*   form, indicating whether or not a particular rendering device has    */
/*   been compiled into the library -- for example "GLX", "SDL" or "TXT". */
/* The list comes from the file: "vr_visren.opts.h".                      */
/*                                                                        */
/* Presently the configuration function name is:                          */
/*   * existsVisrenDeviceType()                                           */
char *_ConfigFunctionExistsVisrenDeviceType(char *args)
{
	if (vrVisrenDeviceInList(args))
		return (vrShmemStrDup("1"));
	else	return (vrShmemStrDup("0"));
}


/***********************************************************************/
/* The _ConfigFunctionNumberProcsOfType() function is an internal      */
/*   function that ... */
/* NOTE: I'm trying to move away from the use of system-wide globals,  */
/*   but for now I'm using the vrContext global to get access to the   */
/*   existing configuration.  To fix this, all these config-functions  */
/*   would need to take the vrContext as an argument, which would need */
/*   to be added in vr_parse.c:_FunctionString().                      */
/*                                                                     */
/* Presently the configuration function name is:                       */
/*   * numberProcsOfType()                                             */
char *_ConfigFunctionNumberProcsOfType(char *args)
{
static	char		*whitespace = " \t\r\b";
	vrContextInfo	*context = vrContext;	/* Ideally this would be passed in as an argument */
	vrSystemInfo	*system;
	vrProcessInfo	*proc;
	vrProcessType	proc_type = VRPROC_NONE;
	char		*parse;			/* pointer to the string being parsed */
	char		parse_word[64];		/* parse one argument from the args string */
	char		result_string[64];	/* Used to convert the result integer to a string */
	int		str_len;
	int		result = 0;
	int		count;

	/* First run the argument string through variable substitution */
	parse = vrSubstituteVariablesInString(context, args);

#if 0 /* delete me */
	vrPrintf("numberProcsOfType: args = '%s'\n", args);
#endif

	/* Parse out the system name from the function arguments */
	parse += strspn(parse, whitespace);     	/* skip whitespace */
#if defined(__APPLE__) || defined(__CYGWIN__)
	/* Apple & Cygwin don't have strchrnul(), so need to use a conditional */
	if (strchr(parse, ',') != NULL) 
		(strchr(parse, ','))[0] = '\0';
#else
	(strchrnul(parse, ','))[0] = '\0';		/* terminate at the comma */
#endif
	str_len = strcspn(parse, whitespace);		/* count non-whitespace */
	if (str_len > sizeof(parse_word)-1)		/* ensure it's not too big */
		str_len = sizeof(parse_word)-1;
	strncpy(parse_word, parse, str_len);
	parse_word[str_len] = '\0';
	system = (vrSystemInfo *)vrObjectSearch(context, VROBJECT_SYSTEM, parse_word);
	if (system == NULL) {
		vrDbgPrintfN(ALWAYS_DBGLVL, "\nnumberProcsOfType(): Warning: system '%s' not defined yet.\n", parse_word);
		return (vrShmemStrDup("0"));
	}

#if 0 /* (delete me) */
	vrPrintf("Found system at %p (%s) -- it has %d procs\n", system, system->name, system->num_procs);
#endif

	/* Parse out the process type from the function arguments */
	parse += str_len + 1;
//vrPrintf("remaining string to parse: '%s' (%d)\n", parse, strlen(parse));
	parse += strspn(parse, whitespace);     	/* skip whitespace */
#if defined(__APPLE__) || defined(__CYGWIN__)
	/* Apple & Cygwin don't have strchrnul(), so need to use a conditional */
	if (strchr(parse, ',') != NULL) 
		(strchr(parse, ','))[0] = '\0';
#else
	(strchrnul(parse, ')'))[0] = '\0';		/* terminate at the close-paren */
#endif
	str_len = strcspn(parse, whitespace);		/* count non-whitespace */
	if (str_len > sizeof(parse_word)-1)		/* ensure it's not too big */
		str_len = sizeof(parse_word)-1;
	strncpy(parse_word, parse, str_len);
	parse_word[str_len] = '\0';
	proc_type = vrProcessTypeValue(parse_word);
//vrPrintf("Parsed process type as '%s'(%d)\n", parse_word, proc_type);

	for (count = 0; count < system->num_procs; count++) {
#if 0 /* I was trying this, but when the configuration is being read, the system->procs[] array hasn't been instantiated yet, so we have to go by the names. */
		proc = system->procs[count];
#else
		proc = (vrProcessInfo *)vrObjectSearch(context, VROBJECT_PROCESS, system->proc_names[count]);
#endif
		if (proc != NULL) {
			if (proc->type == proc_type) {
				result++;
			}
		} else {
			vrDbgPrintfN(ALWAYS_DBGLVL, "numberProcsOfType(): Warning: process '%s' not defined yet.\n", system->proc_names[count]);
		}
	}

	/* Now return the result as a string */
//vrPrintf("result is %d -- proc[0] is '%s', proc[1] is '%s'\n", result, system->proc_names[0], system->proc_names[1]);
	snprintf(result_string, sizeof(result_string), "%d", result);
	return (vrShmemStrDup(result_string));
}


/**************************************************************/
/* _ConfigFuncMap[]: array that maps vrFunctionInfos to their */
/*   corresponding string representations.                    */
/* These are functions that can be called from within the     */
/*   configuration process.                                   */
static vrFunctionInfo _ConfigFuncMap[] = {
		{ _ConfigFunctionExecStatus,	"execstatus" },
		{ _ConfigFunctionExecPipe,	"execpipe" },
		{ _ConfigFunctionExistsInputDeviceType,"existsinputdevicetype" },
		{ _ConfigFunctionExistsVisrenDeviceType,"existsvisrendevicetype" },
		{ _ConfigFunctionNumberProcsOfType,"numberprocsoftype" },
};


/************************************************************/
/* NOTE: we have a choice of whether to handle the built-in */
/*   variable name before or after checking the environment */
/*   variables.  It is nice to know that a variable always  */
/*   means the same thing -- which favors doing the built-in*/
/*   check first, however by checking the existence of the  */
/*   env-var first, we have to opportunity to override the  */
/*   built-in values.  For now we'll go with the former.    */
/* NOTE: this function also can handle special variable     */
/*   operations specified by a character that precedes the  */
/*   actual name of the variable.  Currently only the '?'   */
/*   operator has been implemented, which returns "0" or "1"*/
/*   depending on whether the variable exists or not.       */
char *vrEvaluateVariable(vrContextInfo *context, char *variable_name)
{
static	char		temp_string[1024];
	char		*return_string = NULL;
	char		*return_value;
	int		existence_check = 0;	/* by default, don't do an existence check */

	/**********************************************/
	/* Do some sanity checks on the variable name */
	if (variable_name == NULL) {
		vrDbgPrintfN(ALWAYS_DBGLVL, "vrEvaluateVariable(): Null variable name\n");
	}

	if (variable_name[0] == '\0') {
		vrDbgPrintfN(ALWAYS_DBGLVL, "vrEvaluateVariable(): Empty variable name\n");
	}

	vrDbgPrintfN(VARIABLE_DBGLVL, "vrEvaluateVariable(): evaluating variable '%s'\n", variable_name);

	/******************************************/
	/* Check for a special operator character */
	if (variable_name[0] == '?') {
		/* doing a variable existence check */
		existence_check = 1;
		variable_name++;	/* skip the '?' */

		vrDbgPrintfN(VARIABLE_DBGLVL, "vrEvaluateVariable(): checking for variable existence\n");
	}

	/***************************************************/
	/* First check whether variable is a built-in name */
	if ((!strcmp(variable_name, "machine")) || (!strcmp(variable_name, "hostname"))) {
		/* return the hostname of the machine we're running on */
		gethostname(temp_string, 1024);
		return_string = vrShmemStrDup(temp_string);
	} else

	if (!strcmp(variable_name, "version")) {
		/* return the FreeVR version string */
		/* The version string is a special case, in that we only want   */
		/*   the actual version code -- not the words "FreeVR Version". */
		/*   However, we want the extra words in the string in order to */
		/*   quickly determine against which version an application was */
		/*   compiled.                                                  */

		char	*last_space;
		char	*last_tab;
		char	*last_return;
		char	*last_newline;
		char	*last_backspace;
		char	*begin_version;

		last_space = strrchr(context->version, ' ');
		last_tab = strrchr(context->version, '\t');
		last_return = strrchr(context->version, '\r');
		last_newline = strrchr(context->version, '\n');
		last_backspace = strrchr(context->version, '\b');

		begin_version = context->version;
		if (last_space > begin_version) begin_version = last_space+1;
		if (last_tab > begin_version) begin_version = last_tab+1;
		if (last_return > begin_version) begin_version = last_return+1;
		if (last_newline > begin_version) begin_version = last_newline+1;
		if (last_backspace > begin_version) begin_version = last_backspace+1;

		/* begin the version string with a 'v', so it won't be parsed as a number */
		temp_string[0] = 'v';
		strcpy(&temp_string[1], begin_version);

		return_string = temp_string;
	} else

	if (!strcmp(variable_name, "fullversion")) {
		/* return the FreeVR library architecture (and possibly abi format) */
		return_string = context->version;
	} else

	if (!strcmp(variable_name, "versionnum")) {
		/* return the value of the numeric form of the FreeVR version code */
		snprintf(temp_string, sizeof(temp_string), "%d", FREEVR_VERSION);
		return_string = temp_string;
	} else

	if (!strcmp(variable_name, "name")) {
		/* return the name of the FreeVR application -- as specified by the authors (or not) */
		return_string = context->name;
	} else

	if ((!strcmp(variable_name, "author")) || (!strcmp(variable_name, "authors"))) {
		/* return the author(s) of the FreeVR application -- as specified by the authors (or not) */
		return_string = context->authors;
	} else

	if ((!strcmp(variable_name, "appstatus"))) {
		/* return the programmer specified status of the FreeVR application */
		return_string = context->status_info;
	} else

	if (!strcmp(variable_name, "freevrhomedir")) {
		/* return the home directory of the FreeVR installation */
		return_string = context->homedir;
	} else

	if ((!strcmp(variable_name, "arch")) || (!strcmp(variable_name, "binaryformat"))) {
		/* return the FreeVR library architecture (and possibly abi format) */
		return_string = context->arch;
	} else

	if (!strcmp(variable_name, "system")) {
		/* return the name of the running system */
		return_string = context->config->system_name;
	} else

	if (!strcmp(variable_name, "def_visrenmode")) {
		/* return the name of the default visrenmode setting */
		return_string = vrVisrenModeName(context->config->defaults.visrenmode);
	} else

	if (!strcmp(variable_name, "startup_error")) {
		/* return the value of the system startup error */
		snprintf(temp_string, sizeof(temp_string), "%d", context->startup_error);
		return_string = temp_string;
	} else

	if (!strcmp(variable_name, "memsize")) {
		/* return the size of the shared memory */
		snprintf(temp_string, sizeof(temp_string), "%d", (int)(context->shmem_size));
		return_string = temp_string;
	} else

	if (!strcmp(variable_name, "sysstatus")) {
		/* return the value of the system status */
		snprintf(temp_string, sizeof(temp_string), "%d", context->status);
		return_string = temp_string;
	} else

	if (!strcmp(variable_name, "compile")) {
		/* return the FreeVR library compilation data */
		return_string = context->compile;
	} else

	if (!strcmp(variable_name, "target")) {
		/* return the FreeVR library Makefile target */
		return_string = context->compile_target;
	} else

	if (!strcmp(variable_name, "time_immemorial")) {
		/* return the value of the system time immemorial */
		snprintf(temp_string, sizeof(temp_string), "%.2lf", context->time_immemorial);
		return_string = temp_string;
	} else

	if ((!strcmp(variable_name, "sim_time")) || (!strcmp(variable_name, "time"))) {
		/* return the value of the system simulation time */
		snprintf(temp_string, sizeof(temp_string), "%.2lf", vrCurrentSimTime());
		return_string = temp_string;
	} else

	if (!strcmp(variable_name, "wall_time")) {
		/* return the value of the system wall time */
		snprintf(temp_string, sizeof(temp_string), "%.2lf", vrCurrentWallTime());
		return_string = temp_string;
	} else

	if (!strcmp(variable_name, "debuglevel")) {
		/* return the value of the global debuglevel variable */
		snprintf(temp_string, sizeof(temp_string), "%d", context->config->defaults.debug_level);
		return_string = temp_string;
	} else

	if (!strcmp(variable_name, "debugthistoo")) {
		/* return the value of the global debugthistoo variable */
		snprintf(temp_string, sizeof(temp_string), "%d", context->config->defaults.debug_exact);
		return_string = temp_string;
	} else

	if (!strcmp(variable_name, "win_style")) {
		/* return the name of the windowing interface (Win32 or GLX) */
#if defined(WIN_GLX)
		return_string = vrShmemStrDup("GLX");
#elif defined(WIN_WGL)
		return_string = vrShmemStrDup("Win32-GL");
#else
		return_string = vrShmemStrDup("Win-Unk");
#endif
	} else

	if (!strcmp(variable_name, "gfx_style")) {
		/* return the name of the windowing interface (Win32 or GLX) */
#if defined(GFX_PERFORMER)
		return_string = vrShmemStrDup("Performer");
#elif defined(GFX_OSG)
		return_string = vrShmemStrDup("OSG");
#else
		return_string = vrShmemStrDup("OpenGL");
#endif
	} else

	if (!strcmp(variable_name, "sem_style")) {
		/* return the name of the semaphore method of implementation */
#if defined(SEM_IRIX)
		return_string = vrShmemStrDup("IRIX");
#elif defined(SEM_POSIX)
		return_string = vrShmemStrDup("POSIX");
#elif defined(SEM_SYSVIPC)
		return_string = vrShmemStrDup("SYSVIPC");
#elif defined(SEM_WIN32)
		return_string = vrShmemStrDup("WIN32");
#elif defined(SEM_TCP)
		return_string = vrShmemStrDup("TCP");
#else
		return_string = vrShmemStrDup("Sem-Unk");
#endif
	} else

	if (!strcmp(variable_name, "shm_style")) {
		/* return the name of the shared memory implementation */
#if defined(SHM_SVR4MMAP)
		return_string = vrShmemStrDup("SVR4MMAP");
#elif defined(SHM_BSDANONMMAP)
		return_string = vrShmemStrDup("BSDANONMMAP");
#elif defined(SHM_SYSVIPC)
		return_string = vrShmemStrDup("SYSVIPC");
#elif defined(SHM_PF_ARENA)
		return_string = vrShmemStrDup("PF_ARENA");
#elif defined(SHM_NONE)
		return_string = vrShmemStrDup("NONE");
#else
		return_string = vrShmemStrDup("Shm-Unk");
#endif
	} else

	if (!strcmp(variable_name, "sock_style")) {
		/* return the name of the shared memory implementation */
#if defined(SOCK_W32)
		return_string = vrShmemStrDup("WIN32");
#else
		return_string = vrShmemStrDup("BSD");
#endif
	} else

	if (!strcmp(variable_name, "mp_style")) {
		/* return the name of the multi-processing implementation */
#if defined(MP_PTRHEADS)
		return_string = vrShmemStrDup("PTHREADS");
#elif defined(MP_PTHREADS2)
		return_string = vrShmemStrDup("PTHREADS2");
#else
		return_string = vrShmemStrDup("FORK");
#endif
	} else

	if (!strcmp(variable_name, "compile_options")) {
		/* return the all the compile options used during compilation */
		return_string = vrShmemStrDup("");
#if defined(WIN_GLX)
		return_string = vrShmemStrCat(return_string, " -DWIN_GLX");
#endif
#if defined(WIN_WGL)
		return_string = vrShmemStrCat(return_string, " -DWIN_WGL");
#endif
#if defined(GFX_PERFORMER)
		return_string = vrShmemStrCat(return_string, " -DGFX_PERFORMER");
#endif
#if defined(GFX_OSG)
		return_string = vrShmemStrCat(return_string, " -DGFX_OSG");
#endif
#if defined(SHM_DUMMY)
		return_string = vrShmemStrCat(return_string, " -DSHM_DUMMY");
#endif
#if defined(SHM_NONE)
		return_string = vrShmemStrCat(return_string, " -DSHM_NONE");
#endif
#if defined(SHM_SVR4MMAP)
		return_string = vrShmemStrCat(return_string, " -DSHM_SVR4MMAP");
#endif
#if defined(SHM_BSDANONMMAP)
		return_string = vrShmemStrCat(return_string, " -DSHM_BSDANONMMAP");
#endif
#if defined(SHM_SYSVIPC)
		return_string = vrShmemStrCat(return_string, " -DSHM_SYSVIPC");
#endif
#if defined(SHM_PF_ARENA)
		return_string = vrShmemStrCat(return_string, " -DSHM_PF_ARENA");
#endif
#if defined(SEM_IRIX)
		return_string = vrShmemStrCat(return_string, " -DSEM_IRIX");
#endif
#if defined(SEM_POSIX)
		return_string = vrShmemStrCat(return_string, " -DSEM_POSIX");
#endif
#if defined(SEM_SYSVIPC)
		return_string = vrShmemStrCat(return_string, " -DSEM_SYSVIPC");
#endif
#if defined(SEM_WIN32)
		return_string = vrShmemStrCat(return_string, " -DSEM_WIN32");
#endif
#if defined(SEM_TCP)
		return_string = vrShmemStrCat(return_string, " -DSEM_TCP");
#endif
#if defined(MP_PTHREADS)
		return_string = vrShmemStrCat(return_string, " -DMP_PTHREADS");
#endif
#if defined(MP_PTHREADS2)
		return_string = vrShmemStrCat(return_string, " -DMP_PTHREADS2");
#endif
#if defined(SOCK_W32)
		return_string = vrShmemStrCat(return_string, " -DSOCK_W32");
#endif
	} else

	/* See if the variable name matches any of the debug levels */
	if (vrDbgStringValue(variable_name) >= 0) {
		/* return the value of the debug level name */
		snprintf(temp_string, sizeof(temp_string), "%d", vrDbgStringValue(variable_name));
		return_string = temp_string;
	} else


	/**************************************************************/
	/* If not a built-in variable, then use environment variables */
	if (getenv(variable_name) != NULL) {
		return_string = getenv(variable_name);
		vrDbgPrintfN(VARIABLE_DBGLVL, "vrEvaluateVariable(): using environment variable '%s'\n", variable_name);
	} else

	/****************************************************/
	/* otherwise report that the variable doesn't exist */
	{
		vrDbgPrintfN(DEFAULT_DBGLVL, "vrEvaluateVariable(): No such variable: '%s'\n", variable_name);
	}

	/********************************/
	/* return the appropriate value */

	/* NOTE: I'm not sure we need to put the string into shared memory. */
	/*   These strings are used only once, and will never be deleted.   */
	/*   TODO: perhaps try without shared memory duplication.           */
	if (existence_check) {
		if (return_string == NULL) {
			return_value = vrShmemStrDup("0");
		} else {
			return_value = vrShmemStrDup("1");
		}
	} else {
		if (return_string == NULL) {
			return_value = NULL;
		} else {
			return_value = vrShmemStrDup(return_string);
		}
	}

	vrDbgPrintfN(VARIABLE_DBGLVL, "vrEvaluateVariable(): returning value '%s' for variable '%s'\n", return_value, variable_name);
	return (return_value);
}


/************************************************************************/
/* vrSubstituteVariablesInString(): this function takes a string (along */
/*   with the FreeVR system context), and outputs that string, but with */
/*   all the variables (as denoted by leading dollar signs ('$'))       */
/*   replaced by their values.  The actual variable processing is       */
/*   by vrEvaluateVariable().                                           */
/* Some special cases are:                                              */
/*   "$$" -- outputs a single "$"                                       */
/*   "$<whitespace> -- outputs a single "$"                             */
/*   "$<nul>" (end of string) -- outputs the "$"                        */
char *vrSubstituteVariablesInString(vrContextInfo *context, char *string)
{
static	char		*whitespace = " \t\r\b";
static	char		*strnumchars = "_:abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ.0123456789";	/* characters that can be in a variable name */
	char		*string_copy;		/* pointer to the new string */
	char		*temp_string;		/* used to point to the previous string when allocating more memory */
	int		string_length;		/* length of the incoming string */
	char		*var_begin;		/* where the variable name begins in string */
	int		var_len;		/* length of variable name (less one) */
	char		variable_name[256];	/* copy of the full variable name */
	char		*variable_value;	/* returned value of the variable */
	int		max_size;		/* size of the allocated string */
	int		index = 0;		/* point to copy next character */
	int		count;

	string_length = strlen(string);
	max_size = string_length + 128;

	/* Our first guess for the needed amount of memory for the new      */
	/*   string is the length of the original string, plus a few bytes. */
//vrPrintf("Initial allocation of %d\n", max_size);	/* delete me */
	string_copy = vrShmemAlloc(max_size);

	for (count = 0; count < string_length; count++) {
		if (string[count] == '\0') {
			/* we're done -- but copy the terminal byte */
			string_copy[index] = string[count];
			index++;
			return (string_copy);
		} else if (string[count] != '$') {
			/* not a variable, so just copy the next character */
			/* check to make sure our allocated memory is sufficient */
			if (max_size < index) {
				vrDbgPrintfN(ALWAYS_DBGLVL, "vrSubstituteVariablesInString(): we overran memory for the new string.  This shouldn't happen.\n");
				return (string_copy);
			}

			/* for all other characters (except the '$' which indicates a variable), just copy it */
			string_copy[index] = string[count];
			index++;

		} else {
			/* This means we're at the beginning of a variable */
			var_begin = &string[count+1];	/* variable name begins right after '$' */

			/* maybe -- if the next character is whitespace, or another '$', then just echo it */
			if ((strspn(var_begin, whitespace) > 0) || (var_begin[0] == '\0')) {
				/* if whitespace (or <nul>) follows the '$' then just print the '$' and move on */
				string_copy[index] = string[count];
				index++;
			} else if (var_begin[0] == '$') {
				/* if it's two '$' in a row, copy 1 '$' */
				string_copy[index] = '$';
				count++;	/* to skip one '$' */
				index++;
			} else {
				/* otherwise treat the next character and all alphanums following as variable name */
				/* NOTE: this will pass on non-alphanumerics in the first position */
				/*   for special operations such as the $?<var> construct.         */
				var_len = strspn(var_begin+1, strnumchars);
				if (var_len > sizeof(variable_name)) {
					vrDbgPrintfN(ALWAYS_DBGLVL, "vrSubstituteVariablesInString(): Variable name too long (%*s), ignoring.\n", sizeof(variable_name)-1, var_begin);
					count += var_len;	/* skip the variable name */
				}
				strncpy(variable_name, var_begin, var_len+1);
				variable_name[var_len+1] = '\0';
				count += var_len+1;		/* skip the variable name */

				variable_value = vrEvaluateVariable(context, variable_name);
				if (variable_value == NULL)
					variable_value = "<null>";

				/* Check for the end of allocated memory */
				if (max_size < index + strlen(variable_value) + 1) {
					/* allocate enough more to cover the size of the variable's value */
					max_size += strlen(variable_value) + 2;
					temp_string = string_copy;
//vrPrintf("1: New allocation of %d\n", max_size);	/* delete me */
					string_copy = vrShmemAlloc(max_size);
					strncpy(string_copy, temp_string, index);
				}

				/* Copy the variable result into the string */
//vrPrintf("Copying variable (%s) value -- '%s'\n", variable_name, variable_value);	/* delete me */
//vrPrintf("index points to ... '%c' (%d) '%c'(%d) *%c*(%d)\n", string_copy[index-2], (int)(string_copy[index-2]), string_copy[index-1], (int)(string_copy[index-1]), string_copy[index], (int)(string_copy[index]));	/* delete me */
				strncpy(&(string_copy[index]), variable_value, strlen(variable_value));
				index += strlen(variable_value);
//vrPrintf("String is now (%d): '%s'\n", index, string_copy);	/* delete me */
			}
		}

		/* When we hit the end of the allocated memory, allocate more.*/
		if (max_size == index) {
			max_size *= 2;
			temp_string = string_copy;
//vrPrintf("2: New allocation of %d\n", max_size);	/* delete me */
			string_copy = vrShmemAlloc(max_size);
			strncpy(string_copy, temp_string, index);
		}
	}

	/* return the result -- which is copied into the overall memory pool. */
//vrPrintf("3: returning allocation of %d -- (%d) '%s'\n", max_size, index, string_copy);	/* delete me */
	return (string_copy);
}




	/*************************************************************/
	/*** Algorithmic syntax parsing of the FreeVR config file. ***/
	/*************************************************************/


/*****************/
/*** UseSystem ***/
/********************************************************************/
/*  Format is: 'USESYSTEM' { <string> | 'caveconfig' <filename> } ; */
static void _ConfigParseUsesystem(vrConfigInfo *config, vrParseInfo *parse)
{
	vrTokenInfo	token;

	/** next token is name of system to use **/
	token = vrParseNextToken(parse);

	if (token.token == VRTOKEN_SEMICOLON) {
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "'UseSystem' option prematurely terminated.\n" NORM_TEXT,
			token.linenum);
	} else if (token.token == VRTOKEN_SYSTEM_CAVECONFIG) {
		vrDbgPrintf("TODO: CAVEconfig system option not yet implemented.\n");
		/** next token is name of caveconfig file to read **/
		token = vrParseNextToken(parse);
		if (token.token == VRTOKEN_SEMICOLON) {
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "'UseSystem' option prematurely terminated.\n" NORM_TEXT,
				token.linenum);
		} else if (token.token != VRTOKEN_STRING) {
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "config filename should follow the 'UseSystem caveconfig' option.\n" NORM_TEXT,
				token.linenum);
		}
		return;

	} else if (token.token != VRTOKEN_STRING) {
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "string should follow the 'UseSystem' option.\n" NORM_TEXT,
			token.linenum);
	}

	/* Here's where we actually set the value! */
	//config->system_name = strdup(token.string);	/* test and delete me */
	config->system_name = vrShmemStrDup(token.string);
	vrDbgPrintfN(AALWAYS_DBGLVL, "FreeVR config: " BOLD_TEXT "Using system '%s'\n" NORM_TEXT, config->system_name);

	/** next token should be a statement terminator (ie. ';') **/
	token = vrParseNextToken(parse);
	if (token.token != VRTOKEN_SEMICOLON) {
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "'UseSystem' option not terminated with a ';'.\n" NORM_TEXT,
			token.linenum);
	}
}


/**************/
/*** System ***/
/*****************************************************************************************/
/*  Format is: { system { "=" | "+=" <systemnames>* [;] | inputmap "=" <mapname> [";"] } */

/******************************************************************************/
/* _ConfigParseSystemOption(): receives the first token to parse, which       */
/*   should be one of the system options (ie. "procs" or "inputmap", and      */
/*   continues to read new tokens until hitting a semicolon, or a closecurl.) */
/* Returns the last token read.                                               */
/*                                                                            */
/* For all _ConfigParse<type>Option() functions:                              */
/*   It may also be a general option, including conditionals (if), or the     */
/*   beginning of a "system-option-list" in the language of "freevr.bnf" in   */
/*   other words a block of statements (an open-curly), in which case it      */
/*   will fall through and call _ConfigParseGlobalOption() which will handle  */
/*   the if/curly/etc., and then co-recursively call this function via the    */
/*   "*objparser" function argument.                                          */
/*                                                                            */
/* Arguments:                                                                 */
/*   * vrConfigInfo *config -- the top-level configuration structure          */
/*   * vrParseInfo *parse   -- state of the parser (linenum, tokenmap, etc.)  */
/*   * vrTokenInfo token    -- the parsed token currently being handled       */
/*   * vrObjectInfo *object -- existing "system" into which options are set   */
/*   * _ConfigParseOptionList objparser -- function that parses this obj-type.*/
static vrTokenInfo _ConfigParseSystemOption(vrConfigInfo *config, vrParseInfo *parse, vrTokenInfo token, vrObjectInfo *object, _ConfigParseOptionList objparser)
{
	vrSystemInfo	*system;
	vrObject	*src_object;			/* used for copying an existing object */
	vrTokenInfo	orig_token = token;		/* store the incoming token for debug info */

	vrTrace("_ConfigParseSystemOption()", RED_TEXT "entering" NORM_TEXT);
//vrPrintf("TEMP: top of _ConfigParseSystemOption(): token = %d, line = %d, string = '%s'\n", token.token, token.linenum, token.string); /* delete me */

	system = (vrSystemInfo *)object;

	if (system == NULL) {
		/* NOTE: this should never be true.  If so, then fix where */
		/*   this is called by using a temporary system object.    */
		vrErrPrintf("FreeVR config: Error, Hey, 'system' is NULL no place to put the values\n" NORM_TEXT);
	}

	switch (token.token) {

#if 0 /* delete me */
					/***************/
	case VRTOKEN_SEMICOLON:		/* Format: ";" */
					/***************/
		/* extra semicolons are never a problem */
		token = vrParseNextToken(parse);
		break;
#endif

#if 0 /* delete me */
					/***************/		/* need matching "{" */
	case VRTOKEN_CLOSECURL:		/* Format: "}" */
					/***************/
		/* if we hit the end of the list, then we're done */
		/*   (if this happens here, then it's probably an empty list) */
vrPrintf("TEMP: CLOSE_CURL IN _ConfigParseSystemOption\n");
		vrDbgPrintfN(CONFIG_DBGLVL, "Config Parse (line %d): read '}' in system option -- empty list?\n", token.linenum);
		break;
#endif

					/****************************************************/
	case VRTOKEN_MALLEABLE:		/* Format: "Malleable" assignment-expr number [";"] */
					/****************************************************/
		token = vrParseSingleIntegerExpr(&(system->malleable), "System Malleable", parse);
		break;

					/************************************/
	case VRTOKEN_COPYOBJECT:	/* Format: "copyof" object-name ";" */
					/************************************/
		/*** read the name of an object for copying ***/
		token = vrParseNextToken(parse);
		src_object = vrObjectSearch(config->context, object->object_type, token.string);
		if (src_object == NULL) {
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "Can't find System object \"%s\" for copying.\n" NORM_TEXT,
				token.linenum, token.string);
		} else {
			vrObjectCopy(config->context, object, src_object);
			if (vrDbgDo(CONFIG_DBGLVL)) {
				vrFprintObjectInfo(stdout, object, verbose);
			}
		}
		/* a copyof operation should be terminated by a semicolon. */
		token = vrParseNextToken(parse);
		if (token.token != VRTOKEN_SEMICOLON) {
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "Improper termination to the object copy option.\n" NORM_TEXT,
				token.linenum);
		}
		break;


	/*************************************/
	/* these fields are only for systems */

					/**********************************************************/
	case VRTOKEN_TYPE:		/* Format: "Type" assignment-expr { "sim" | "cave" | ... }*/
					/**********************************************************/
		token = vrParseSingleEnumerator((int *)(&(system->type)), vrUIStyleValue, "System Type", parse);
		break;

					/***************************************************/
	case VRTOKEN_INPUTMAP:		/* Format: "InputMap" assignment-expr string [";"] */
	case VRTOKEN_SYSTEM_INPUTMAP:	/***************************************************/
		token = vrParseSingleStringExpr(&(system->input_map_name), "System InputMap", parse);
		break;

					/*****************************************************/
	case VRTOKEN_PROCESS:		/* Format: "Procs" assignment-expr string-list [";"] */
	case VRTOKEN_SYSTEM_PROCS:	/*****************************************************/
		token = vrParseStringListExpr(&(system->proc_names), &(system->num_procs), "System Procs-list", parse);
		break;

					/*************************************************/
	case VRTOKEN_SYSTEM_MASTER:	/* Format: "Master" assignment-expr string [";"] */
					/*************************************************/
		token = vrParseSingleStringExpr(&(system->master_name), "System Master", parse);
		break;

					/******************************************************/
	case VRTOKEN_SYSTEM_SLAVES:	/* Format: "Slaves" assignment-expr string-list [";"] */
					/******************************************************/
		token = vrParseStringListExpr(&(system->slave_names), &(system->num_slaves), "System Slaves-list", parse);
		break;

					/***********************************************/
	/*****************************************************/
	/* these are fields that can be inherited/overridden */

					/*********************************************/
	case VRTOKEN_SYSTEM_PRECONTEXT:	/* Format: "PreContext" printstyle-value ";" */
					/*********************************************/
		token = vrParseSingleEnumerator((int *)&(config->defaults.pre_context_print), vrPrintStyleValue, "Global PreContextPrint", parse);
		break;

					/********************************************/
	case VRTOKEN_SYSTEM_PRECONFIG:	/* Format: "PreConfig" printstyle-value ";" */
					/********************************************/
		token = vrParseSingleEnumerator((int *)&(config->defaults.pre_config_print), vrPrintStyleValue, "Global PreConfigPrint", parse);
		break;

					/*******************************************/
	case VRTOKEN_SYSTEM_PREINPUT:	/* Format: "PreInput" printstyle-value ";" */
					/*******************************************/
		token = vrParseSingleEnumerator((int *)&(config->defaults.pre_input_print), vrPrintStyleValue, "Global PreInputPrint", parse);
		break;

					/**********************************************/
	case VRTOKEN_SYSTEM_POSTCONTEXT:/* Format: "PostContext" printstyle-value ";" */
					/**********************************************/
		token = vrParseSingleEnumerator((int *)&(config->defaults.post_context_print), vrPrintStyleValue, "Global PostContextPrint", parse);
		break;

					/*********************************************/
	case VRTOKEN_SYSTEM_POSTCONFIG:	/* Format: "PostConfig" printstyle-value ";" */
					/*********************************************/
		token = vrParseSingleEnumerator((int *)&(config->defaults.post_config_print), vrPrintStyleValue, "Global PostConfigPrint", parse);
		break;

					/********************************************/
	case VRTOKEN_SYSTEM_POSTINPUT:	/* Format: "PostInput" printstyle-value ";" */
					/********************************************/
		token = vrParseSingleEnumerator((int *)&(config->defaults.post_input_print), vrPrintStyleValue, "Global PostInputPrint", parse);
		break;

					/****************************************************/
	case VRTOKEN_EXECSTART:		/* Format: "ExecAtStart" assignment-expr string ";" */
					/****************************************************/
		token = vrParseSingleStringExpr(&(system->settings.exec_start), "System ExecAtStart", parse);
		break;

					/***************************************************/
	case VRTOKEN_EXECSTOP:		/* Format: "ExecAtStop" assignment-expr string ";" */
					/***************************************************/
		token = vrParseSingleStringExpr(&(system->settings.exec_stop), "System ExecAtStop", parse);
		break;

					/******************************************************/
	case VRTOKEN_EXECUPONERROR:	/* Format: "ExecUponError" assignment-expr string ";" */
					/******************************************************/
		token = vrParseSingleStringExpr(&(system->settings.exec_uponerror), "System ExecUponError", parse);
		break;

					/****************************************************************************/
	case VRTOKEN_EXITUPONERROR:	/* Format: "ExitUponError" assignment-expr { "yes" | "no" | "default" } ";" */
					/****************************************************************************/
		token = vrParseSingleIntegerExpr(&(system->settings.exit_uponerror), "System ExitUponError", parse);
		break;

					/**********************************************************************/
	case VRTOKEN_SYSTEM_LOCKCPU:	/* Format: "LockCPU" assignment-expr { "yes" | "no" | "default" } ";" */
					/**********************************************************************/
		token = vrParseSingleIntegerExpr(&(system->settings.lock_proc), "System LockCPU", parse);
		break;

					/********************************************************************/
	case VRTOKEN_SYSTEM_VISMODE:	/* Format: "VisrenMode" assignment-expr { "mono" | "dualfb" | ... } */
					/********************************************************************/
		token = vrParseSingleEnumerator((int *)&(system->settings.visrenmode), vrVisrenModeValue, "System VisrenMode", parse);
		break;

					/**************************************************/
	case VRTOKEN_EYELIST:		/* Format: "EyeList" assignment-expr string [";"] */
					/**************************************************/
		token = vrParseSingleStringExpr(&(system->settings.eyelist_name), "System EyeList", parse);
		break;

					/***********************************************/
	case VRTOKEN_NEARCLIP:		/* Format: "NEAR" assignment-expr number [";"] */
					/***********************************************/
		token = vrParseSingleFloatExpr(&(system->settings.near_clip), "System NearClip", parse);
		break;

					/**********************************************/
	case VRTOKEN_FARCLIP:		/* Format: "FAR" assignment-expr number [";"] */
					/**********************************************/
		token = vrParseSingleFloatExpr(&(system->settings.far_clip), "System FarClip", parse);
		break;

					/***********************************************************/
	case VRTOKEN_TRAVEL_TRANSFORM:	/* Format: "travel_transform" assignment-expr num*16 [";"] */
					/***********************************************************/
		token = vrParseMatrixTransform(&system->settings.user_travel, "System Travel Transform", parse);
		break;

					/****************************************************************/
	case VRTOKEN_TRAVEL_TRANSLATE:	/* Format: "travel_translate" assignment-expr num num num [";"] */
					/****************************************************************/
		token = vrParseMatrixTranslate(&system->settings.user_travel, "System Travel Translate", parse);
		break;

					/*****************************************************************/
	case VRTOKEN_TRAVEL_ROTATE:	/* Format: "travel_rotate" assignment-expr num num num num [";"] */
					/*****************************************************************/
		token = vrParseMatrixRotate(&system->settings.user_travel, "System Travel Rotate", parse);
		break;

					/**************************************************************/
	case VRTOKEN_WORLD_TRANSLATE:	/* Format: "world_translate" assignment-expr num num num [";"] */
					/**************************************************************/
		token = vrParseMatrixTranslate(&system->settings.world_xform, "System World Translate", parse);
		break;

					/*************************/
	default:			/* Error: unknown option */
					/*************************/
		/* First test whether the option is a global config command */
//vrPrintf("\n\nTEMP: System unknown option: token = '%s' (%d) @ %d\n", token.string, token.token, token.linenum); /* delete me */


//vrPrintf("\nTEMP: _ConfigParseSystemOption: Unknown option: Here I am!!! token='%s' @%d -- %d =? %d -- Calling _ConfigParseGlobalOption()\n", token.string, token.linenum, token.token, VRTOKEN_UNKNOWN); /* delete me */
		token = _ConfigParseGlobalOption(config, parse, token, object, objparser);
//vrPrintf("TEMP: _ConfigParseSystemOption(): _ConfigParseGlobalOption() returned token ('%s') -- %d\n", token.string, token.token); /* delete me */

		if (token.token == VRTOKEN_UNKNOWN) {
			/* If the unknown token is returned, then print a warning */
			vrDbgPrintfN(CONFIG_WARN_DBGLVL, "Config Parse Warning (line %d): "
				RED_TEXT "Unknown System option -- got '%s'\n" NORM_TEXT,
				token.linenum, token.string);

			/* NOTE: there is no need to call vrParseToEOS() here to read and skip to the next */
			/*   statement, as that is handled within the call to _ConfigParseGlobalOption().  */
		} else {
			vrDbgPrintfN(CONFIG_DBGLVL, "Config Parse (line %d): read global option '%s'\n",
				token.linenum, token.string);
		}
		break;
	}

	if (vrDbgDo(CONFIG_DBGLVL)) {
		vrFprintf(stdout, "_ConfigParseSystemOption(): Just parsed '%s' token on line %d.\n", orig_token.string, orig_token.linenum);
		vrFprintf(stdout, "_ConfigParseSystemOption(): Most immediate token parsed '%s' token on line %d.\n", token.string, token.linenum);
		vrFprintObjectInfo(stdout, (vrObjectInfo *)system, verbose);
	}
	vrTrace("_ConfigParseSystemOption()", RED_TEXT "exiting" NORM_TEXT);
	return token;
}


/***************/
/*** Process ***/
/*************************************************************************************/
/*  Format is: { procs { "=" | "+=" <procname>* [;] | inputmap "=" <mapname> [";"] } */

/******************************************************************************/
/* _ConfigParseProcOption(): receives the first token to parse, which         */
/*   should be one of the Process definition options. (ie. "type" or "lock",  */
/*   and continues to read new tokens until hitting a semicolon, or a         */
/*   closecurl.)                                                              */
/* Returns the last token read.                                               */
/*                                                                            */
/* For all _ConfigParse<type>Option() functions:                              */
/*   It may also be a general option, including conditionals (if), or the     */
/*   beginning of a "process-option-list" in the language of "freevr.bnf" in  */
/*   other words a block of statements (an open-curly), in which case it      */
/*   will fall through and call _ConfigParseGlobalOption() which will handle  */
/*   the if/curly/etc., and then co-recursively call this function via the    */
/*   "*objparser" function argument.                                          */
/*                                                                            */
/* Arguments:                                                                 */
/*   * vrConfigInfo *config -- the top-level configuration structure          */
/*   * vrParseInfo *parse   -- state of the parser (linenum, tokenmap, etc.)  */
/*   * vrTokenInfo token    -- the parsed token currently being handled       */
/*   * vrObjectInfo *object -- existing "system" into which options are set   */
/*   * _ConfigParseOptionList objparser -- function that parses this obj-type.*/
static vrTokenInfo _ConfigParseProcOption(vrConfigInfo *config, vrParseInfo *parse, vrTokenInfo token, vrObjectInfo *object, _ConfigParseOptionList objparser)
{
	vrProcessInfo	*proc;
	vrObject	*src_object;			/* used for copying an existing object */
	vrTokenInfo	orig_token = token;		/* store the incoming token for debug info */
	int		temp_int;

	vrTrace("_ConfigParseProcOption()", RED_TEXT "entering" NORM_TEXT);

	proc = (vrProcessInfo *)object;

	if (proc == NULL) {
		/* NOTE: this should never be true.  If so, then fix where */
		/*   this is called by using a temporary proc object.      */
		vrErrPrintf("FreeVR config: Error, Hey, 'proc' is NULL no place to put the values\n" NORM_TEXT);
	}

	switch (token.token) {

					/****************************************************/
	case VRTOKEN_MALLEABLE:		/* Format: "Malleable" assignment-expr number [";"] */
					/****************************************************/
		token = vrParseSingleIntegerExpr(&(proc->malleable), "Proc Malleable", parse);
		break;

					/************************************/
	case VRTOKEN_COPYOBJECT:	/* Format: "copyof" object-name ";" */
					/************************************/
		/*** read the name of an object for copying ***/
		token = vrParseNextToken(parse);
		src_object = vrObjectSearch(config->context, object->object_type, token.string);
		if (src_object == NULL) {
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "Can't find Process object \"%s\" for copying.\n" NORM_TEXT,
				token.linenum, token.string);
		} else {
			vrObjectCopy(config->context, object, src_object);
			if (vrDbgDo(CONFIG_DBGLVL)) {
				vrFprintObjectInfo(stdout, object, verbose);
			}
		}
		/* a copyof operation should be terminated by a semicolon. */
		token = vrParseNextToken(parse);
		if (token.token != VRTOKEN_SEMICOLON) {
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "Improper termination to the object copy option.\n" NORM_TEXT,
				token.linenum);
		}
		break;


	/********************************************/
	/* the rest of the cases are only for procs */

					/*********************************************/
	case VRTOKEN_TYPE:		/* Format: "Type" assignment-expr enum [";"] */
					/*********************************************/
		token = vrParseNextToken(parse);
		if (token.token != VRTOKEN_SETEQUALS)
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "Expecting an assignment operator for Proc Type.\n" NORM_TEXT,
				token.linenum);
		token = vrParseNextToken(parse);

		switch (token.token) {
		case VRTOKEN_PROCESS_MAIN:
			proc->type = VRPROC_MAIN;
			break;
		case VRTOKEN_PROCESS_COMPUTE:
			proc->type = VRPROC_COMPUTE;
			break;
		case VRTOKEN_PROCESS_TELNET:
			proc->type = VRPROC_TELNET;
			break;
		case VRTOKEN_INPUT:
			proc->type = VRPROC_INPUT;
			break;
		case VRTOKEN_PROCESS_VISREN:
			proc->type = VRPROC_VISREN;
			break;
		case VRTOKEN_PROCESS_AUDREN:
			proc->type = VRPROC_AUDREN;
			break;
		case VRTOKEN_PROCESS_HAPREN:
			proc->type = VRPROC_HAPREN;
			break;
		case VRTOKEN_PROCESS_OLFREN:
			proc->type = VRPROC_OLFREN;
			break;
		case VRTOKEN_PROCESS_TASREN:
			proc->type = VRPROC_TASREN;
			break;
		case VRTOKEN_PROCESS_VESREN:
			proc->type = VRPROC_VESREN;
			break;

		default:
			proc->type = VRPROC_UNKNOWN;
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "Invalid Process type given (%s).\n" NORM_TEXT,
				token.linenum, token.string);
		}
		token = vrParseNextToken(parse);
		if (token.token != VRTOKEN_SEMICOLON && token.token != VRTOKEN_CLOSECURL)
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "Process can only be one type.\n" NORM_TEXT,
				token.linenum);
		break;

					/***********************************************/
	case VRTOKEN_ARGS:		/* Format: "Args" assignment-expr string [";"] */
					/***********************************************/
		token = vrParseSingleStringExpr(&(proc->args), "Proc Args", parse);
		break;

					/************************************************/
	case VRTOKEN_PROCESS_STATSARGS:	/* Format: "Stats" assignment-expr string [";"] */
					/************************************************/
		token = vrParseSingleStringExpr(&(proc->stats_args), "Proc Stats", parse);
		break;

					/***********************************************/
	case VRTOKEN_PROCESS_SYNC:	/* Format: "Sync" assignment-expr number [";"] */
					/***********************************************/
		token = vrParseSingleIntegerExpr(&(proc->sync_id), "Proc Sync", parse);
		break;

					/**************************************************/
	case VRTOKEN_PROCESS_MACHINE:	/* Format: "Machine" assignment-expr string [";"] */
					/**************************************************/
		token = vrParseSingleStringExpr(&(proc->machine_name), "Proc Machine", parse);
		break;

					/*****************************************************/
	case VRTOKEN_DEBUG_LEVEL:	/* Format: "DebugLevel" assignment-expr number [";"] */
					/*****************************************************/
		token = vrParseSingleIntegerExpr(&temp_int, "Proc DebugLevel", parse);
		if (getenv("FREEVR_DEBUGLEVEL_NO") == NULL) {
			proc->settings.debug_level = temp_int;
			vrDbgPrintfN(CONFIG_DBGLVL, "FreeVR config: Setting %s Process DebugLevel to %s (%d).\n",
				proc->name, vrDbgValueString(proc->settings.debug_level), proc->settings.debug_level);
		} else {
			vrDbgPrintfN(CONFIG_WARN_DBGLVL, "FreeVR config: Request to set Process %s DebugLevel to %d denied.\n",
				proc->name, temp_int);
		}
		break;

					/***************************************************/
	case VRTOKEN_DEBUG_EXACT:	/* Format: "DebugExact" assignment-expr number ";" */
					/***************************************************/
		token = vrParseSingleIntegerExpr(&temp_int, "Proc DebugExact", parse);
		if (getenv("FREEVR_DEBUGEXACT_NO") == NULL) {
			proc->settings.debug_exact = temp_int;
			vrDbgPrintfN(CONFIG_DBGLVL, "FreeVR config: Setting %s Process DebugExact to %s (%d).\n",
				proc->name, vrDbgValueString(proc->settings.debug_exact), proc->settings.debug_exact);
		} else {
			vrDbgPrintfN(CONFIG_WARN_DBGLVL, "FreeVR config: Request to set Process %s DebugExact to %d denied.\n",
				proc->name, temp_int);
		}
		break;

					/**************************************************/
	case VRTOKEN_PROCESS_USEC:	/* Format: "UsecMin" assignment-expr number [";"] */
					/**************************************************/
		token = vrParseSingleIntegerExpr(&(proc->usec_min), "Proc UsecMin", parse);
		break;

					/*****************************************************/
	case VRTOKEN_PROCESS_COLOR:	/* Format: "PrintColor" assignment-expr number [";"] */
					/*****************************************************/
		token = vrParseSingleIntegerExpr(&(proc->print_color), "Proc PrintColor", parse);
		break;

					/****************************************************/
	case VRTOKEN_PROCESS_STRING:	/* Format: "PrintString" assignment-expr string ";" */
					/****************************************************/
		token = vrParseSingleStringExpr(&(proc->print_string), "Proc PrintString", parse);
#if 0 /* 05/25/2006: I began implementing this (using code from vr_utils.c:vrShellCmd(), but then noticed it requires an extra string holder space, so decided to hold off a bit */
		/* TODO: since "proc->print_string" will be used in a printf, */
		/*   all instances of '%' need to be changed to '%%'.         */

		/* replace all occurances of '%' w/ "%%" first (from vr_utils.c:vrShellCmd() function) */
		while ((temp_char_ptr = strchr(proc->print_string, '%')) != NULL) {
			strncpy(buffer, proc->print_string, (size_t)(temp_char_ptr - proc->print_string));
			strcat(buffer, temp_char_ptr);
		}
#endif

		break;

					/**************************************************/
	case VRTOKEN_PROCESS_DBGFILE:	/* Format: "PrintFile" assignment-expr string ";" */
					/**************************************************/
		token = vrParseSingleStringExpr(&(proc->print_filename), "Proc PrintFile", parse);
		break;

					/**********************************************************************/
	case VRTOKEN_SYSTEM_LOCKCPU:	/* Format: "LockCPU" assignment-expr { "yes" | "no" | "default" } ";" */
					/**********************************************************************/
		token = vrParseSingleIntegerExpr(&(proc->settings.lock_proc), "Process LockCPU", parse);
		break;


	/** TODO: perhaps have process - dependent commands for locking and unlocking **/


					/****************************************************/
	case VRTOKEN_EXECSTART:		/* Format: "ExecAtStart" assignment-expr string ";" */
					/****************************************************/
		token = vrParseSingleStringExpr(&(proc->settings.exec_start), "Proc ExecAtStart", parse);
		break;

					/***************************************************/
	case VRTOKEN_EXECSTOP:		/* Format: "ExecAtStop" assignment-expr string ";" */
					/***************************************************/
		token = vrParseSingleStringExpr(&(proc->settings.exec_stop), "Proc ExecAtStop", parse);
		break;

					/*******************************************************/
	case VRTOKEN_PROCESS_OBJECT:	/* Format: "Objects" assignment-expr string-list [";"] */
	case VRTOKEN_PROCESS_OBJECTS:	/*******************************************************/
		/* TODO: perhaps we could also do VROTKEN_WINDOWS & VRTOKEN_DEVICES */
		/*   and verify that the token matches the type of process.         */
		token = vrParseStringListExpr(&(proc->thing_names), &(proc->num_thing_names), "Proc Objects-list", parse);
		break;

					/*************************/
	default:			/* Error: unknown option */
					/*************************/
		/* First test whether the option is a global config command */
		token = _ConfigParseGlobalOption(config, parse, token, object, objparser);

		if (token.token == VRTOKEN_UNKNOWN) {
			/* If the unknown token is returned, then print a warning */
			vrDbgPrintfN(CONFIG_WARN_DBGLVL, "Config Parse Warning (line %d): "
				RED_TEXT "Unknown Process option -- got '%s'\n" NORM_TEXT,
				token.linenum, token.string);

			/* NOTE: there is no need to call vrParseToEOS() here to read and skip to the next */
			/*   statement, as that is handled within the call to _ConfigParseGlobalOption().  */

		} else {
			vrDbgPrintfN(CONFIG_DBGLVL, "Config Parse (line %d): read global option '%s'\n",
				token.linenum, token.string);
		}
		break;
	}

	if (vrDbgDo(CONFIG_DBGLVL)) {
		vrFprintf(stdout, "_ConfigParseProcOption(): Just parsed '%s' token on line %d.\n", orig_token.string, orig_token.linenum);
		vrFprintf(stdout, "_ConfigParseprocOption(): Most immediate token parsed '%s' token on line %d.\n", token.string, token.linenum);
		vrFprintObjectInfo(stdout, (vrObjectInfo *)proc, verbose);
	}
	vrTrace("_ConfigParseProcOption()", RED_TEXT "exiting" NORM_TEXT);
	return token;
}


/**************/
/*** Window ***/
/****************************************************************************************/
/*  Format is: { window { "=" | "+=" <windowname>* [;] | inputmap "=" <mapname> [";"] } */

/******************************************************************************/
/* _ConfigParseWindowOption(): receives the first token to parse, which       */
/*   should be one of the Window definition options. (ie. "rw2w_coords" or    */
/*   "showFPS", and continues to read new tokens until hitting a semicolon,   */
/*   or a closecurl.)                                                         */
/* Returns the last token read.                                               */
/*                                                                            */
/* For all _ConfigParse<type>Option() functions:                              */
/*   It may also be a general option, including conditionals (if), or the     */
/*   beginning of a "window-option-list" in the language of "freevr.bnf" in   */
/*   other words a block of statements (an open-curly), in which case it      */
/*   will fall through and call _ConfigParseGlobalOption() which will handle  */
/*   the if/curly/etc., and then co-recursively call this function via the    */
/*   "*objparser" function argument.                                          */
/*                                                                            */
/* Arguments:                                                                 */
/*   * vrConfigInfo *config -- the top-level configuration structure          */
/*   * vrParseInfo *parse   -- state of the parser (linenum, tokenmap, etc.)  */
/*   * vrTokenInfo token    -- the parsed token currently being handled       */
/*   * vrObjectInfo *object -- existing "system" into which options are set   */
/*   * _ConfigParseOptionList objparser -- function that parses this obj-type.*/
static vrTokenInfo _ConfigParseWindowOption(vrConfigInfo *config, vrParseInfo *parse, vrTokenInfo token, vrObjectInfo *object, _ConfigParseOptionList objparser)
{
	vrWindowInfo	*window;
	vrObject	*src_object;			/* used for copying an existing object */
	vrTokenInfo	orig_token = token;		/* store the incoming token for debug info */
	char		*parse_string = NULL;

	vrTrace("_ConfigParseWindowOption()", RED_TEXT "entering" NORM_TEXT);

	window = (vrWindowInfo *)object;
	if (window == NULL) {
		/* NOTE: this should never be true.  If so, then fix where */
		/*   this is called by using a temporary proc object.      */
		vrErrPrintf("FreeVR config: Error, Hey, 'window' is NULL no place to put the values\n" NORM_TEXT);
	}

	switch (token.token) {

					/****************************************************/
	case VRTOKEN_MALLEABLE:		/* Format: "Malleable" assignment-expr number [";"] */
					/****************************************************/
		token = vrParseSingleIntegerExpr(&(window->malleable), "Window Malleable", parse);
		break;

					/************************************/
	case VRTOKEN_COPYOBJECT:	/* Format: "copyof" object-name ";" */
					/************************************/
		/*** read the name of an object for copying ***/
		token = vrParseNextToken(parse);
		src_object = vrObjectSearch(config->context, object->object_type, token.string);
		if (src_object == NULL) {
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "Can't find Window object \"%s\" for copying.\n" NORM_TEXT,
				token.linenum, token.string);
		} else {
			vrObjectCopy(config->context, object, src_object);
			if (vrDbgDo(CONFIG_DBGLVL)) {
				vrFprintObjectInfo(stdout, object, verbose);
			}
		}
		/* a copyof operation should be terminated by a semicolon. */
		token = vrParseNextToken(parse);
		if (token.token != VRTOKEN_SEMICOLON) {
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "Improper termination to the object copy option.\n" NORM_TEXT,
				token.linenum);
		}
		break;

	/**********************************************/
	/* the rest of the cases are only for windows */

					/************************************************/
	case VRTOKEN_WINDOW_MOUNT:	/* Format: "Mount" assignment-expr string [";"] */
					/************************************************/
		token = vrParseSingleStringExpr(&(parse_string), "Window MountType", parse);
		window->mount = vrWindowType(parse_string);
		break;

					/**************************************************/
	case VRTOKEN_EYELIST:		/* Format: "EyeList" assignment-expr string [";"] */
					/**************************************************/
		token = vrParseSingleStringExpr(&(window->settings.eyelist_name), "Window EyeList", parse);
		break;

					/***************************************************/
	case VRTOKEN_WINDOW_GRAPHICS:	/* Format: "Graphics" assignment-expr string [";"] */
					/***************************************************/
		token = vrParseSingleStringExpr(&(window->graphics), "Window GraphicsType", parse);
		break;

					/***********************************************/
	case VRTOKEN_ARGS:		/* Format: "Args" assignment-expr string [";"] */
					/***********************************************/
		token = vrParseSingleStringExpr(&(window->args), "Window Args", parse);
		break;

					/********************************************************************/
	case VRTOKEN_SYSTEM_VISMODE:	/* Format: "VisrenMode" assignment-expr { "mono" | "dualfb" | ... } */
					/********************************************************************/
		token = vrParseSingleEnumerator((int *)(&(window->settings.visrenmode)), vrVisrenModeValue, "Window VisrenMode", parse);
		break;

					/****************************************************/
	case VRTOKEN_WINDOW_EYE_COORDS:	/* Format: "e2w_coords" assignment-expr num*9 [";"] */
					/****************************************************/
		/* TODO: is it reasonable to overload the use of the "coords_ll,lr,ul" fields? */
//vrPrintf("#### Window Eye coordinates for " RED_TEXT "\"%s\":\n" NORM_TEXT, window->name); /* delete me -- NIST floor rotation bug fixed! */
		token = vrParseMatrixFromCoords(&window->e2w_xform, window->coords_ll, window->coords_lr, window->coords_ul, "Window Coordinates", parse);
		break;

					/********************************************************/
	case VRTOKEN_WINDOW_COORDS:	/* Format: "rw2w_coords" assignment-expr num*9 [";"] */
					/********************************************************/
//vrPrintf("#### Window coordinates for " RED_TEXT "\"%s\":\n" NORM_TEXT, window->name); /* delete me -- NIST floor rotation bug fixed! */
		token = vrParseMatrixFromCoords(&window->rw2w_xform, window->coords_ll, window->coords_lr, window->coords_ul, "Window Coordinates", parse);
		break;

					/********************************************************/
	case VRTOKEN_WINDOW_TRANSFORM:	/* Format: "rw2w_transform" assignment-expr num*16 [";"] */
					/********************************************************/
		token = vrParseMatrixTransform(&window->rw2w_xform, "Window Transform", parse);
		/* NOTE: the methods for specifying the rw2w xform other than the coords method don't ultimately */
 		/*   set the coords_{ll,lr,ul} values, which are needed for the perspective calculations!        */
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "rw2w_transform is not yet fully implemented.\n" NORM_TEXT,
			token.linenum);
		vrSleep(5000000);	/* pause to allow user to notice warning */
		break;

					/**************************************************************/
	case VRTOKEN_WINDOW_TRANSLATE:	/* Format: "rw2w_translate" assignment-expr num num num [";"] */
					/**************************************************************/
#ifdef EXPERIMENTAL /* working on this 04/21/21 */
		token = vrParseMatrixTranslateWithCoords(&window->rw2w_xform, window->coords_ll, window->coords_lr, window->coords_ul, "Window Translate", parse);
#else
		token = vrParseMatrixTranslate(&window->rw2w_xform, "Window Translate", parse);
#endif
		/* NOTE: the methods for specifying the rw2w xform other than the coords method don't ultimately */
 		/*   set the coords_{ll,lr,ul} values, which are needed for the perspective calculations!        */
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "rw2w_translate is not yet fully implemented.\n" NORM_TEXT,
			token.linenum);
		vrSleep(500000);	/* pause to allow user to notice warning */
		break;

					/***************************************************************/
	case VRTOKEN_WINDOW_ROTATE:	/* Format: "rw2w_rotate" assignment-expr num num num num [";"] */
					/***************************************************************/
		token = vrParseMatrixRotate(&window->rw2w_xform, "Window Rotate", parse);
		/* NOTE: the methods for specifying the rw2w xform other than the coords method don't ultimately */
 		/*   set the coords_{ll,lr,ul} values, which are needed for the perspective calculations!        */
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "rw2w_rotate is not yet fully implemented.\n" NORM_TEXT,
			token.linenum);
		vrSleep(500000);	/* pause to allow user to notice warning */
		break;

					/***********************************************/
	case VRTOKEN_NEARCLIP:		/* Format: "NEAR" assignment-expr number [";"] */
					/***********************************************/
		token = vrParseSingleFloatExpr(&(window->settings.near_clip), "Window NearClip", parse);
		break;

					/**********************************************/
	case VRTOKEN_FARCLIP:		/* Format: "FAR" assignment-expr number [";"] */
					/**********************************************/
		token = vrParseSingleFloatExpr(&(window->settings.far_clip), "Window FarClip", parse);
		break;

					/**************************************************/
	case VRTOKEN_WINDOW_UISHOW:	/* Format: "showUI" assignment-expr number [";"] */
					/**************************************************/
		token = vrParseSingleIntegerExpr(&(window->ui_show), "Window Show-UI", parse);
		break;

					/**********************************************************/
	case VRTOKEN_WINDOW_UILOC:	/* Format: "placeUI" assignment-expr number number [";"] */
					/**********************************************************/
		token = vrParseFloatListExpr(window->ui_loc, 2, "Window Place-UI", parse);
		break;

					/*****************************************************************/
	case VRTOKEN_WINDOW_UICOL:	/* Format: "colorUI" assignment-expr number number number [";"] */
					/*****************************************************************/
		token = vrParseFloatListExpr(window->ui_color, 3, "Window Color-UI", parse);
		break;

					/**************************************************/
	case VRTOKEN_WINDOW_FPSSHOW:	/* Format: "showFps" assignment-expr number [";"] */
					/**************************************************/
		token = vrParseSingleIntegerExpr(&(window->fps_show), "Window Show-FPS", parse);
		break;

					/**********************************************************/
	case VRTOKEN_WINDOW_FPSLOC:	/* Format: "placeFps" assignment-expr number number [";"] */
					/**********************************************************/
		token = vrParseFloatListExpr(window->fps_loc, 2, "Window Place-FPS", parse);
		break;

					/*****************************************************************/
	case VRTOKEN_WINDOW_FPSCOL:	/* Format: "colorFps" assignment-expr number number number [";"] */
					/*****************************************************************/
		token = vrParseFloatListExpr(window->fps_color, 3, "Window Color-FPS", parse);
		break;

					/****************************************************/
	case VRTOKEN_WINDOW_STATSSHOW:	/* Format: "showStats" assignment-expr number [";"] */
					/****************************************************/
		token = vrParseSingleIntegerExpr(&(window->stats_show), "Window Show-Stats", parse);
		break;

					/*****************************************************/
	case VRTOKEN_WINDOW_STATSPROCS:	/* Format: "statsProcs" assignment-expr string [";"] */
					/*****************************************************/
		token = vrParseSingleStringExpr(&(window->stats_procs), "Window statsProcs", parse);
		break;

					/****************************************************/
	case VRTOKEN_WINDOW_FRAMESHOW:	/* Format: "showFrame" assignment-expr number [";"] */
					/****************************************************/
		token = vrParseSingleIntegerExpr(&(window->show_in_simulator), "Window Show-Frame", parse);
		break;

					/**************************************************/
	case VRTOKEN_WINDOW_SIMMASK:	/* Format: "simMask" assignment-expr number [";"] */
					/**************************************************/
		token = vrParseSingleIntegerExpr(&(window->simulator_mask), "Window Simulator-Mask", parse);
		break;

					/************************************************************************/
	case VRTOKEN_WINDOW_VIEWPORT:	/* Format: "viewport[:both]" assignment-expr number number number number [";"] */
					/************************************************************************/
		token = vrParseIntegerListExpr((int *)&window->viewport_left, 4, "Window Viewport", parse);
		window->viewport_right =  window->viewport_left;
		break;

					   /***********************************************************************/
	case VRTOKEN_WINDOW_VIEWPORT_LEFT: /* Format: "viewport:left" assignment-expr number number number number [";"] */
					   /***********************************************************************/
		token = vrParseIntegerListExpr((int *)&window->viewport_left, 4, "Window Viewport:Left", parse);
		break;

					    /******************************************************************************/
	case VRTOKEN_WINDOW_VIEWPORT_RIGHT: /* Format: "viewport:right" assignment-expr number number number number [";"] */
					    /******************************************************************************/
		token = vrParseIntegerListExpr((int *)&window->viewport_right, 4, "Window Viewport:Right", parse);
		break;

					/************************************************************************/
	case VRTOKEN_WINDOW_FVIEWPORT:	/* Format: "viewportf[:both]" assignment-expr number number number number [";"] */
					/************************************************************************/
		token = vrParseFloatListExpr((float *)&window->viewportF_left, 4, "Window ViewportF", parse);
		window->viewportF_right =  window->viewportF_left;
		break;

					   /***********************************************************************/
	case VRTOKEN_WINDOW_FVIEWPORT_LEFT: /* Format: "viewportf:left" assignment-expr number number number number [";"] */
					   /***********************************************************************/
		token = vrParseFloatListExpr((float *)&window->viewportF_left, 4, "Window ViewportF:Left", parse);
		break;

					    /******************************************************************************/
	case VRTOKEN_WINDOW_FVIEWPORT_RIGHT: /* Format: "viewportf:right" assignment-expr number number number number [";"] */
					    /******************************************************************************/
		token = vrParseFloatListExpr((float *)&window->viewportF_right, 4, "Window ViewportF:Right", parse);
		break;


					/******************************/
					/* Error: Not Yet Implemented */
					/******************************/
	case VRTOKEN_WINDOW_VIEWMASK:
	case VRTOKEN_WINDOW_COLORMASK:
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "Config Parse Warning (line %d): "
			RED_TEXT "Unimplemented Window option -- got '%s'\n" NORM_TEXT,
			token.linenum, token.string);

		/* read and skip the rest of the statement */
		token = vrParseToEOS(parse);
		vrDbgPrintfN(CONFIG_DBGLVL, "Config Parse (line %d): skipping '%s'\n",
			token.linenum, token.string);
		break;


					/*************************/
	default:			/* Error: unknown option */
					/*************************/
		/* First test whether the option is a global config command */
		token = _ConfigParseGlobalOption(config, parse, token, object, objparser);

		if (token.token == VRTOKEN_UNKNOWN) {
			/* If the unknown token is returned, then print a warning */
			vrDbgPrintfN(CONFIG_WARN_DBGLVL, "Config Parse Warning (line %d): "
				RED_TEXT "Unknown Window option -- got '%s'\n" NORM_TEXT,
				token.linenum, token.string);

			/* NOTE: there is no need to call vrParseToEOS() here to read and skip to the next */
			/*   statement, as that is handled within the call to _ConfigParseGlobalOption().  */

		} else {
			vrDbgPrintfN(CONFIG_DBGLVL, "Config Parse (line %d): read global option '%s'\n",
				token.linenum, token.string);
		}
		break;
	}

	if (vrDbgDo(CONFIG_DBGLVL)) {
		vrFprintf(stdout, "_ConfigParseWindowOption(): Just parsed '%s' token on line %d.\n", orig_token.string, orig_token.linenum);
		vrFprintf(stdout, "_ConfigParseWindowOption(): Most immediate token parsed '%s' token on line %d.\n", token.string, token.linenum);
		vrFprintObjectInfo(stdout, (vrObjectInfo *)window, verbose);
	}
	vrTrace("_ConfigParseWindowOption()", RED_TEXT "exiting" NORM_TEXT);
	return token;
}


/***************/
/*** Eyelist ***/
/******************************************************************************************/
/*  Format is: { eyelist { "=" | "+=" <eyelistname>* [;] | inputmap "=" <mapname> [";"] } */

/******************************************************************************/
/* _ConfigParseEyelistOption(): receives the first token to parse, which      */
/*   should be one of the Eyelist definition options. (ie. "monofb" or        */
/*  "leftfb" and continues to read new tokens until hitting a semicolon, or a */
/*  closecurl.)                                                               */
/* Returns the last token read.                                               */
/*                                                                            */
/* For all _ConfigParse<type>Option() functions:                              */
/*   It may also be a general option, including conditionals (if), or the     */
/*   beginning of a "eyelist-option-list" in the language of "freevr.bnf" in  */
/*   other words a block of statements (an open-curly), in which case it      */
/*   will fall through and call _ConfigParseGlobalOption() which will handle  */
/*   the if/curly/etc., and then co-recursively call this function via the    */
/*   "*objparser" function argument.                                          */
/*                                                                            */
/* Arguments:                                                                 */
/*   * vrConfigInfo *config -- the top-level configuration structure          */
/*   * vrParseInfo *parse   -- state of the parser (linenum, tokenmap, etc.)  */
/*   * vrTokenInfo token    -- the parsed token currently being handled       */
/*   * vrObjectInfo *object -- existing "system" into which options are set   */
/*   * _ConfigParseOptionList objparser -- function that parses this obj-type.*/
static vrTokenInfo _ConfigParseEyelistOption(vrConfigInfo *config, vrParseInfo *parse, vrTokenInfo token, vrObjectInfo *object, _ConfigParseOptionList objparser)
{
	vrEyelistInfo	*eyelist;
	vrObject	*src_object;			/* used for copying an existing object */
	vrTokenInfo	orig_token = token;		/* store the incoming token for debug info */

	vrTrace("_ConfigParseEyelistOption()", RED_TEXT "entering" NORM_TEXT);

	eyelist = (vrEyelistInfo *)object;
	if (eyelist == NULL) {
		/* NOTE: this should never be true.  If so, then fix where */
		/*   this is called by using a temporary proc object.      */
		vrErrPrintf("FreeVR config: Error, Hey, 'eyelist' is NULL no place to put the values\n" NORM_TEXT);
	}

	switch (token.token) {

					/****************************************************/
	case VRTOKEN_MALLEABLE:		/* Format: "Malleable" assignment-expr number [";"] */
					/****************************************************/
		token = vrParseSingleIntegerExpr(&(eyelist->malleable), "Eyelist Malleable", parse);
		break;

					/************************************/
	case VRTOKEN_COPYOBJECT:	/* Format: "copyof" object-name ";" */
					/************************************/
		/*** read the name of an object for copying ***/
		token = vrParseNextToken(parse);
		src_object = vrObjectSearch(config->context, object->object_type, token.string);
		if (src_object == NULL) {
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "Can't find Eyelist object \"%s\" for copying.\n" NORM_TEXT,
				token.linenum, token.string);
		} else {
			vrObjectCopy(config->context, object, src_object);
			if (vrDbgDo(CONFIG_DBGLVL)) {
				vrFprintObjectInfo(stdout, object, verbose);
			}
		}
		/* a copyof operation should be terminated by a semicolon. */
		token = vrParseNextToken(parse);
		if (token.token != VRTOKEN_SEMICOLON) {
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "Improper termination to the object copy option.\n" NORM_TEXT,
				token.linenum);
		}
		break;


	/***********************************************/
	/* the rest of the cases are only for eyelists */

					/******************************************************/
	case VRTOKEN_EYELIST_MONOFB:	/* Format: "MonoFB" assignment-expr string-list [";"] */
					/******************************************************/
		token = vrParseStringListExpr(&(eyelist->monofb_names), &(eyelist->num_monofb_names), "MonoFB list", parse);
		break;

					/******************************************************/
	case VRTOKEN_EYELIST_LEFTFB:	/* Format: "LeftFB" assignment-expr string-list [";"] */
					/******************************************************/
		token = vrParseStringListExpr(&(eyelist->leftfb_names), &(eyelist->num_leftfb_names), "LeftFB list", parse);
		break;

					/*******************************************************/
	case VRTOKEN_EYELIST_RIGHTFB:	/* Format: "RightFB" assignment-expr string-list [";"] */
					/*******************************************************/
		token = vrParseStringListExpr(&(eyelist->rightfb_names), &(eyelist->num_rightfb_names), "RightFB list", parse);
		break;

					/*****************************************************/
	case VRTOKEN_EYELIST_LEFTVP:	/* Format: "LeftVP" assignment-expr string-list [";"] */
					/*****************************************************/
		token = vrParseStringListExpr(&(eyelist->leftvp_names), &(eyelist->num_leftvp_names), "LeftVP list", parse);
		break;

					/********************************************************/
	case VRTOKEN_EYELIST_RIGHTVP:	/* Format: "RightVP" assignment-expr string-list [";"] */
					/********************************************************/
		token = vrParseStringListExpr(&(eyelist->rightvp_names), &(eyelist->num_rightvp_names), "RightVP list", parse);
		break;

					/*******************************************************/
	case VRTOKEN_EYELIST_ANAGLFB:	/* Format: "AnaglFB" assignment-expr string-list [";"] */
					/*******************************************************/
		token = vrParseStringListExpr(&(eyelist->anaglfb_names), &(eyelist->num_anaglfb_names), "AnaglFB list", parse);
		break;


					/*************************/
	default:			/* Error: unknown option */
					/*************************/
		/* First test whether the option is a global config command */
		token = _ConfigParseGlobalOption(config, parse, token, object, objparser);

		if (token.token == VRTOKEN_UNKNOWN) {
			/* If the unknown token is returned, then print a warning */
			vrDbgPrintfN(CONFIG_WARN_DBGLVL, "Config Parse Warning (line %d): "
				RED_TEXT "Unknown Eyelist option -- got '%s'\n" NORM_TEXT,
				token.linenum, token.string);

			/* NOTE: there is no need to call vrParseToEOS() here to read and skip to the next */
			/*   statement, as that is handled within the call to _ConfigParseGlobalOption().  */

		} else {
			vrDbgPrintfN(CONFIG_DBGLVL, "Config Parse (line %d): read global option '%s'\n",
				token.linenum, token.string);
		}
		break;
	}

	if (vrDbgDo(CONFIG_DBGLVL)) {
		vrFprintf(stdout, "_ConfigParseEyelistOption(): Just parsed '%s' token on line %d.\n", orig_token.string, orig_token.linenum);
		vrFprintf(stdout, "_ConfigParseEyelistOption(): Most immediate token parsed '%s' token on line %d.\n", token.string, token.linenum);
		vrFprintObjectInfo(stdout, (vrObjectInfo *)eyelist, verbose);
	}
	vrTrace("_ConfigParseEyelistOption()", RED_TEXT "exiting" NORM_TEXT);
	return token;
}


/************/
/*** User ***/
/***************************************************************************************/
/*  Format is: { user { "=" | "+=" <eyelistname>* [;] | inputmap "=" <mapname> [";"] } */

/******************************************************************************/
/* _ConfigParseUserOption(): receives the first token to parse, which         */
/*   should be one of the User definition options. (e.g. "iod" or "color",    */
/*   and continues to read new tokens until hitting a semicolon, or a         */
/*   closecurl.)                                                              */
/* Returns the last token read.                                               */
/*                                                                            */
/* For all _ConfigParse<type>Option() functions:                              */
/*   It may also be a general option, including conditionals (if), or the     */
/*   beginning of a "user-option-list" in the language of "freevr.bnf" in     */
/*   other words a block of statements (an open-curly), in which case it      */
/*   will fall through and call _ConfigParseGlobalOption() which will handle  */
/*   the if/curly/etc., and then co-recursively call this function via the    */
/*   "*objparser" function argument.                                          */
/*                                                                            */
/* Arguments:                                                                 */
/*   * vrConfigInfo *config -- the top-level configuration structure          */
/*   * vrParseInfo *parse   -- state of the parser (linenum, tokenmap, etc.)  */
/*   * vrTokenInfo token    -- the parsed token currently being handled       */
/*   * vrObjectInfo *object -- existing "system" into which options are set   */
/*   * _ConfigParseOptionList objparser -- function that parses this obj-type.*/
static vrTokenInfo _ConfigParseUserOption(vrConfigInfo *config, vrParseInfo *parse, vrTokenInfo token, vrObjectInfo *object, _ConfigParseOptionList objparser)
{
	vrUserInfo	*user;
	vrObject	*src_object;			/* used for copying an existing object */
	vrTokenInfo	orig_token = token;		/* store the incoming token for debug info */
#ifdef UNPICKY_COMPILER
	char		*parse_string = NULL;
#endif

	vrTrace("_ConfigParseUserOption()", RED_TEXT "entering" NORM_TEXT);

	user = (vrUserInfo *)object;
	if (user == NULL) {
		/* NOTE: this should never be true.  If so, then fix where */
		/*   this is called by using a temporary proc object.      */
		vrErrPrintf("FreeVR config: Error, Hey, 'user' is NULL no place to put the values\n" NORM_TEXT);
	}

	switch (token.token) {

					/****************************************************/
	case VRTOKEN_MALLEABLE:		/* Format: "Malleable" assignment-expr number [";"] */
					/****************************************************/
		token = vrParseSingleIntegerExpr(&(user->malleable), "User Malleable", parse);
		break;

					/************************************/
	case VRTOKEN_COPYOBJECT:	/* Format: "copyof" object-name ";" */
					/************************************/
		/*** read the name of an object for copying ***/
		token = vrParseNextToken(parse);
		src_object = vrObjectSearch(config->context, object->object_type, token.string);
		if (src_object == NULL) {
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "Can't find User object \"%s\" for copying.\n" NORM_TEXT,
				token.linenum, token.string);
		} else {
			vrObjectCopy(config->context, object, src_object);
			if (vrDbgDo(CONFIG_DBGLVL)) {
				vrFprintObjectInfo(stdout, object, verbose);
			}
		}
		/* a copyof operation should be terminated by a semicolon. */
		token = vrParseNextToken(parse);
		if (token.token != VRTOKEN_SEMICOLON) {
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "Improper termination to the object copy option.\n" NORM_TEXT,
				token.linenum);
		}
		break;


	/********************************************/
	/* the rest of the cases are only for users */

					/**********************************************/
	case VRTOKEN_USER_IOD:		/* Format: "IOD" assignment-expr number [";"] */
					/**********************************************/
		token = vrParseSingleFloatExpr(&(user->iod), "User IOD", parse);
		break;

					/*****************************************************/
	case VRTOKEN_USER_COLOR:	/* Format: "Color" assignment-expr num num num [";"] */
					/*****************************************************/
		token = vrParseFloatListExpr(user->color, 3, "User Color", parse);
		break;

					/***********************************************/
	case VRTOKEN_NEARCLIP:		/* Format: "NEAR" assignment-expr number [";"] */
					/***********************************************/
		token = vrParseSingleFloatExpr(&(user->settings.near_clip), "User NearClip", parse);
		break;

					/**********************************************/
	case VRTOKEN_FARCLIP:		/* Format: "FAR" assignment-expr number [";"] */
					/**********************************************/
		token = vrParseSingleFloatExpr(&(user->settings.far_clip), "User FarClip", parse);
		break;

					/***********************************************************/
	case VRTOKEN_TRAVEL_TRANSFORM:	/* Format: "travel_transform" assignment-expr num*16 [";"] */
					/***********************************************************/
		token = vrParseMatrixTransform(&user->settings.user_travel, "User Travel Transform", parse);
		break;

					/****************************************************************/
	case VRTOKEN_TRAVEL_TRANSLATE:	/* Format: "travel_translate" assignment-expr num num num [";"] */
					/****************************************************************/
		token = vrParseMatrixTranslate(&user->settings.user_travel, "User Travel Translate", parse);
		break;

					/*****************************************************************/
	case VRTOKEN_TRAVEL_ROTATE:	/* Format: "travel_rotate" assignment-expr num num num num [";"] */
					/*****************************************************************/
		token = vrParseMatrixRotate(&user->settings.user_travel, "User Travel Rotate", parse);
		break;

					/******************************/
					/* Error: Not Yet Implemented */
					/******************************/
	case VRTOKEN_USER_HEADSENSOR:
	case VRTOKEN_USER_BODYSENSOR:
	case VRTOKEN_USER_REFFRAME_TRANSFORM:
	case VRTOKEN_USER_REFFRAME_TRANSLATE:
	case VRTOKEN_USER_REFFRAME_ROTATE:
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "Config Parse Warning (line %d): "
			RED_TEXT "Unimplemented User option -- got '%s'\n" NORM_TEXT,
			token.linenum, token.string);

		/* read and skip the rest of the statement */
		token = vrParseToEOS(parse);
		vrDbgPrintfN(CONFIG_DBGLVL, "Config Parse (line %d): skipping '%s'\n",
			token.linenum, token.string);
		break;


					/*************************/
	default:			/* Error: unknown option */
					/*************************/
		/* First test whether the option is a global config command */
		token = _ConfigParseGlobalOption(config, parse, token, object, objparser);

		if (token.token == VRTOKEN_UNKNOWN) {
			/* If the unknown token is returned, then print a warning */
			vrDbgPrintfN(CONFIG_WARN_DBGLVL, "Config Parse Warning (line %d): "
				RED_TEXT "Unknown User option -- got '%s'\n" NORM_TEXT,
				token.linenum, token.string);

			/* NOTE: there is no need to call vrParseToEOS() here to read and skip to the next */
			/*   statement, as that is handled within the call to _ConfigParseGlobalOption().  */

		} else {
			vrDbgPrintfN(CONFIG_DBGLVL, "Config Parse (line %d): read global option '%s'\n",
				token.linenum, token.string);
		}
		break;
	}

	if (vrDbgDo(CONFIG_DBGLVL)) {
		vrFprintf(stdout, "_ConfigParseUserOption(): Just parsed '%s' token on line %d.\n", orig_token.string, orig_token.linenum);
		vrFprintf(stdout, "_ConfigParseUserOption(): Most immediate token parsed '%s' token on line %d.\n", token.string, token.linenum);
		vrFprintObjectInfo(stdout, (vrObjectInfo *)user, verbose);
	}
	vrTrace("_ConfigParseUserOption()", RED_TEXT "exiting" NORM_TEXT);
	return token;
}


/************/
/*** Prop ***/
/************/
/* TODO: */
static vrTokenInfo _ConfigParsePropOption(vrObjectInfo *object, vrTokenInfo token, vrParseInfo *parse, vrConfigInfo *config, _ConfigParseOptionList objparser)
{
	vrDbgPrintfN(CONFIG_WARN_DBGLVL, "Config Parse Warning (line %d): " RED_TEXT "Sorry, Parsing of Prop options not yet implemented.\n" NORM_TEXT, token.linenum);

	return token;
}


/*******************/
/*** InputDevice ***/
/*********************************************************************************************/
/*  Format is: { inpudevice { "=" | "+=" <eyelistname>* [;] | inputmap "=" <mapname> [";"] } */

/******************************************************************************/
/* _ConfigParseIndevOption(): receives the first token to parse, which        */
/*   should be one of the Input Device definition options. (e.g. "type" or    */
/*   "input", and continues to read new tokens until hitting a semicolon, or  */
/*   a closecurl.)                                                            */
/* Returns the last token read.                                               */
/*                                                                            */
/* For all _ConfigParse<type>Option() functions:                              */
/*   It may also be a general option, including conditionals (if), or the     */
/*   beginning of a "inputdevice-option-list" in the language of "freevr.bnf" */
/*   in other words a block of statements (an open-curly), in which case it   */
/*   will fall through and call _ConfigParseGlobalOption() which will handle  */
/*   the if/curly/etc., and then co-recursively call this function via the    */
/*   "*objparser" function argument.                                          */
/*                                                                            */
/* Arguments:                                                                 */
/*   * vrConfigInfo *config -- the top-level configuration structure          */
/*   * vrParseInfo *parse   -- state of the parser (linenum, tokenmap, etc.)  */
/*   * vrTokenInfo token    -- the parsed token currently being handled       */
/*   * vrObjectInfo *object -- existing "system" into which options are set   */
/*   * _ConfigParseOptionList objparser -- function that parses this obj-type.*/
static vrTokenInfo _ConfigParseIndevOption(vrConfigInfo *config, vrParseInfo *parse, vrTokenInfo token, vrObjectInfo *object, _ConfigParseOptionList objparser)
{
	vrInputDevice	*device;
	vrObject	*src_object;			/* used for copying an existing object */
	vrTokenInfo	orig_token = token;		/* store the incoming token for debug info */

	vrTrace("_ConfigParseIndevOption()", RED_TEXT "entering" NORM_TEXT);

	device = (vrInputDevice *)object;
	if (device == NULL) {
		/* NOTE: this should never be true.  If so, then fix where */
		/*   this is called by using a temporary proc object.      */
		vrErrPrintf("FreeVR config: Error, Hey, 'inputdevice' is NULL no place to put the values\n" NORM_TEXT);
	}

	switch (token.token) {

					/****************************************************/
	case VRTOKEN_MALLEABLE:		/* Format: "Malleable" assignment-expr number [";"] */
					/****************************************************/
		token = vrParseSingleIntegerExpr(&(device->malleable), "InputDevice Malleable", parse);
		break;

					/************************************/
	case VRTOKEN_COPYOBJECT:	/* Format: "copyof" object-name ";" */
					/************************************/
		/*** read the name of an object for copying ***/
		token = vrParseNextToken(parse);
		src_object = vrObjectSearch(config->context, object->object_type, token.string);
		if (src_object == NULL) {
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "Can't find InputDevice object \"%s\" for copying.\n" NORM_TEXT,
				token.linenum, token.string);
		} else {
			vrObjectCopy(config->context, object, src_object);
			if (vrDbgDo(CONFIG_DBGLVL)) {
				vrFprintObjectInfo(stdout, object, verbose);
			}
		}
		/* a copyof operation should be terminated by a semicolon. */
		token = vrParseNextToken(parse);
		if (token.token != VRTOKEN_SEMICOLON) {
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "Improper termination to the object copy option.\n" NORM_TEXT,
				token.linenum);
		}
		break;


	/****************************************************/
	/* the rest of the cases are only for input devices */

					/***********************************************/
	case VRTOKEN_TYPE:		/* Format: "Type" assignment-expr string [";"] */
					/***********************************************/
		token = vrParseSingleStringExpr(&(device->type), "InputDevice Type", parse);
		break;

					/**************************************************/
	case VRTOKEN_DSOFILE:		/* Format: "DsoFile" assignment-expr string [";"] */
					/**************************************************/
		token = vrParseSingleStringExpr(&(device->dso_file), "InputDevice DsoFile", parse);
		break;

					/**************************************************/
	case VRTOKEN_DSOFUNC:		/* Format: "DsoFunc" assignment-expr string [";"] */
					/**************************************************/
		token = vrParseSingleStringExpr(&(device->dso_func), "InputDevice DsoFunc", parse);
		break;

					/***********************************************/
	case VRTOKEN_ARGS:		/* Format: "Args" assignment-expr string [";"] */
					/***********************************************/
		token = vrParseSingleStringExpr(&(device->args), "InputDevice Args", parse);
		break;

					/*******************************************************/
	case VRTOKEN_INPUT:		/* Format: "Input" "name" assignment-expr string [";"] */
					/*******************************************************/
		token = _ConfigParseInputDescription(config, parse, device);
		break;

					/*********************************************************/
	case VRTOKEN_CONTROL:		/* Format: "Control" "name" assignment-expr string [";"] */
					/*********************************************************/
		token = _ConfigParseControlDescription(config, parse, device);
		break;

					/***************************************************************/
	case VRTOKEN_INDEV_REFFRAME_TRANSFORM:/* Format: "t2rw_transform" assignment-expr num*16 [";"] */
					/***************************************************************/
		token = vrParseMatrixTransform(&device->t2rw_xform, "InputDevice Transform", parse);
		break;

					/********************************************************************/
	case VRTOKEN_INDEV_REFFRAME_TRANSLATE:/* Format: "t2rw_translate" assignment-expr num num num [";"] */
					/********************************************************************/
		token = vrParseMatrixTranslate(&device->t2rw_xform, "InputDevice Translate", parse);
		break;

					/***********************************************************************/
	case VRTOKEN_INDEV_REFFRAME_ROTATE:	/* Format: "t2rw_rotate" assignment-expr num num num num [";"] */
					/***********************************************************************/
		token = vrParseMatrixRotate(&device->t2rw_xform, "InputDevice Rotate", parse);
		break;

					/**************************************************************/
	case VRTOKEN_INDEV_RECEIVER_TRANSFORM:/* Format: "r2e_transform" assignment-expr num*16 [";"] */
					/**************************************************************/
		token = vrParseMatrixTransform(&device->r2e_xform, "InputDevice Transform", parse);
		break;

					/*******************************************************************/
	case VRTOKEN_INDEV_RECEIVER_TRANSLATE:/* Format: "r2e_translate" assignment-expr num num num [";"] */
					/*******************************************************************/
		token = vrParseMatrixTranslate(&device->r2e_xform, "InputDevice Translate", parse);
		break;

					/*****************************************************************/
	case VRTOKEN_INDEV_RECEIVER_ROTATE:/* Format: "r2e_rotate" assignment-expr num num num num [";"] */
					/*****************************************************************/
		token = vrParseMatrixRotate(&device->r2e_xform, "InputDevice Rotate", parse);
		break;

					/**************************************************************/
	case VRTOKEN_INDEV_INITPOS_TRANSFORM:/* Format: "ipos_transform" assignment-expr num*16 [";"] */
					/**************************************************************/
		token = vrParseMatrixTransform(&device->ipos_xform, "6-Sensor Initial Position Transform", parse);
		break;

					/*******************************************************************/
	case VRTOKEN_INDEV_INITPOS_TRANSLATE:/* Format: "ipos_translate" assignment-expr num num num [";"] */
					/*******************************************************************/
		token = vrParseMatrixTranslate(&device->ipos_xform, "6-Sensor Iitial Position Translate", parse);
//vrPrintf("setting ipos translation (%p)\n", device->ipos_xform);	// delete me
//vrPrintMatrix("ipos", device->ipos_xform);	// delete me
		break;

					/*****************************************************************/
	case VRTOKEN_INDEV_INITPOS_ROTATE:/* Format: "ipos_rotate" assignment-expr num num num num [";"] */
					/*****************************************************************/
		token = vrParseMatrixRotate(&device->ipos_xform, "6-Sensor Initial Position Rotate", parse);
		break;

					/******************************/
					/* Error: Not Yet Implemented */
					/******************************/
	case VRTOKEN_INDEV_FUNCTION:
	case VRTOKEN_INDEV_DOF:
	case VRTOKEN_EXECSTART:
	case VRTOKEN_EXECSTOP:
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "Config Parse Warning (line %d): "
			RED_TEXT "Unimplemented InputDevice option -- got '%s'\n" NORM_TEXT,
			token.linenum, token.string);

		/* read and skip the rest of the statement */
		token = vrParseToEOS(parse);
		vrDbgPrintfN(CONFIG_DBGLVL, "Config Parse (line %d): skipping '%s'\n",
			token.linenum, token.string);
		break;


					/*************************/
	default:			/* Error: unknown option */
					/*************************/
		/* First test whether the option is a global config command */
		token = _ConfigParseGlobalOption(config, parse, token, object, objparser);

		if (token.token == VRTOKEN_UNKNOWN) {
			/* If the unknown token is returned, then print a warning */
			vrDbgPrintfN(CONFIG_WARN_DBGLVL, "Config Parse Warning (line %d): "
				RED_TEXT "Unknown InputDevice option -- got '%s'\n" NORM_TEXT,
				token.linenum, token.string);

			/* NOTE: there is no need to call vrParseToEOS() here to read and skip to the next */
			/*   statement, as that is handled within the call to _ConfigParseGlobalOption().  */

		} else {
			vrDbgPrintfN(CONFIG_DBGLVL, "Config Parse (line %d): read global option '%s'\n",
				token.linenum, token.string);
		}
		break;
	}

	if (vrDbgDo(CONFIG_DBGLVL)) {
		vrFprintf(stdout, "_ConfigParseIndevOption(): Just parsed '%s' token on line %d.\n", orig_token.string, orig_token.linenum);
		vrFprintf(stdout, "_ConfigParseIndevOption(): Most immediate token parsed '%s' token on line %d.\n", token.string, token.linenum);
		vrFprintObjectInfo(stdout, (vrObjectInfo *)device, verbose);
	}
	vrTrace("_ConfigParseIndevOption()", RED_TEXT "exiting" NORM_TEXT);
	return token;
}


/*************************/
/*** Input description ***/
/**********************************************************************/
/* The "input" object parsing -- happens within "inputdevice" objects */
/*   basically just parse:                                            */
/*     Format: "Input" "name" set-only-assignment string [";"]        */

/******************************************************************************/
/* _ConfigParseInputDescription(): receives the rest of a statement after a   */
/*   VRTOKEN_INPUT ("input") token is encountered.  The next token is simply  */
/*   a string that is the "name" assigned to the new input, followed by a     */
/*   set-only-assignment token, which means it must be just "=", and then     */
/*   another string, this time which describes the input is stored in the     */
/*   input (container) field "desc_str".  This "description string" itself    */
/*   isn't actually parsed until the input is created, which may or may not   */
/*   happen for any given input, depending on whether it's device is active   */
/*   in the operational configuration (via "usesystem").                      */
/* Returns the input "description string" token.                              */
/*                                                                            */
/* Arguments:                                                                 */
/*   * vrConfigInfo *config -- the top-level configuration structure          */
/*   * vrParseInfo *parse   -- state of the parser (linenum, tokenmap, etc.)  */
/*   * vrInputDevice *device -- the input device to which this input belongs  */
static vrTokenInfo _ConfigParseInputDescription(vrConfigInfo *config, vrParseInfo *parse, vrInputDevice *device)
{
	vrTokenInfo	token;
#ifdef UNPICKY_COMPILER
	vrTokenInfo	assignment_token;
#endif
	vrInput		*input;		/* pointer to the input container */

	vrTrace("_ConfigParseInputDescription()", RED_TEXT "entering" NORM_TEXT);

	/* code example taken from _ConfigParseObjectDefinition(), below */


	/** next token is name of object to define/declare **/
	token = vrParseNextToken(parse);
	if (token.token != VRTOKEN_STRING) {
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "string should follow the '%s' option -- got \"%s\".\n" NORM_TEXT,
			token.linenum, "input", token.string);
		return token;
	}


	/** find an existing input object with the matching name in this list of inputs for this device **/
	input = (vrInput *)vrObjectListSearch(VROBJECT_INPUT, (vrObjectInfo *)(device->inputs), token.string);

	/* if one doesn't exist yet, make a new one (this is the usual scenario). */
	if (input == NULL) {
		/* create a new input object definition (adding it to the inputdevice's private list of input objects) */
		input = (vrInput *)vrObjectListNew(config->context, VROBJECT_INPUT, (vrObjectInfo **)(&(device->inputs)), token.string);
		input->line_created = parse->linenum;
		strncpy(input->file_created, parse->parse_name, sizeof(input->file_created));
		vrDbgPrintfN(CONFIG_DBGLVL, "Created new %s: '%s' (%d)\n", "input object", input->name, input->id);
	} else {
		/* if one did exist, report a warning that the old description will be overwritten */
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "Config Parse Warning (line %d): "
			RED_TEXT "Replacing definition of input \"%s::%s\".\n" NORM_TEXT,
			token.linenum, device->name, token.string);
		vrObjectClear((vrObject *)input);
	}

	/* set the last modified line and file for the input-object */
	input->line_lastmod = parse->linenum;
	strncpy(input->file_lastmod, parse->parse_name, sizeof(input->file_lastmod));

#if 1 /* [09/23/21: TODO: determine whether we actually need to store these values in the input container -- I'm not sure we do.] */
	/* store the current r2e transform for inputs that use this */
	input->r2e_xform = vrMatrixCopy(vrMatrixCreate(), device->r2e_xform);
	input->ipos_xform = vrMatrixCopy(vrMatrixCreate(), device->ipos_xform);
#endif

	token = vrParseSingleStringExpr(&(input->desc_str), "Input Description", parse);

	vrTrace("_ConfigParseInputDescription()", RED_TEXT "exiting" NORM_TEXT);

	return token;
}


/***************************/
/*** Control description ***/
/************************************************************************/
/* the "control" object parsing -- happens within "inputdevice" objects */
/*   basically just parse:                                              */
/*     Format: "Control" "name" set-only-assignment string [";"]        */

/******************************************************************************/
/* _ConfigParseControlDescription(): receives the rest of a statement after a */
/*   VRTOKEN_CONTROL ("control") token is encountered.  The next token is     */
/*   simply  a string that is the "name" assigned to the new control, followed*/
/*   by a set-only-assignment token, which means it must be just "=", and then*/
/*   another string, this time which describes the input is stored in the     */
/*   control (container) field "desc_str".  This "description string" itself  */
/*   isn't actually parsed until the control is created, which may or may not */
/*   happen for any given control, depending on whether it's device is active */
/*   in the operational configuration (via "usesystem").                      */
/* Returns the control "description string" token.                            */
/*                                                                            */
/* Arguments:                                                                 */
/*   * vrConfigInfo *config -- the top-level configuration structure          */
/*   * vrParseInfo *parse   -- state of the parser (linenum, tokenmap, etc.)  */
/*   * vrInputDevice *device -- the input device to which this control belongs*/
static vrTokenInfo _ConfigParseControlDescription(vrConfigInfo *config, vrParseInfo *parse, vrInputDevice *device)
{
	vrTokenInfo	token;
#ifdef UNPICKY_COMPILER
	vrTokenInfo	assignment_token;
#endif
	vrInput		*input;

	vrTrace("_ConfigParseControlDescription()", RED_TEXT "entering" NORM_TEXT);

	/* code example taken from _ConfigParseInputDescription(), above */


	/** next token is name of object to define/declare **/
	token = vrParseNextToken(parse);
	if (token.token != VRTOKEN_STRING) {
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "string should follow the '%s' option -- got \"%s\".\n" NORM_TEXT,
			token.linenum, "control", token.string);
		return token;
	}


	/** find an existing control object with the matching name in this list of controls for this device **/
	input = (vrInput *)vrObjectListSearch(VROBJECT_INPUT, (vrObjectInfo *)(device->self_controls), token.string);

	/* if one doesn't exist yet, make a new one. */
	if (input == NULL) {
		/* create a new control object definition (adding it to the inputdevice's private list of input objects) */
		input = (vrInput *)vrObjectListNew(config->context, VROBJECT_INPUT, (vrObjectInfo **)(&(device->self_controls)), token.string);
		input->line_created = parse->linenum;
		strncpy(input->file_created, parse->parse_name, sizeof(input->file_created));
		vrDbgPrintfN(CONFIG_DBGLVL, "Created new %s: '%s' (%d)\n", "control object", input->name, input->id);
	} else {
		/* if one did exist, report a warning that the old description will be overwritten */
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "Config Parse Warning (line %d): "
			RED_TEXT "input device %s already has a control called '%s'."
			"  Will replace with new definition.\n" NORM_TEXT,
			token.linenum, device->name, token.string);
		vrObjectClear((vrObject *)input);
	}

	/* set the last modified line and file for the input-object */
	input->line_lastmod = parse->linenum;
	strncpy(input->file_lastmod, parse->parse_name, sizeof(input->file_lastmod));

	/* store the current r2e transform for inputs that use this   */
	/* (which I guess is unlikely for controls, but what the hey) */
	input->r2e_xform = vrMatrixCopy(vrMatrixCreate(), device->r2e_xform);

	token = vrParseSingleStringExpr(&(input->desc_str), "Control Description", parse);

	vrTrace("_ConfigParseControlDescription()", RED_TEXT "exiting" NORM_TEXT);

	return token;
}



/****************/
/*** InputMap ***/
/****************/
/* TODO: */
static void _ConfigParseInmap(vrConfigInfo *config, vrParseInfo *parse)
{
	vrDbgPrintfN(CONFIG_WARN_DBGLVL, RED_TEXT "Sorry, InputMap parsing not yet implemented.\n" NORM_TEXT);
}


/*********************************/
/*** Generic Object Definition ***/
/**************************************************************************************************************************/
/*     Format: "system|process|window|eyelist|user|prop|inputdevice" object-name [set-assignment option | option-list ";" */

/******************************************************************************/
/* _ConfigParseObjectDefinition(): is called from _ConfigParseGlobalOption()  */
/*   when the name of an object type is encountered as a global keyword.  For */
/*   instance "system" or "window".  The name that follows that keyword is    */
/*   used to find/create the unique object pointer.  Objects may already      */
/*   exist from previous configuration input, and thus be modified or         */
/*   overwritten.  New objects will be created if the name (next token) is    */
/*   new for the given type of object.                                        */
/* Once the object pointer is established, standard object parsing is done,   */
/*   which is to determine which type of assignment should be actioned (either*/
/*   append or overwrite), and then generally look for the open-curly which   */
/*   begins the block that will contain the options for that particular       */
/*   object.                                                                  */
/*                                                                            */
/* Arguments:                                                                 */
/*   * vrConfigInfo *config -- pointer to the top-level configuration struct  */
/*   * vrParseInfo *parse   -- state of the parser (linenum, tokenmap, etc.)  */
/*   * vrObject objtype     -- type of the object to search for               */
/*   * char *objtypename    -- name of the object type (redun: used for debug)*/
/*       (The "objtypename" string is just used for debugging/warning output. */
/*   * _ConfigParseOptionList objparser -- function that parses this obj-type.*/
/*                                                                            */
/* NOTE: I suppose that last argument isn't needed -- there could be a        */
/*   function that uses a switch to jump to the correct function call.  But   */
/*   in fact that "switch" already happened in _ConfigParseGlobalOption(),    */
/*   thus saving the trouble from doing it here.                              */
static vrTokenInfo _ConfigParseObjectDefinition(vrConfigInfo *config, vrParseInfo *parse, vrObject objtype, char *objtypename, _ConfigParseOptionList objparser)
{
static	char		tmp_malleable_object[2048];	/* memory that can be changed in lieu of an object */
	int		opencurl_linenum;
	vrTokenInfo	token;
	vrTokenInfo	assignment_token;
	vrObjectInfo	*object = NULL;
	vrObjectInfo	*src_object = NULL;
	vrObjectInfo	*real_object = NULL;

	vrTrace("_ConfigParseObjectDefinition()", RED_TEXT "entering" NORM_TEXT);
//vrPrintf("TEMP: _ConfigParseObjectDefinition(): object type = '%d', typename = '%s'\n", objtype, objtypename);

	/** next token is name of object to define/declare **/
	token = vrParseNextToken(parse);
	if (token.token != VRTOKEN_STRING) {
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "string should follow the '%s' option -- got \"%s\".\n" NORM_TEXT,
			token.linenum, objtypename, token.string);
		return token;
	}
#ifdef EXTRA_DEBUG
vrPrintf("MAMA: object name is '%s'\n", token.string);
#endif

	/* find an existing object with the matching name */
	object = vrObjectSearch(config->context, objtype, token.string);

	/* if the object doesn't exist yet, make a new one. */
	if (object == NULL) {
		/* create a new object definition */
		object = vrObjectNew(config->context, objtype, token.string);
		object->line_created = parse->linenum;
		strncpy(object->file_created, parse->parse_name, sizeof(object->file_created));
		vrDbgPrintfN(CONFIG_DBGLVL, "Created new %s: '%s' (%d)\n", objtypename, object->name, object->id);
	} else {
		vrDbgPrintfN(CONFIG_DBGLVL, "Found existing %s: '%s' (%d)\n", objtypename, object->name, object->id);
	}

	/* If the found object is not malleable, then make changes in */
	/*   some temporary memory.                                   */
	if (!object->malleable) {
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "Can't change unmalleable %s object \"%s\".\n" NORM_TEXT,
			token.linenum, vrObjectTypeName(object->object_type), object->name);
		real_object = object;
		object = (vrObjectInfo *)tmp_malleable_object;
		object->object_type = real_object->object_type;
		object->name = vrShmemStrDup("-- unmalleable --");
	}

//vrPrintf("TEMP: _ConfigParseObjectDefinition(): next token should be assignment\n");

	/** next token should be one of ("=" | "+=" | ";") **/
	token = vrParseNextToken(parse);

	if (token.token == VRTOKEN_SEMICOLON) {
		/* declaration of a object -- no settings */
		/*   ie. all the work was done above -- no more work to do */
#ifdef EXTRA_DEBUG
vrPrintf("_ConfigParseObjectDefinition(): declaration only\n");
#endif
		return token;
	}
	assignment_token = token;
	if (assignment_token.token == VRTOKEN_SETEQUALS) {
		/* this will set the values from scratch (so clear old ones) */
		vrObjectClear(object);
	}
	if ((assignment_token.token == VRTOKEN_SETEQUALS) || (assignment_token.token == VRTOKEN_PLUSEQUALS)) {
#if 1 /* (Looks like we're now excluding a considerable amount of this code -- it's now all done by calling _ConfigParseStatement()) { */
#  ifdef EXTRA_DEBUG
vrPrintf("_ConfigParseObjectDefinition(): assignment is '%s'\n", ((assignment_token.token == VRTOKEN_SETEQUALS) ? "set" : "append"));
#  endif
		token = _ConfigParseStatement(config, parse, object, objparser);
#else /* } delete me down to the #endif { */
		/* both '=' and '+=' will read in settings and assign them to the object */

		/** next token should be an option-keyword or an "{" **/  /* need this matching "}" */
		token = vrParseNextToken(parse);

		switch (token.token) {

						/************************************/
		case VRTOKEN_COPYOBJECT:	/* Format: "copyof" object-name ";" */
						/************************************/
			/*** read the name of an object for copying ***/
			if (assignment_token.token == VRTOKEN_PLUSEQUALS) {
				vrErrPrintf("FreeVR config: Error (line %d): "
					RED_TEXT "Can't merge objects with '+=' operation.\n" NORM_TEXT,
					token.linenum);
			}

			token = vrParseNextToken(parse);
			src_object = vrObjectSearch(config->context, objtype, token.string);
			if (src_object == NULL) {
				vrErrPrintf("FreeVR config: Error (line %d): "
					RED_TEXT "Can't find object \"%s\" for copying.\n" NORM_TEXT,
					token.linenum, token.string);
			} else {
				vrObjectCopy(config->context, object, src_object);
				if (vrDbgDo(CONFIG_DBGLVL)) {
					vrFprintObjectInfo(stdout, object, verbose);
				}
			}
			/* a copyof operation should be terminated by a semicolon. */
			token = vrParseNextToken(parse);
			if (token.token != VRTOKEN_SEMICOLON) {
				vrErrPrintf("FreeVR config: Error (line %d): "
					RED_TEXT "Improper termination to the object copy option.\n" NORM_TEXT,
					token.linenum);
			}
			break;

#  if 0 /* let global option handle the open-curl */
						/**************************************/
		case VRTOKEN_OPENCURL:		/* Format: "{" { object-option } "}" */
						/**************************************/
			/* NOTE: this *case* may not be necessary, as it may  */
			/*   be handled by the _ConfigParseGlobalOption(),    */
			/*   but I'm going to leave well-enough alone for now.*/

			/*** read a bunch of keyword options ***/
//			token = vrParseNextToken(parse); /* skip past the "{" */  /* old VIM needs matching "}" */
			opencurl_linenum = token.linenum;
			do {
				token = (*objparser)(config, parse, token, object, objparser);
			//} while (token.token != VRTOKEN_CLOSECURL && token.token != VRTOKEN_OPENCURL && token.token != VRTOKEN_EOF);
			} while (token.token != VRTOKEN_CLOSECURL && token.token != VRTOKEN_EOF);
/* TODO: deal with the above line */

			/* if the above loop ended because of an EOF, then the */
			/*   opencurl was unmatched, so print an error message.*/
			if (token.token == VRTOKEN_EOF) {
				vrErrPrintf("FreeVR config: Error (line %d): "
					RED_TEXT "Unmatched '{'\n" NORM_TEXT,   /* old VIM needs matching '}' */
					opencurl_linenum);
			}

			/* if the above loop ended because of an opencurl, then */
			/*   there is a nested section (or missing closecurl),  */
			/*   so printed an error message.                       */
			if (token.token == VRTOKEN_OPENCURL) {
				vrErrPrintf("FreeVR config: Error (line %d): "
					RED_TEXT "Encountered a '{' -- nesting of blocks not allowed\n"
					"\t(still in block started on line %d).\n" NORM_TEXT,
					token.linenum, opencurl_linenum);  /* old VIM needs matching '}' */
			}
			break;
#  endif
						/*****************************/
		default:			/* Format: object-option ";" */
						/*****************************/
			/* NOTE: a list of options in a {} block is handled   */
			/*   above -- though in theory, blocks should be      */
			/*   correctly handled by the overall parsing process.*/

			/*** read a single keyword option ***/
vrPrintf(ANSI_TEXTMAGENTA "NEW: _ConfigParseObjectDefinition() before objparse\n" NORM_TEXT);
			token = (*objparser)(config, parse, token, object, objparser);
vrPrintf(ANSI_TEXTMAGENTA "NEW: _ConfigParseObjectDefinition() after objparse '%s'(%d) @ %d\n" NORM_TEXT, token.string, token.token, token.linenum);

			/* a single keyworld should be terminated by a semicolon. */
			//if (token.token != VRTOKEN_SEMICOLON) {
			if (token.token != VRTOKEN_SEMICOLON && token.token != VRTOKEN_CLOSECURL) { /* TODO: check for "IsTerminal?" */
				vrErrPrintf("FreeVR config: Error (line %d): "
					RED_TEXT "HEY: Improper termination to the '%s' option.\n" NORM_TEXT,
					token.linenum, objtypename);
			}
			/* TODO: do I want to put the close-curl back on the parsing stack? */
		}
#endif /* } (I think most of this section is excluded) */

		/* set the last modified line and file for the object */
		object->line_lastmod = parse->linenum;
		strncpy(object->file_lastmod, parse->parse_name, sizeof(object->file_lastmod));
	}

	vrTrace("_ConfigParseObjectDefinition()", RED_TEXT "exiting" NORM_TEXT);
	return token;
}


/*************************************************/
/*** Default and Global (inheritable) Settings ***/
/************************************************************************/
/*     Format: "set|setDefault" setting-name set-assignment value [";"] */

/*****************************************************************************/
/* _ConfigParseAssignments(): this function parses all the global assignment */
/*   settings, including setting the default values for parameters that can  */
/*   be inherited from the global, down through objects -- for example, the  */
/*   visual-rendering-mode (visrenmode) is inherited by the "visren" process */
/*   and then down to the "eyelist" of that process.  (And at any point, the */
/*   lower-level structure can override the default.)                        */
/*                                                                           */
/* NOTE: "set" and "setDefault" are treated exactly the same.                */
static void _ConfigParseAssignments(vrConfigInfo *config, vrParseInfo *parse)
{
	vrTokenInfo	token;
	int		temp_int;

	/** next token is name of setting to set **/
	token = vrParseNextToken(parse);
//vrPrintf("temp: _ConfigParseAssignments(): got token '%s' (%d) @ %d  //////\n", token.string, token.token, token.linenum);

	switch (token.token) {
					/***************************************************/
	case VRTOKEN_DEBUG_LEVEL:	/* Format: "DebugLevel" assignment-expr number ";" */
					/***************************************************/
		temp_int =  config->defaults.debug_level;	/* set the current value in case there's an error */
		token = vrParseSingleIntegerExpr(&temp_int, "Global DebugLevel", parse);
		if (getenv("FREEVR_DEBUGLEVEL_NO") == NULL) {
			config->defaults.debug_level = temp_int;
			vrDbgPrintfN(CONFIG_DBGLVL, "FreeVR config: Setting Global DebugLevel to %s (%d).\n", vrDbgValueString(config->defaults.debug_level), config->defaults.debug_level);
		} else {
			vrDbgPrintfN(CONFIG_WARN_DBGLVL, "FreeVR config: Request to set Global DebugLevel to %d denied.\n", temp_int);
		}
		break;

					/***************************************************/
	case VRTOKEN_DEBUG_EXACT:	/* Format: "DebugExact" assignment-expr number ";" */
					/***************************************************/
		temp_int =  config->defaults.debug_exact;	/* set the current value in case there's an error */
		token = vrParseSingleIntegerExpr(&temp_int, "Global DebugExact", parse);
		if (getenv("FREEVR_DEBUGEXACT_NO") == NULL) {
			config->defaults.debug_exact = temp_int;
			vrDbgPrintfN(CONFIG_DBGLVL, "FreeVR config: Setting Global DebugExact to %s (%d).\n", vrDbgValueString(config->defaults.debug_exact), config->defaults.debug_exact);
		} else {
			vrDbgPrintfN(CONFIG_WARN_DBGLVL, "FreeVR config: Request to set Global DebugExact to %d denied.\n", temp_int);
		}
		break;

					/*********************************************/
	case VRTOKEN_SYSTEM_PRECONTEXT:	/* Format: "PreContext" printstyle-value ";" */
					/*********************************************/
		token = vrParseSingleEnumerator((int *)&(config->defaults.pre_context_print), vrPrintStyleValue, "Global PreContextPrint", parse);
		break;

					/********************************************/
	case VRTOKEN_SYSTEM_PRECONFIG:	/* Format: "PreConfig" printstyle-value ";" */
					/********************************************/
		token = vrParseSingleEnumerator((int *)&(config->defaults.pre_config_print), vrPrintStyleValue, "Global PreConfigPrint", parse);
		break;

					/*******************************************/
	case VRTOKEN_SYSTEM_PREINPUT:	/* Format: "PreInput" printstyle-value ";" */
					/*******************************************/
		token = vrParseSingleEnumerator((int *)&(config->defaults.pre_input_print), vrPrintStyleValue, "Global PreInputPrint", parse);
		break;

					/**********************************************/
	case VRTOKEN_SYSTEM_POSTCONTEXT:/* Format: "PostContext" printstyle-value ";" */
					/**********************************************/
		token = vrParseSingleEnumerator((int *)&(config->defaults.post_context_print), vrPrintStyleValue, "Global PostContextPrint", parse);
		break;

					/*********************************************/
	case VRTOKEN_SYSTEM_POSTCONFIG:	/* Format: "PostConfig" printstyle-value ";" */
					/*********************************************/
		token = vrParseSingleEnumerator((int *)&(config->defaults.post_config_print), vrPrintStyleValue, "Global PostConfigPrint", parse);
		break;

					/********************************************/
	case VRTOKEN_SYSTEM_POSTINPUT:	/* Format: "PostInput" printstyle-value ";" */
					/********************************************/
		token = vrParseSingleEnumerator((int *)&(config->defaults.post_input_print), vrPrintStyleValue, "Global PostInputPrint", parse);
		break;

					/****************************************************/
	case VRTOKEN_EXECSTART:		/* Format: "ExecAtStart" assignment-expr string ";" */
					/****************************************************/
		token = vrParseSingleStringExpr(&(config->defaults.exec_start), "Global ExecAtStart", parse);
		break;

					/***************************************************/
	case VRTOKEN_EXECSTOP:		/* Format: "ExecAtStop" assignment-expr string ";" */
					/***************************************************/
		token = vrParseSingleStringExpr(&(config->defaults.exec_stop), "Global ExecAtStop", parse);
		break;

					/******************************************************/
	case VRTOKEN_EXECUPONERROR:	/* Format: "ExecUponError" assignment-expr string ";" */
					/******************************************************/
		token = vrParseSingleStringExpr(&(config->defaults.exec_uponerror), "Global ExecUponError", parse);
		break;

					/****************************************************************************/
	case VRTOKEN_EXITUPONERROR:	/* Format: "ExitUponError" assignment-expr { "yes" | "no" | "default" } ";" */
					/****************************************************************************/
		token = vrParseSingleIntegerExpr(&(config->defaults.exit_uponerror), "Global ExitUponError", parse);
		break;

					/*****************************************************************/
	case VRTOKEN_SYSTEM_LOCKCPU:	/* Format: "LockCPU" assignment-expr ("yes" | "no" | number) ";" */
					/*****************************************************************/
		token = vrParseSingleIntegerExpr(&(config->defaults.lock_proc), "Global LockCPU", parse);
		break;

					/****************************************************/
	case VRTOKEN_SYSTEM_LOCKCMD:	/* Format: "LockCommand" assignment-expr string ";" */
					/****************************************************/
		token = vrParseSingleStringExpr(&(config->defaults.proc_lock_cmd), "Global LockCommand", parse);
		break;

					/******************************************************/
	case VRTOKEN_SYSTEM_UNLOCKCMD:	/* Format: "UnlockCommand" assignment-expr string ";" */
					/******************************************************/
		token = vrParseSingleStringExpr(&(config->defaults.proc_unlock_cmd), "Global UnlockCommand", parse);
		break;

					/********************************************************************/
	case VRTOKEN_SYSTEM_VISMODE:	/* Format: "VisrenMode" assignment-expr { "mono" | "dualfb" | ... } */
					/********************************************************************/
		token = vrParseSingleEnumerator((int *)&(config->defaults.visrenmode), vrVisrenModeValue, "Global VisrenMode", parse);
		break;

					/********************************************/
	case VRTOKEN_EYELIST:		/* Format: "EyeList" assignment-expr string */
					/********************************************/
		token = vrParseSingleStringExpr(&(config->defaults.eyelist_name), "Global Eyelist", parse);
		break;

					/***********************************************/
	case VRTOKEN_NEARCLIP:		/* Format: "NEAR" assignment-expr number [";"] */
					/***********************************************/
		token = vrParseSingleFloatExpr(&(config->defaults.near_clip), "Global NearClip", parse);
		break;

					/**********************************************/
	case VRTOKEN_FARCLIP:		/* Format: "FAR" assignment-expr number [";"] */
					/**********************************************/
		token = vrParseSingleFloatExpr(&(config->defaults.far_clip), "Global FarClip", parse);
		break;

					/***********************************************************/
	case VRTOKEN_TRAVEL_TRANSFORM:	/* Format: "travel_transform" assignment-expr num*16 [";"] */
					/***********************************************************/
		token = vrParseMatrixTransform(&config->defaults.user_travel, "Global Travel Transform", parse);
		break;

					/****************************************************************/
	case VRTOKEN_TRAVEL_TRANSLATE:	/* Format: "travel_translate" assignment-expr num num num [";"] */
					/****************************************************************/
		token = vrParseMatrixTranslate(&config->defaults.user_travel, "Global Travel Translate", parse);
		break;

					/*****************************************************************/
	case VRTOKEN_TRAVEL_ROTATE:	/* Format: "travel_rotate" assignment-expr num num num num [";"] */
					/*****************************************************************/
		token = vrParseMatrixRotate(&config->defaults.user_travel, "Global Travel Rotate", parse);
		break;

					/**************************************************************/
	case VRTOKEN_WORLD_TRANSLATE:	/* Format: "world_translate" assignment-expr num num num [";"] */
					/**************************************************************/
		token = vrParseMatrixTranslate(&config->defaults.world_xform, "Global World Translate", parse);
		break;

					/******************************/
					/* Error: Not Yet Implemented */
					/******************************/
	case VRTOKEN_WORLD_TRANSFORM:	/* may become deprecated */
	case VRTOKEN_WORLD_ROTATE:	/* may become deprecated */
	case VRTOKEN_ENDOFTOKENLIST:	/* placeholder for any other NYI token */
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "Config Parse Warning (line %d): "
			RED_TEXT "Unimplemented global setting \"%s\".\n" NORM_TEXT,
			token.linenum, token.string);

		/* read and skip the rest of the statement */
		token = vrParseToEOS(parse);
		vrDbgPrintfN(CONFIG_DBGLVL, "Config Parse (line %d): skipping '%s'\n",
			token.linenum, token.string);
		break;

					/**************************/
	default:			/* Error: unknown setting */
					/**************************/
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "Config Parse Warning (line %d): "
			RED_TEXT "Unknown global setting \"%s\".\n" NORM_TEXT,
			token.linenum, token.string);

		/* read and skip the rest of the statement */
		token = vrParseToEOS(parse);
		vrDbgPrintfN(CONFIG_DBGLVL, "Config Parse (line %d): skipping '%s'\n",
			token.linenum, token.string);
		break;
	}
}


/***********************************/
/*** PrintObject command parsing ***/
/*************************************************************/
/*  Format is: "PRINTOBJECT" <object-type> <object-name> ";" */

/*****************************************************************************/
/* _ConfigParsePrintobject(): this function outputs to the terminal the      */
/*   structure information for the given object (based on the type-name pair */
/*   given as the next two tokens.                                           */
static vrTokenInfo _ConfigParsePrintobject(vrConfigInfo *config, vrParseInfo *parse)
{
	vrTokenInfo	token_type;
	vrTokenInfo	token_name;
	vrTokenInfo	token;
	vrObjectInfo	*object = NULL;
	char		*type_name = "unknown";

	/* TODO: need to check this parsing for format errors */
	/** next token is name of object type to print, followed by the name of the object **/
	token_type = vrParseNextToken(parse);
	token_name = vrParseNextToken(parse);

	switch (token_type.token) {
	case VRTOKEN_SYSTEM:
		type_name = "system";
		object = vrObjectSearch(config->context, VROBJECT_SYSTEM, token_name.string);
		break;

	case VRTOKEN_PROCESS:
		type_name = "process";
		object = vrObjectSearch(config->context, VROBJECT_PROCESS, token_name.string);
		break;

	case VRTOKEN_WINDOW:
		type_name = "window";
		object = vrObjectSearch(config->context, VROBJECT_WINDOW, token_name.string);
		break;

	case VRTOKEN_EYELIST:
		type_name = "eyelist";
		object = vrObjectSearch(config->context, VROBJECT_EYELIST, token_name.string);
		break;

	case VRTOKEN_USER:
		type_name = "user";
		object = vrObjectSearch(config->context, VROBJECT_USER, token_name.string);
		break;

	case VRTOKEN_PROP:
#if 0 /*  I'm temporarily using "prop" to mean the overall system defaults, as I don't have a token for that, and not sure what name to use */
		type_name = "prop";
		object = vrObjectSearch(config->context, VROBJECT_PROP, token_name.string);
#else
		vrFprintSystemSettings(stdout, &config->defaults, verbose);
#endif
		break;

	case VRTOKEN_INPUTDEV:
		type_name = "input device";
		object = vrObjectSearch(config->context, VROBJECT_INDEV, token_name.string);
		break;

	case VRTOKEN_INPUT:
		type_name = "input object";
		object = vrObjectSearch(config->context, VROBJECT_INPUT, token_name.string);
		break;

	default:
		type_name = "unknown";
		object = NULL;
		break;
	}

	if (object == NULL) {
		vrErrPrintf(RED_TEXT "Warning, No %s object '%s'\n", type_name, token_name.string);
	} else if (vrDbgDo(AALWAYS_DBGLVL)) {
		vrFprintf(stdout, "Config (PrintObject): %s '%s' = ", type_name, token_name.string);
		vrFprintObjectInfo(stdout, object, verbose);
	}

	/** next token should be a statement terminator (ie. ';') **/
	token = vrParseNextToken(parse);
	if (token.token != VRTOKEN_SEMICOLON) {
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "'PrintObject' option not terminated with a ';'.\n" NORM_TEXT,
			token.linenum);
	}

	return token;
}


/***********************************/
/*** PrintConfig command parsing ***/
/***************************************************/
/*  Format is: "PRINTCONFIG" <filename-string> ";" */

/****************************************************************************/
/* _ConfigParsePrintconfig(): this function outputs to a file the structure */
/*   information for the overall configuration.                             */
static void _ConfigParsePrintconfig(vrConfigInfo *config, vrParseInfo *parse)
{
	vrTokenInfo	file_name;
	vrTokenInfo	token;

	/* TODO: need to check this parsing for format errors */
	/** next token is name of the file to which to print **/
	file_name = vrParseNextToken(parse);

	/* Make sure a filename was given */
	if (file_name.token != VRTOKEN_STRING) {
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "No filename given for 'PrintConfig' option.\n" NORM_TEXT,
			file_name.linenum);

		return;
	}

	/** next token should be a statement terminator (ie. ';') **/
	token = vrParseNextToken(parse);
	if (token.token != VRTOKEN_SEMICOLON) {
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "'PrintConfig' option not terminated with a ';'.\n" NORM_TEXT,
			token.linenum);
	}

	/******************************/
	/*** now handle the request ***/
	vrDbgPrintf("Config (PrintConfig): will print to file '%s'\n", file_name.string);
	config->config_print = strdup(file_name.string);
}


/********************************/
/*** GdbPause command parsing ***/
/********************************/
/*  Format is: "GDBPAUSE" ";" */

/*****************************************************************************/
/* _ConfigParseGdbpause(): puts the running FreeVR application into "pause"  */
/*   mode (an OS-level call).  This is to allow the system/application to be */
/*   debugged by connecting with a debugging tool such as "gdb".  Before     */
/*   pausing an output message is given that explains how to connect to the  */
/*   process -- assuming the system has "gdb".                               */
static void _ConfigParseGdbpause(vrConfigInfo *config, vrParseInfo *parse)
{
	vrTokenInfo	token;

	/** next token should be a statement terminator (ie. ';') **/
	token = vrParseNextToken(parse);
	if (token.token != VRTOKEN_SEMICOLON) {
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "'GdbPause' option not terminated with a ';'.\n" NORM_TEXT,
			token.linenum);
	}

	/* PAUSE to allow debugger to attach to the process. */
	vrMsgPrintf("Process pausing, use 'gdb -p %d' to debug.\n", getpid());
	pause();
}


/******************************/
/*** Setenv command parsing ***/
/********************************************************/
/*  Format is: "SETENV" <name> ("=" | "+=") <value> ";" */

/****************************************************************************/
/* _ConfigParseSetenv(): parses the following three tokens as "name", "=",  */
/*   and "value" and assigns to the working environment variable named      */
/*   "name" the given "value".  As the FreeVR parsing system recognizes any */
/*   "variable" that isn't internal as possibly coming from the environment,*/
/*   then this acts as though FreeVR were handling variables on it's own.   */
/*                                                                          */
/* NOTE: the "+=" operation means concatenation, not mathematical addition. */
static void _ConfigParseSetenv(vrConfigInfo *config, vrParseInfo *parse)
{
	char		*value_string;
	char		*assignment_string;
	vrTokenInfo	token_varname;
	vrTokenInfo	token_terminator;

	/* get the variable and it's value */
	token_varname = vrParseNextToken(parse);
	value_string = getenv(token_varname.string); /* We get the value in case we are doing an augmentation operation -- e.g. "*=" */

	/* use standard single string setting parser (this captures the assignment operator) */
	/*    NOTE, it's okay if value_string is NULL -- vrParseSingleStringExpr() will allocate */
	token_terminator = vrParseSingleStringExpr(&value_string, "Setenv", parse);

	/* get the assignment operation */
	/*   NOTE: must be done in non-shared memory since it will be stored in the */
	/*   system.  Fortunately though, we haven't forked yet, so all processes   */
	/*   will have these environment variable settings.                         */
	assignment_string = malloc(strlen(token_varname.string) + strlen(value_string) + 2);
	sprintf(assignment_string, "%s=%s", token_varname.string, value_string);

	/** the last token should be a statement terminator (ie. ';') **/
	if (token_terminator.token != VRTOKEN_SEMICOLON) {
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "'Setenv' option not terminated with a ';'.\n" NORM_TEXT,
			token_terminator.linenum);
	}

	/* do the assignment */
	vrDbgPrintfN(CONFIG_DBGLVL FORCE_DBG, "ConfigParse: setenv setting '%s'\n", assignment_string);
	if (putenv(assignment_string)) {
		vrErrPrintf(RED_TEXT "_ConfigParseSetenv: Unable to set the variable\n" NORM_TEXT);
	}
	vrDbgPrintfN(CONFIG_DBGLVL FORCE_DBG, "_ConfigParseSetenv: getenv(%s) now returns '%s'\n",
		token_varname.string, getenv(token_varname.string));
}


/********************************/
/*** UnSetenv command parsing ***/
/*************************************/
/*  Format is: "UNSETENV" <name> ";" */

/*************************************************************************/
/* _ConfigParseUnsetenv(): parses the following tokens as an environment */
/*   variable "name" and removes it from the working environment.  This  */
/*   will likewise make it unavailable for future reference in FreeVR.   */
static void _ConfigParseUnsetenv(vrConfigInfo *config, vrParseInfo *parse)
{
	char		*assignment_string = vrShmemAlloc0(1024);
	vrTokenInfo	token_varname;
	vrTokenInfo	token_terminator;

	/* TODO: need to check this parsing for more format errors */
	token_varname = vrParseNextToken(parse);

	snprintf(assignment_string, 1024, "%s", token_varname.string);
	vrDbgPrintfN(CONFIG_DBGLVL, "ConfigParse: unsetting '%s'\n", assignment_string);
	if (putenv(assignment_string)) {
		vrErrPrintf(RED_TEXT "_ConfigParseSetenv: Unable to unset the variable\n" NORM_TEXT);
	}
	vrDbgPrintfN(CONFIG_DBGLVL, "_ConfigParseSetenv: getenv(%s) now returns '%s'\n",
		token_varname.string,
		(getenv(token_varname.string) ? getenv(token_varname.string) : "<null>"));

	/** next token should be a statement terminator (ie. ';') **/
	token_terminator = vrParseNextToken(parse);
	if (token_terminator.token != VRTOKEN_SEMICOLON) {
		vrErrPrintf("FreeVR config: Error (line %d): "
			RED_TEXT "'Unsetenv' option not terminated with a ';'.\n" NORM_TEXT,
			token_terminator.linenum);
	}
}



/**************************************/
/*** If conditional command parsing ***/
/***************************************************************************/
/*  Format is: "IF" <condition> <statement(s)> ["ELSE" <statement(s)>] ";" */

/********************************************************************/
/* _ConfigParseIf(): this function is called when a VRTOKEN_IF is   */
/*   encountered, and looks next for a <condition> result which is  */
/*   parsed as a math-expression, and if the expression result is   */
/*   non-zero it is considered to be "true" and the next statement, */
/*   or statement block (a collections of statements inside curly   */
/*   braces) is executed.                                           */
/* After the first statement or statement block, the parsing looks  */
/*   ahead to see whether there is a VRTOKEN_ELSE.  If the "else"   */
/*   token is not found, then there is no more for this function to */
/*   handle.  If the "else" token is next, and the result of the    */
/*   math-expression is "false" (zero), then the next statement or  */
/*   statement-block is executed.  Otherwise, if "else" exists and  */
/*   the initial result is "true" then this/these are skipped.      */
static vrTokenInfo _ConfigParseIf(vrConfigInfo *config, vrParseInfo *parse, vrObjectInfo *object, _ConfigParseOptionList objparser)
{
	vrTokenInfo	token;
	vrTokenInfo	statement_token = { VRTOKEN_NOP, "no if statement" };
	vrTokenInfo	skip_token;		/* NOTE: only used for printing debugging info */
	vrTokenInfo	else_token;
	int		math_value;

	/** next token(s) are the math expression **/
	token = vrParseMathExpression(parse);
	math_value = vrAtoI(token.string);
	vrDbgPrintfN(PARSE_DBGLVL, "value of 'if' math-expression is '%s' (%d)\n", token.string, math_value);
	vrDbgPrintfN(VARIABLE_DBGLVL, "value of 'if' math-expression is '%s' (%d)\n", token.string, math_value);
#ifdef EXTRA_DEBUG
vrPrintf(ANSI_TEXTBLUE "value of 'if' math-expression is '%s' (%d)\n" NORM_TEXT, token.string, math_value);
#endif

	/** then read and perhaps execute the statement(s) -- the "then" clause **/
	if (math_value) {
		statement_token = _ConfigParseStatement(config, parse, object, objparser);
		vrDbgPrintfN(CONDITIONAL_DEBUG PARSE_DBGLVL, ANSI_TEXTBLUE"Then-clause: Parsed statement '%s'\n" NORM_TEXT, statement_token.string);
	} else {
		skip_token = vrParseToEOS(parse);
		vrDbgPrintfN(CONDITIONAL_DEBUG PARSE_DBGLVL, "Then-clause: skipped statement '%s'\n", skip_token.string);
	}

	/** now get the next token to see whether it's an "else" token **/
	else_token = vrParseNextToken(parse);
	vrDbgPrintfN(CONDITIONAL_DEBUG PARSE_DBGLVL, ANSI_TEXTBLUE"(Possible) Else-token is '%s' (%d), @ %d\n" NORM_TEXT, else_token.string, else_token.token, else_token.linenum);

	/* If the next token is not "else", then put it back into the parser, and return */
	if (else_token.token != VRTOKEN_ELSE) {
		vrDbgPrintfN(CONDITIONAL_DEBUG PARSE_DBGLVL, "Else-check: returning non-else token\n");
		vrParseReturnPreviousToken(else_token, parse);
		return statement_token;
	}

	/* now do or skip the next statement -- the "else" clause (negative condition from above) */
	if (!math_value) {
		statement_token = _ConfigParseStatement(config, parse, object, objparser);
		vrDbgPrintfN(CONDITIONAL_DEBUG PARSE_DBGLVL, "Else-clause: Parsed statement '%s'\n", statement_token.string);
	} else {
		skip_token = vrParseToEOS(parse);
		vrDbgPrintfN(CONDITIONAL_DEBUG PARSE_DBGLVL, ANSI_TEXTBLUE"Else-clause: Skipped statement '%s'\n"NORM_TEXT, skip_token.string);
	}

	statement_token.token = VRTOKEN_EOS;

	return statement_token;
}


/******************************************/
/*** IfExec conditional command parsing ***/
/*******************************************************************************/
/*  Format is: "IFEXEC" <shell cmd> <statement(s)> ["ELSE" <statement(s)>] ";" */

/*****************************************************************************/
/* _ConfigParseIfexec(): this function is nearly functionally equivalent to  */
/*   the above _ConfigParseIf() function, with the exception that instead of */
/*   basing the conditional's truth-value from a math expression, it instead */
/*   makes a system call (i.e. a shell-command), and based on the status     */
/*   result of that command performs the "then" and "else" statement(s).     */
/*                                                                           */
/* NOTE: it might actually already be possible to use "if" in combination    */
/*   with "execStatus" to accomplish this, but leaving this here for now.    */
static vrTokenInfo _ConfigParseIfexec(vrConfigInfo *config, vrParseInfo *parse, vrObjectInfo *object, _ConfigParseOptionList objparser)
{
	vrTokenInfo	token;
	vrTokenInfo	statement_token = { VRTOKEN_NOP, "no if statement" };
	vrTokenInfo	skip_token;		/* NOTE: only used for printing debugging info */
	vrTokenInfo	else_token;
	int		exec_value;

	/** next token is the shell command **/
	token = vrParseNextToken(parse);
	if (token.token == VRTOKEN_STRING) {
		exec_value = system(token.string);
	} else {
		vrErrPrintf(RED_TEXT "Error, expected string to execute (line %d), skipping '%s'\n",
			token.linenum, token.string);
	}
	vrDbgPrintfN(PARSE_DBGLVL, "value of 'ifexec' shell command is '%s' (%d)\n", token.string, exec_value);

	/** then read and perhaps execute the statement(s) -- the "then" clause **/
	if (!exec_value) {
		statement_token = _ConfigParseStatement(config, parse, object, objparser);
		vrDbgPrintfN(CONDITIONAL_DEBUG PARSE_DBGLVL, "Parsed statement '%s'\n", statement_token.string);
	} else {
		skip_token = vrParseToEOS(parse);
		vrDbgPrintfN(CONDITIONAL_DEBUG PARSE_DBGLVL, "Skipped statement '%s'\n", skip_token.string);
	}

	/** now get the next token to see whether it's an "else" token **/
	else_token = vrParseNextToken(parse);
	vrDbgPrintfN(CONDITIONAL_DEBUG PARSE_DBGLVL, ANSI_TEXTBLUE"(Possible) Else-token is '%s' (%d), @ %d\n" NORM_TEXT, else_token.string, else_token.token, else_token.linenum);

	/* If the next token is not "else", then put it back into the parser, and return */
	if (else_token.token != VRTOKEN_ELSE) {
		vrDbgPrintfN(CONDITIONAL_DEBUG PARSE_DBGLVL, "Else-check: returning non-else token\n");
		vrParseReturnPreviousToken(else_token, parse);
		return statement_token;
	}

	/* now do or skip the next statement -- the "else" clause (negative condition from above) */
	if (exec_value) {
		statement_token = _ConfigParseStatement(config, parse, object, objparser);
		vrDbgPrintfN(CONDITIONAL_DEBUG PARSE_DBGLVL, "Else: Parsed statement '%s'\n", statement_token.string);
	} else {
		skip_token = vrParseToEOS(parse);
		vrDbgPrintfN(CONDITIONAL_DEBUG PARSE_DBGLVL, "Else: Skipped statement '%s'\n", skip_token.string);
	}

	statement_token.token = VRTOKEN_EOS;

	return statement_token;
}



/******************/
/*** ReadConfig ***/
/********************************************************************/
/*  Format is: 'READCONFIG' { <filename> } ; */

/******************************************************************************/
/* _ConfigParseReadconfig(): this function reads the next token as a string   */
/*   filename from which to directly read additional configuration commands.  */
/*   The _ConfigReadFile() function is called to actually read and parse the  */
/*   file -- it will read that file to completion, and then parsing will      */
/*   return to the point where this operation was encountered.                */
static void _ConfigParseReadconfig(vrConfigInfo *config, vrParseInfo *parse)
{
	vrTokenInfo	token;
	int		result = 1;	/* assume success if result isn't set */

	/** next token is name of file to read **/
	token = vrParseNextToken(parse);

#if 0 /* Since this wasn't implemented, and we're now using '|' for bitwise or, and we have another means to read pipes, we'll disable this proposed feature */
	if (token.token == VRTOKEN_VERTBAR) {
		token = vrParseNextToken(parse);
		if (token.token == VRTOKEN_STRING)
			vrErrPrintf(RED_TEXT "Sorry, Shell Command Config Input not yet implemented.\n" NORM_TEXT);
		else	vrErrPrintf(RED_TEXT "Error, expected a string (line %d), skipping '%s'\n",
				token.linenum, token.string);
	} else
#endif
	if (token.token == VRTOKEN_STRING) {
		/* First try to read the file in the global space, but if it */
		/*   doesn't exist, then read it from the home directory, and */
		/*   if that doesn't exist, then read it from the local directory. */
		/* TODO: consider whether this is the proper ordering. */
		switch (token.string[0]) {
		case '/':
			result = _ConfigReadFile(config, NULL, token.string);
			break;
		case '~':
			/* A tilde followed by a slash treats the tilde as the home directory */
			if (token.string[1] == '/') {
				result = _ConfigReadFile(config, getenv("HOME"), &(token.string[2]));
				break;
			}
		default:
			if (!_ConfigReadFile(config, config->context->homedir, token.string))
				if (!_ConfigReadFile(config, getenv("HOME"), token.string))
					result = _ConfigReadFile(config, ".", token.string);
			break;
		}

		/* check whether we got a good read result */
		if (result == 0) {
			vrErrPrintf(RED_TEXT "Error, unable to read file '%s' (line %d)\n",
				token.string, token.linenum);
		}
	} else {
		vrErrPrintf(RED_TEXT "Error, expected a filename string (line %d), skipping '%s'\n",
			token.linenum, token.string);
	}
}


/******************/
/*** ReadConfig ***/
/********************************************************************/
/*  Format is: 'READCONFIGSTRING' { <string> } ; */

/****************************************************************************/
/* _ConfigParseReadconfig(): this function reads the next token as a string */
/*   from which to directly parse additional configuration commands.  The   */
/*   _ConfigReadString() function is called to actually and parse the file. */
/*   The string will be parsed to completion, and then parsing will return  */
/*   to the point where this operation was encountered.                     */
static void _ConfigParseReadconfigstring(vrConfigInfo *config, vrParseInfo *parse)
{
	vrTokenInfo	token;
	char		info_string[256];	/* string used to indicate source of config data */

	/** next token is name of string to parse **/
	token = vrParseNextToken(parse);

	if (token.token == VRTOKEN_STRING) {
		snprintf(info_string, sizeof(info_string), "on line %d of '%s'", parse->linenum, parse->parse_name);
		if (!_ConfigReadString(config, info_string, token.string)) {
			vrErrPrintf(RED_TEXT "Error, unable to parse configuration string (line %d), skipping '%s'\n",
				parse->linenum, token.string);
		}
	} else {
		vrErrPrintf(RED_TEXT "Error, expected a configuration string (line %d), skipping '%s'\n",
			token.linenum, token.string);
	}
}


/*************************************/
/*** Global option parsing routine ***/
/*************************************/

/***************************************************************************/
/* _ConfigParseGlobalOption(): Parse all the global options.  I.e. the     */
/*   options that aren't specific to any object type.  Things like "echo", */
/*   "setenv", or "if".  It parses a *single* option, though that includes */
/*   options group into a block -- which is handled through a co-recursive */
/*   call to _ConfigParseStatement(). */
/*                                                                            */
/* TODO: determine whether we then call the *objparser or not -- this is what I'm debugging right now */
/*                                                                            */
/* NOTE: unlike the specific object parsing routines, this function does NOT  */
/*   use the vrObjectInfo argument.  Since these options operate on the       */
/*   general configuration of the system.  However, it does conduit it through*/
/*   for options such as the conditional statements which may be encountered  */
/*   while in the middle of parsing some object.                              */
/*   Note also that unlike the statement parsing function, this function DOES */
/*   take a vrTokenInfo argument, which is passed from either                 */
/*   _ConfigParseStatement() itself,or from one of the specific object parsers.*/
/*                                                                             */
/* NOTE also: this parsing routine (unlike the specific object parsers) does   */
/*   NOT advance to the next token, thus it will generally return the token    */
/*   that was passed to it, so care must be taken to make sure that advancement*/
/*   is made, or you can end up in an infinite loop (well, a loop that won't   */
/*   end until the config file or string is entirely consumed, producing lots  */
/*   of error messages in the process.                                         */
static vrTokenInfo _ConfigParseGlobalOption(vrConfigInfo *config, vrParseInfo *parse_data, vrTokenInfo token, vrObjectInfo *object, _ConfigParseOptionList objparser)
{
	vrTokenInfo	skip_token;

	vrTrace("_ConfigParseGlobalOption()", RED_TEXT "entering" NORM_TEXT);
	vrDbgPrintfN(PARSE_DBGLVL, "_ConfigParseGlobalOption(): line %d: string '%s'\n", token.linenum, token.string);

	switch (token.token) {
	/**********************/
	/** Error conditions **/
	case VRTOKEN_PARSEERROR:
		vrErrPrintf("ConfigParse: Encountered an error on line %d of %s: '%s'\n", -1,
			"<fill this in>", token.string);
		break;


	/***************************/
	/** Statement Terminators **/
	case VRTOKEN_SEMICOLON:
	case VRTOKEN_EOS:
		/* just skip them to read the next keyword */
		vrDbgPrintfN(23+0*PARSE_DBGLVL, "Config Parse (line %d): skipping '%s' in global option -- empty statement?\n", token.linenum, token.string);
		break;

	/****************************/
	/** Tokens with no effects **/
	case VRTOKEN_NOP:
		vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "ConfigParse: got **** NOP ****\n");
		break;

	/******************************/
	/** Tokens with side effects **/
	case VRTOKEN_EOF:
		vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "ConfigParse: got **** EOF ****\n");
		break;

	case VRTOKEN_EXIT:
		vrDbgPrintfN(PARSE_DETAIL_DBGLVL, "ConfigParse: got **** Exit token ****\n");
		break;

	case VRTOKEN_ABORT:
		vrDbgPrintfN(ALWAYS_DBGLVL, "ConfigParse: got **** Termination ****\n");
		//vrExit();
		exit(-2);

	/****************************************/
	/** Tokens for builtin/system commands **/
	case VRTOKEN_EXEC:
		token = vrParseNextToken(parse_data);
		if (token.token == VRTOKEN_STRING) {
			system(token.string);
		} else {
			vrErrPrintf(RED_TEXT "Error, expected string to execute (line %d), skipping '%s'\n",
				token.linenum, token.string);
		}

		/* check that command is terminated */
		token = vrParseNextToken(parse_data);
		if (token.token != VRTOKEN_SEMICOLON) {
			vrErrPrintf("FreeVR config: Error (line %d): "
				RED_TEXT "'Exec' option not terminated with a ';'.\n" NORM_TEXT,
				token.linenum);
		}

		break;

	case VRTOKEN_ECHO:
		vrMsgPrintf("FreeVR Config" BOLD_TEXT ">>" NORM_TEXT);
		token = vrParseNextToken(parse_data);
		while (token.token != VRTOKEN_SEMICOLON && token.token != VRTOKEN_CLOSECURL && token.token != VRTOKEN_EOF) {
			/* NOTE: would prefer to use !_TokenIsTerminal(token), but it is not defined in this source file! */
#if 0 /* whether or not to put separating whitespace between tokens -- '0' does */
			vrMsgPrintf("%s ", token.string);
#else
			vrMsgPrintf("%s", token.string);
#endif
			token = vrParseNextToken(parse_data);
		}
		vrMsgPrintf("\n");
		if (token.token == VRTOKEN_CLOSECURL) {
			/* TODO: only return the close-curl?  or also the other terminal tokens? */
			vrParseReturnPreviousToken(token, parse_data);
		}
		break;

	case VRTOKEN_SETENV:
		_ConfigParseSetenv(config, parse_data);
		break;

	case VRTOKEN_UNSETENV:
		_ConfigParseUnsetenv(config, parse_data);
		break;

	case VRTOKEN_PRINTOBJ:
		token = _ConfigParsePrintobject(config, parse_data);
		break;

	case VRTOKEN_PRINTCONFIG:
		_ConfigParsePrintconfig(config, parse_data);
		break;

	case VRTOKEN_GDBPAUSE:
		_ConfigParseGdbpause(config, parse_data);
		break;

	case VRTOKEN_IF:
		token = _ConfigParseIf(config, parse_data, object, objparser);
		break;

	case VRTOKEN_IFEXEC:
		/* TODO: also should reset the token -- needed for exit */
		_ConfigParseIfexec(config, parse_data, object, objparser);
		break;

	case VRTOKEN_ELSE:
		/* The "else" token is only expected by _ConfigParseIf() or _ConfigParseIfExec() */
		skip_token = vrParseToEOS(parse_data);
		vrDbgPrintfN(CONFIG_ERROR_DBGLVL, "Configuration Error (line %d): "
			RED_TEXT "\"else\" keyword not inside an if-conditional -- skipping.\n" NORM_TEXT, token.linenum);
		token.token = VRTOKEN_UNKNOWN;	/* return the "unknown" token to indicate an error has occurred */
		break;

	/********************************************************/
	/** Tokens for Setting Global Assignments and defaults **/
	case VRTOKEN_SET:
	case VRTOKEN_SETDEFAULT:
		_ConfigParseAssignments(config, parse_data);
		break;

	/*****************************/
	/** Tokens for main options **/
	case VRTOKEN_USESYSTEM:
		_ConfigParseUsesystem(config, parse_data);
		break;

	/* token.string provides the name of the object for all these */
	case VRTOKEN_SYSTEM:
		token = _ConfigParseObjectDefinition(config, parse_data,
			VROBJECT_SYSTEM, token.string, (void *)_ConfigParseSystemOption);
		break;

	case VRTOKEN_PROCESS:
		_ConfigParseObjectDefinition(config, parse_data,
			VROBJECT_PROCESS, token.string, (void *)_ConfigParseProcOption);
		break;

	case VRTOKEN_WINDOW:
		_ConfigParseObjectDefinition(config, parse_data,
			VROBJECT_WINDOW, token.string, (void *)_ConfigParseWindowOption);
		break;

	case VRTOKEN_EYELIST:
		_ConfigParseObjectDefinition(config, parse_data,
			VROBJECT_EYELIST, token.string, (void *)_ConfigParseEyelistOption);
		break;

	case VRTOKEN_USER:
		_ConfigParseObjectDefinition(config, parse_data,
			VROBJECT_USER, token.string, (void *)_ConfigParseUserOption);
		break;

	case VRTOKEN_PROP:
		_ConfigParseObjectDefinition(config, parse_data,
			VROBJECT_PROP, token.string, (void *)_ConfigParsePropOption);
		break;

	case VRTOKEN_INPUTDEV:
		_ConfigParseObjectDefinition(config, parse_data,
			VROBJECT_INDEV, token.string, (void *)_ConfigParseIndevOption);
		break;

	case VRTOKEN_INPUTMAP:
		_ConfigParseInmap(config, parse_data);
		break;

	case VRTOKEN_READCONFIG:
		_ConfigParseReadconfig(config, parse_data);
		break;

	case VRTOKEN_READCONFIGSTR:
		_ConfigParseReadconfigstring(config, parse_data);
		break;

	default:
		token.token = VRTOKEN_UNKNOWN;
		skip_token = vrParseToEOS(parse_data);
		vrDbgPrintfN(CONFIG_ERROR_DBGLVL, "Configuration Error (line %d): "
			RED_TEXT "Unknown %s keyword -- skipping: '%s %s'\n" NORM_TEXT,
			token.linenum,
			(object == NULL ? "Global" : vrObjectTypeName(object->object_type)),
			token.string, skip_token.string);
		break;
	}

	vrTrace("_ConfigParseGlobalOption()", RED_TEXT "exiting" NORM_TEXT);
	vrDbgPrintfN(PARSE_DBGLVL, "_ConfigParseGlobalOption(): returning token %d: line %d: string '%s'\n", token.token, token.linenum, token.string);
	return token;
}



/**********************************************/
/*** MAIN configuration file parse routines ***/
/**********************************************/

/***********************************************************************/
/* _ConfigParseStatement() reads the next token from the input stream, */
/*   and then calls either _ConfigParseGlobalOption() when at the top  */
/*   level of the configuration, or the given "*objparer" when in the  */
/*   middle of parsing a particular object.                            */
/* The returned value is the last token from the option-handler.       */
/*   In the primary _ConfigParseMain(), the returned value is used to  */
/*   determine whether to continue looping, or that the current input  */
/*   stream is finished (either intentionally or by error).            */
static vrTokenInfo _ConfigParseStatement(vrConfigInfo *config, vrParseInfo *config_parse, vrObjectInfo *object, _ConfigParseOptionList objparser)
{
	vrTokenInfo	token;
	vrTokenInfo	return_token;
	vrTokenInfo	block_start_token;
static	int		block_depth = 0;/* how deep are we in {} block levels?*/
	char		*old_string_to_free;

	/*******************/
	/** get the token **/
	token = vrParseNextToken(config_parse);
	vrDbgPrintfN(CONFIG_DBGLVL, RED_TEXT "_ConfigParseStatement(): (depth: %d) (line %d) got token **** (%3d) '%s' ****\n" NORM_TEXT,
		block_depth,
		token.linenum, token.token, token.string);

	/**********************/
	/** Error conditions **/
	//case VRTOKEN_PARSEERROR:
	if (token.token == VRTOKEN_PARSEERROR) {
		vrErrPrintf("ConfigParse: Encountered an error on line %d of %s: '%s'\n", -1,
			"<fill this in>", token.string);
		//break;
		return token;
	}


	/* TODO: perhaps check for a semicolon here, and if received, just return. */
	/***************************/
	/** Statement Terminators **/
	//case VRTOKEN_SEMICOLON:
	//case VRTOKEN_EOS:
	if (token.token == VRTOKEN_SEMICOLON || token.token == VRTOKEN_EOS) {
		/* just skip them to read the next keyword */
		vrDbgPrintfN(PARSE_DBGLVL, RED_TEXT "_ConfigParseStatement(): (line %d): skipping '%s' -- empty statement?\n" NORM_TEXT, token.linenum, token.string);
		//break;
		return token;
	}




	/***************************/
	/* TODO: perhaps check for an open-curl here, and if received, call ourself recursively. */
	if (token.token == VRTOKEN_OPENCURL) {
char	msg[128];
		block_depth++;
		block_start_token = token;
		//block_start_token.string = vrShmemStrDup(token.string);
snprintf(msg, sizeof(msg), "%s(@%d)", token.string, token.linenum);
block_start_token.string = vrShmemStrDup(msg);
		do {
#define OPTION1	/* define/undef -- two options for how to handle the close-curlys to close a block */ /* delete me -- after confirming OPTION1 */
#ifdef OPTION1
			token = vrParseNextToken(config_parse);
			if (token.token == VRTOKEN_CLOSECURL) {

			/* build the token's string to be all the statement tokens until the close */
			/* NOTE: we only get the terminating tokens here, so this string isn't very interesting */
			old_string_to_free = block_start_token.string;
			block_start_token.string = vrShmemStrCat(block_start_token.string, " ");		/* add whitespace */
			block_start_token.string = vrShmemStrCat(block_start_token.string, token.string);
snprintf(msg, sizeof(msg), "(@%d)", token.linenum);
block_start_token.string = vrShmemStrCat(block_start_token.string, msg);
			vrShmemFree(old_string_to_free);				/* free the previous copy */
				block_depth--;
				return token;
			}
			vrParseReturnPreviousToken(token, config_parse);
#endif
			token = _ConfigParseStatement(config, config_parse, object, objparser);

			/* build the token's string to be all the statement tokens until the close */
			/* NOTE: we only get the terminating tokens here, so this string isn't very interesting */
			old_string_to_free = block_start_token.string;
			block_start_token.string = vrShmemStrCat(block_start_token.string, " ");		/* add whitespace */
			block_start_token.string = vrShmemStrCat(block_start_token.string, token.string);
snprintf(msg, sizeof(msg), "(@%d)", token.linenum);
block_start_token.string = vrShmemStrCat(block_start_token.string, msg);
			vrShmemFree(old_string_to_free);				/* free the previous copy */

		//} while (token.token != VRTOKEN_CLOSECURL && token.token != VRTOKEN_EOF);
		} while (token.token != VRTOKEN_EOF);
block_start_token.linenum = -100;
		return token;
	}


	/***************************/
	if (token.token == VRTOKEN_CLOSECURL) {
#ifndef OPTION1
		/* TODO: consider assembling the entire block into the token string. */
		block_depth--;
		token = vrParseNextToken(config_parse);
#else
	/* TODO: warn that this means there wasn't an opening block */
vrPrintf(RED_TEXT "}depth = %d -- error\n" NORM_TEXT, block_depth);
#endif
		return token;
	}


	/***************************/
	/* When there's an object parser, we're in the middle of parsing  */
	/*   that type of object, so parse that instead of Global options.*/
	if (objparser == NULL) {
		return_token = _ConfigParseGlobalOption(config, config_parse, token, object, objparser);
	} else {
		return_token = (*objparser)(config, config_parse, token, object, objparser);
	}

#ifdef EXTRA_DEBUG
vrPrintf(BOLD_TEXT "_ConfigParseStatement(): statement parsed from token '%s' on line %d to '%s' on line %d.\n",
token.string, token.linenum,
return_token.string, return_token.linenum);
#endif
	vrDbgPrintfN(CONFIG_DBGLVL, BOLD_TEXT "_ConfigParseStatement(): (line %d) returning token **** (%3d) '%s' ****\n" NORM_TEXT,
		return_token.linenum, return_token.token, return_token.string);

	return return_token;
}



/**********************************************************************/
/* _ConfigParseMain(): as the name implies, this is the primary call  */
/*   for parsing a stream of input (either via file or from a string. */
/*   Thus, this function is called either by _ConfigReadFile() or by  */
/*   _ConfigReadString().  The use of the later is mainly for reading */
/*   the default configuration which is in _DefaultConfig(), or can   */
/*   be used to parse command-line-arguments or environment variables */
/*   containing configuration instructions -- in the latter cases a   */
/*   common use might be to set the "usesystem <system>" option.      */
/* Arguments:                                                         */
/*   * config -- the overall configuration data-structure.  This is   */
/*       what will be filled in with all the configuration options.   */
/*   * config_parse -- a data-structure containing the stream of      */
/*       characters to be parsed, and all the associated meta-data,   */
/*       including (in the case of files), from whence to read more   */
/*       characters.  Also includes information such as line numbers, */
/*       what's being parsed, and the list of keywords and functions. */
static void _ConfigParseMain(vrConfigInfo *config, vrParseInfo *config_parse)
{
	vrTokenInfo	token;

	/* Read every line of the config input, returning when reaching the */
	/*   end, encountering an error, or encountering the config "exit"  */
	/*   {command|operation}.                                           */
	do {
#ifdef EXTRA_DEBUG
vrPrintf(">>>>>>>>>>>>>>>>>>>>>>> in main loop reading next statement\n");
#endif
		token = _ConfigParseStatement(config, config_parse, NULL, NULL);
	} while (token.token != VRTOKEN_EOF && token.token != VRTOKEN_EXIT && token.token != VRTOKEN_PARSEERROR);

	vrDbgPrintfN(CONFIG_WARN_DBGLVL, "FreeVR config: Done Parsing \"%s\".  Read %d lines.\n",
		config_parse->parse_name, config_parse->linenum);

	return;
}


/**************************************************************************/
/* _ConfigReadFile(): this function opens the file provided by the "file" */
/*   argument (using "path" to possibly augment the filename), and then   */
/*   sets some values in the config_parse (type vrParseInfo) structure    */
/*   which is used for keeping track of the parsing process, and is       */
/*   created here to put the value on the stack and thus enable recursive */
/*   calling of this (and related) functions.                             */
/* After the new parsing structure is initialized _ConfigParseMain() is   */
/*   called to do the actual parsing.                                     */
static int _ConfigReadFile(vrConfigInfo *config, char *path, char *file)
{
	FILE		*fp_configfile;
	char		*fullfilename;
	vrParseInfo	config_parse;

	if (path == NULL) {
		fullfilename = file;
	} else {
		fullfilename = (char *)vrShmemAlloc(strlen(path)+strlen(file)+3);
		sprintf(fullfilename, "%s/%s", path, file);
	}

	fp_configfile = fopen(fullfilename, "r");
	if (fp_configfile == NULL) {
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "FreeVR config: could not open file \"%s\".\n", fullfilename);
		return 0;
	}
	vrDbgPrintfN(ALWAYS_DBGLVL, "FreeVR config: " BOLD_TEXT "Reading file \"%s\".\n" NORM_TEXT, fullfilename);

	/* initialize the parsing info structure for reading a config file */
	config_parse.linenum = 0;
	config_parse.parse_name = vrShmemAlloc(20 + strlen(fullfilename));
	sprintf(config_parse.parse_name, "Config file: '%s'", fullfilename);
	config_parse.filename = vrShmemStrDup(fullfilename);
	config_parse.fp = fp_configfile;
	memset(config_parse.input, '\0', MAX_PARSE_INPUT);
	config_parse.next_input = "";
	config_parse.preparsed = NULL;			/* No tokens in preparse holding bin */
	config_parse.suppress_inline = 0;

	/* TODO: we may want to put these four lines at the top of _ConfigParseMain() (redundant w/ _ConfigReadString()) */
	config_parse.tsmap_list = _ConfigTSMap;
	config_parse.nummaps = (sizeof(_ConfigTSMap) / sizeof(_ConfigTSMap[0]));
	config_parse.funcmap_list = _ConfigFuncMap;
	config_parse.numfuncs = (sizeof(_ConfigFuncMap) / sizeof(_ConfigFuncMap[0]));

	_ConfigParseMain(config, &config_parse);

	return 1;
}


/*****************************************************************************/
/* _ConfigReadString(): this function takes a string (conf_string) argument, */
/*   which may be very long, and then sets some values in the config_parse   */
/*   (type vrParseInfo) structure which is used for keeping track of the     */
/*   parsing process, and is created here to put the value on the stack and  */
/*   thus enable recursive calling of this (and related) functions.          */
/* After the new parsing structure is initialized _ConfigParseMain() is      */
/*   called to do the actual parsing.                                        */
/*                                                                           */
/* NOTE: this is actually how the default configuration is handled -- a      */
/*   very long string (later in this source file) is provided as an argument */
/*   to this function.                                                       */
/*                                                                           */
/* Arguments:                                                                */
/*  * config -- pointer to the current configuration object being constructed*/
/*  * info_string -- informational name providing the config info origin     */
/*  * conf_string -- a (perhaps long) string with the configuration settings */
/*                                                                           */
/* For example, the internal default string is set with the call:            */
/*    _ConfigReadString(config, "Default Config String", def_string);        */
static int _ConfigReadString(vrConfigInfo *config, char *info_string, char *conf_string)
{
	vrParseInfo	config_parse;

	if (strlen(conf_string) > MAX_PARSE_INPUT) {
		vrErrPrintf("_ConfigReadString(): too many input characters in \"%s\" (%d > %d).\n", info_string, strlen(conf_string), MAX_PARSE_INPUT);
		return 0;
	}
	vrDbgPrintfN(ALWAYS_DBGLVL, "FreeVR config: " BOLD_TEXT "Reading string \"%s\".\n" NORM_TEXT, info_string);

	config_parse.linenum = 1;
	if (info_string == NULL)
		config_parse.parse_name = vrShmemStrDup("Config string");
	else	config_parse.parse_name = vrShmemStrDup(info_string);
	config_parse.filename = vrShmemStrDup(info_string);	/* 05/12/21: it seems as though the "filename" field is never actually used -- test me, perhaps delete me */
	config_parse.fp = NULL;		/* this indicates reading a string. */
	strncpy(config_parse.input, conf_string, MAX_PARSE_INPUT);
	config_parse.next_input = config_parse.input;
	config_parse.preparsed = NULL;			/* No tokens in preparse holding bin */
	config_parse.suppress_inline = 0;

	/* TODO: we may want to put these four lines at the top of _ConfigParseMain() (redundant w/ _ConfigReadFile()) */
	config_parse.tsmap_list = _ConfigTSMap;
	config_parse.nummaps = (sizeof(_ConfigTSMap) / sizeof(_ConfigTSMap[0]));
	config_parse.funcmap_list = _ConfigFuncMap;
	config_parse.numfuncs = (sizeof(_ConfigFuncMap) / sizeof(_ConfigFuncMap[0]));

	_ConfigParseMain(config, &config_parse);

	return 1;
}


/**********************************************************************/
/* _ConfigReadStringList(): reads a list of of strings in argv[],     */
/*   removing arguments that are used -- so the application can then  */
/*   parse the rest.  Argc is dereferenced so it can be updated to    */
/*   reflect the number of remaining arguments -- though another way  */
/*   to implement this would simply be to set any used arguments to   */
/*   a zero-length string.                                            */
/* A NULL-terminated string-list of unknown length can also be parsed.*/
/*   This is indicated by passing a NULL pointer as argc.  For this,  */
/*   the last argument of argv[] should be NULL,                      */
static void _ConfigReadStringList(vrConfigInfo *config, int *argc, char **argv)
{
	char	info_string[1024];
	int	count;

	if (argv == NULL)
		return;

	if (argc == NULL) {
		for (count = 0; argv[count] != NULL; count++) {
			snprintf(info_string, sizeof(info_string), "Argp[%d] config string", count);
			_ConfigReadString(config, info_string, argv[count]);
		}
	} else {
		for (count = 1; count < *argc; count++) {
			snprintf(info_string, sizeof(info_string), "Argv[%d] config string", count);
			_ConfigReadString(config, info_string, argv[count]);
		}
	}

	return;
}


/***********************************************************************/
/* _ConfigReadEnvVariable(): reads a string from the given environment */
/*   variable name (the "name" argument to the function), and then     */
/*   parses that string into the configuration using the function      */
/*   _ConfigReadString().                                              */
/* Presently this is called once during the configuration process, and */
/*   that is with the VRCONFIG_RC_ENVVAR ("FREEVR") variable.          */
static void _ConfigReadEnvVariable(vrConfigInfo *config, char *name)
{
	char	*env_string;
	char	info_string[1024];

	env_string = getenv(name);
	if (env_string == NULL) {
		vrDbgPrintfN(CONFIG_DBGLVL, "_ConfigReadEnvVariable(): " RED_TEXT "environment variable '%s' does not exist.\n" NORM_TEXT, name);
		return;
	}
	snprintf(info_string, sizeof(info_string), "Envvar[%s] config string", name);
	_ConfigReadString(config, info_string, env_string);

	return;
}


/***************************************************************************/
/* _DefaultConfig(): a simple (but very long) function that contains the   */
/*   default FreeVR configuration in a long string which is then passed to */
/*   _ConfigReadString() for the parsing.                                  */
/* Of course, it's very important that this configuration be well tested   */
/*   as it should not generate any error messages!                         */
/* The default configuration is a simulator.                               */
static void _DefaultConfig(vrConfigInfo *config)
{
static	char	*def_string =

		/* NOTE: Since FreeVR configuration is not based on particular whitespace,*/
		/*   the use of newlines isn't really necessary.  However, it does serve  */
		/*   to assist in finding which line is the culprit since errors are      */
		/*   reported based on line number -- hence the usage of newslines here.  */
		/*   Of course, the conditionally compiled stuff still causes havoc.      */
		/* NOTE also that there are two systems defined here: "default" and       */
		/*   "simulator".  The default system that is used is "simulator", but    */
		/*   when the existing system isn't found, it will instead default to     */
		/*   "default."                                                           */
		"setDefault EyeList = \"" DEFAULT_EYELIST_NAME "\";\n"
		"setDefault VisrenMode = mono;\n"
		"\n"
		"#set DebugLevel = 49;\n"
		"\n"
		"#setDefault PreContextPrint = \"brief\";\n"
		"#setDefault PostContextPrint = \"brief\";\n"
		"setDefault PostConfigPrint = \"brief\";\n"
		"\n"
		"usesystem \"simulator\";\n"
		"\n"
		"system \"default\" = {\n"
		"	malleable = yes;\n"
		"	type = \"simulator\";\n"
		"	procs = \"" DEFAULT_VISRENPROC_NAME "\", \"" DEFAULT_INPUTPROC_NAME "\";\n"
		"	inputmap = \"default\";\n"
		"	visrenmode = \"default\";\n"
		"}\n"
		"\n"
		"system \"simulator\" = {\n"
		"	malleable = yes;\n"
		"	type = \"simulator\";\n"
		"	procs = \"" DEFAULT_VISRENPROC_NAME "\", \"simulator-input\";\n"
		"	inputmap = \"default\";\n"
		"	visrenmode = \"default\";\n"
		"}\n"
		"\n"
		"process \"" DEFAULT_INPUTPROC_NAME "\" = {\n"
		"	malleable = no;\n"
		"	type = input;\n"
		"	usecmin = 15000;\n"
		"	printcolor = 35;\n"
		"	objects = \"" DEFAULT_INDEV_NAME "\";\n"
		"}\n"
		""
		"process \"simulator-input\" = {\n"
		"	malleable = yes;\n"		/* s/b "no" set to "yes" for testing of barriers */
		"	type = input;\n"
		"	usecmin = 15000;\n"
		"	printcolor = 35;\n"
		"	objects = \"simulator-indev\";\n"
#if defined(WIN_WGL)
		/* TODO: Until the Win32 device has 6-sensors, we also need to include the static device */
		"	objects += \"" DEFAULT_INDEV_NAME "\";\n"
#endif
		"}\n"
		"\n"
		"process \"" DEFAULT_VISRENPROC_NAME "\" = {\n"
		"	malleable = yes;\n"	/* [6/11/02] -- made malleable for testing, TODO: should it remain malleable? */
		"	type = visren;\n"
		"	sync = 1;\n"
		"	printcolor = 34;\n"
		"	objects = \"" DEFAULT_WINDOW_NAME "\";\n"
		"}\n"
		"\n"
		"process \"" DEFAULT_TELNETPROC_NAME "\" = {\n"
		"	malleable = yes;\n"
		"	type = telnet;\n"
		"	usecmin = 5000;\n"
		"	printcolor = 32;\n"
		"	args = \"port = 3000;\";\n"
		"	args += \"portrange = 0;\";\n"
		"	#objects = \"??\";	# no use for this (yet)\n"
		"}\n"
		"\n"
		"eyelist \"" DEFAULT_EYELIST_NAME "\" = {\n"
		"	malleable = no;\n"
		"	monofb = \"" DEFAULT_USER_NAME ":cyclops\";\n"
		"	leftfb = \"" DEFAULT_USER_NAME ":lefteye\";\n"
		"	rightfb = \"" DEFAULT_USER_NAME ":righteye\";\n"
		"	leftvp = \"" DEFAULT_USER_NAME ":lefteye\";\n"
		"	rightvp = \"" DEFAULT_USER_NAME ":righteye\";\n"
		"	anaglfb = \"" DEFAULT_USER_NAME ":lefteye:blue\", \"" DEFAULT_USER_NAME ":righteye:red\";\n"
		"}\n"
		"\n"
		"window \"" DEFAULT_WINDOW_NAME "\" = {\n"
		"	malleable = yes;\n"	/* unlike most of the built-in defaults, this one is malleable */
		"	GraphicsType = \"glx\";\n"
		"	mount = \"simulator\";\n"
	/* TODO: the default geometry should probably be "undetermined" allowing */ "\n"
	/*   the app-user to drag out the window.                                */ "\n"
		"	args = \"geometry=350x350+400+400; decoration=window\";\n"
		"	visrenmode = \"default\";\n"
		"	statsProcs = \"self, main, simulator-input\";\n"	/* NOTE: this generates a warning when the "default" system is used (ie. rather than "simulator"). */
		"}\n"
		"\n"
#if 0
		/* TODO: still need to figure props out */ "\n"
		"prop \"" DEFAULT_PROP_NAME "\" = {\n"
#if 0
		"	malleable = no;\n"
#endif
"\n\n\n\n"
		"}\n"
		"\n"
#endif
		"user \"" DEFAULT_USER_NAME "\" = {\n"
		"	malleable = yes;\n"
		"	iod = 0.23;\n"
		"	color = .5647, .3254, .1764;\n"
		"}\n"
		"\n"
		"inputdevice \"" DEFAULT_INDEV_NAME "\" = {\n"
		"	type = \"static\";\n"
		"	args = \"y=5\";\n"
		"	input \"2switch[-]\" = \"2switch(constant[0])\";\n"
		"	input \"2switch[1]\" = \"2switch(constant[0])\";\n"
		"	input \"2switch[2]\" = \"2switch(constant[0])\";\n"
		"	input \"2switch[3]\" = \"2switch(constant[0])\";\n"
		"	input \"valuator[x]\" = \"valuator(constant[0.0])\";\n"
		"	input \"valuator[y]\" = \"valuator(constant[0.0])\";\n"
		"	input \"6sensor[0]\" = \"6sensor(constant[])\";\n"
		"	input \"6sensor[1]\" = \"6sensor(constant[id])\";\n"
		"}\n"
		"\n"
#if defined(WIN_GLX) /* {  { */
		/* Define an X11 input device for GLX windowing systems */ "\n\n"
		"inputdevice \"simulator-indev\" = {\n"
		"	type = \"xwindows\";\n"
		"	args = \"window = " DEFAULT_WINDOW_NAME ";\";\n"
		"	input \"Escape Key\" = \"2switch(keyboard:key[Escape])\";\n"
		"	input \"Left Mouse Button\" = \"2switch(mouse:button[left])\";\n"
		"	input \"Middle Mouse Button\" = \"2switch(mouse:button[middle])\";\n"
		"	input \"Right Mouse Button\" = \"2switch(mouse:button[right])\";\n"
		"	input \"Mouse Left-Right\" = \"valuator(pointer[x])\";\n"
		"	input \"Mouse Up-Down\" = \"valuator(pointer[-y])\";\n"
		"	ipos_translate = 0.0, 5.0, 0.0;\n"
		"	input \"Simulated Head\" = \"6sensor(sim6[0,,ipos])\";\n"
		"	ipos_translate = 0.5, 4.0,-0.5;\n"
#  if 1
		"	input \"Simulated Right Wand\" = \"6sensor(sim6[1,,ipos])\";\n"
		"	ipos_translate = -0.5, 4.0,-0.5;\n"
		"	input \"Simulated Left Wand\" = \"6sensor(sim6[2,,ipos])\";\n"
#  else
		"	input \"Simulated Wand\" = \"6sensor(sim6[1,,ipos])\";\n"
#  endif
		"\n"
		"	control \"system_pause_toggle\" = \"2switch(keyboard:key[pause])\";\n"
		"	control \"print_devinfo\" = \"2switch(keyboard:key[print])\";\n"
		"	control \"print_help\" = \"2switch(keyboard:key[?])\";	# ie. '/'\n"
		"\n"
		"	control \"sensor_next\" = \"2switch(keyboard:key[n])\";\n"	/* fixed 03/17/11 */
		"	control \"setsensor(0)\" = \"2switch(keyboard:key[s])\";\n"
		"	control \"setsensor(1)\" = \"2switch(keyboard:key[w])\";\n"
		"	control \"sensor_reset\" = \"2switch(keyboard:key[r])\";\n"
		"	control \"sensor_resetall\" = \"2switch(keyboard:key[8])\";\n"
		"	control \"sensor_left\" = \"2switch(keyboard:key[leftarrow])\";\n"
		"	control \"sensor_right\" = \"2switch(keyboard:key[rightarrow])\";\n"
		"	control \"sensor_in\" = \"2switch(keyboard:key[uparrow])\";\n"
		"	control \"sensor_out\" = \"2switch(keyboard:key[downarrow])\";\n"
		"	control \"sensor_cw\" = \"2switch(keyboard:key[.])\";\n"
		"	control \"sensor_ccw\" = \"2switch(keyboard:key[,])\";\n"
		"	control \"sensor_nazim\" = \"2switch(keyboard:key[h])\";\n"
		"	control \"sensor_nelev\" = \"2switch(keyboard:key[j])\";\n"
		"	control \"sensor_pelev\" = \"2switch(keyboard:key[k])\";\n"
		"	control \"sensor_pazim\" = \"2switch(keyboard:key[l])\";\n"
		"	control \"sensor_up\" = \"2switch(keyboard:key[f])\";\n"
		"	control \"sensor_down\" = \"2switch(keyboard:key[v])\";\n"
		"	control \"sensor_swap_upin\" = \"2switch(keyboard:key[leftshift])\";\n"
		"	control \"sensor_rotate_sensor\" = \"2switch(keyboard:key[leftalt])\";\n"
		"	control \"pointer_valuator\" = \"2switch(keyboard:key[space])\";\n"
		"	control \"pointer_rot_override\" = \"2switch(keyboard:key[tab])\";\n"
		"	control \"pointer_xy_override\" = \"2switch(keyboard:key[q])\";\n"
		"	control \"pointer_xz_override\" = \"2switch(keyboard:key[a])\";\n"
		"	control \"toggle_relative\" = \"2switch(keyboard:key[1])\";\n"
		"	control \"toggle_space_limit\" = \"2switch(keyboard:key[2])\";\n"
		"	control \"toggle_return_to_zero\" = \"2switch(keyboard:key[3])\";\n"
		"	control \"bell\" = \"2switch(keyboard:key[b])\";\n"	/* added 09/30/21 */
// Experiments:
//"control \"keyval_sign(0)\" = \"2switch(keyboard:key[d])\";\n"
//"control \"keyval_zero(0)\" = \"2switch(keyboard:key[F3])\";\n"
		"}\n"
		"\n"
#elif defined(WIN_WGL) /* } { */
		/* Define a Win32 input device for WGL windowing systems */ "\n\n"
		"inputdevice \"simulator-indev\" = {\n"
		"	type = \"win32\";\n"
		"	args = \"window = default;\";\n"
		"	t2rw_translate = 2.0, 4.0, 0.0;\n"
		"	args += \"y=5; x=-2; z=0; azim = 180; elev = 30; roll = 20;\";\n"
		"\n"
		"	input \"2switch[ESC]\" = \"2switch(keyboard:key[Escape])\";\n"
		"	input \"2switch[left]\" = \"2switch(mouse:button[left])\";\n"
		"	input \"2switch[middle]\" = \"2switch(mouse:button[middle])\";\n"
		"	input \"2switch[right]\" = \"2switch(mouse:button[right])\";\n"
		"\n"
		"	input \"valuator[0]\" = \"valuator(pointer[x])\";\n"
		"	input \"valuator[1]\" = \"valuator(pointer[-y])\";\n"
		"\n"
		"	control \"print_devinfo\" = \"2switch(keyboard:key[print])\";\n"
		"}\n"
#else /* } { */
		/* Define a static input device for unknown windowing systems */ "\n\n"
		"inputdevice \"simulator-indev\" = {\n"
		"	type = \"static\";\n"
		"	args = \"y=5\";\n"
		"	input \"2switch[-]\" = \"2switch(constant[0])\";\n"
		"	input \"2switch[1]\" = \"2switch(constant[0])\";\n"
		"	input \"2switch[2]\" = \"2switch(constant[0])\";\n"
		"	input \"2switch[3]\" = \"2switch(constant[0])\";\n"
		"	input \"valuator[x]\" = \"valuator(constant[0.0])\";\n"
		"	input \"valuator[y]\" = \"valuator(constant[0.0])\";\n"
		"	input \"6sensor[head]\" = \"6sensor(constant[])\";\n"
		"	input \"6sensor[wand]\" = \"6sensor(constant[id])\";\n"
		"}\n"
#endif /* } } */
		"\n";

	vrRemoveAllObjects(config->context);

	_ConfigReadString(config, "Default Config String", def_string);

	return;
}


/**************************************************************************/
/* _SystemConfig(): after the configuration data is read, and the chosen  */
/*   system is being initialized, this function fills in these fields of  */
/*   vrConfigInfo: num_procs, num_windows, num_input_devices, procs[...], */
/*   (and maybe num_users, num_props).                                    */
/* It is called by vrConfigure().                                         */
static void _SystemConfig(vrConfigInfo *config)
{
	int		proccount;
	vrSystemInfo	*systemToUse = NULL;
	vrProcessInfo	*this_proc = NULL;
	char		**orig_proc_names;

	vrTrace("_SystemConfig", "ENTERING");

	/* First, state that nothing has been initialized */
	config->system_init = 0;
	config->procs_init = 0;
	config->windows_init = 0;	/* TODO: currently never set to 1 */
	config->users_init = 0;
	config->props_init = 0;		/* TODO: currently never set to 1 */
	config->inputs_init = 0;

	/* TODO: systemToUse should come from doing a search of the list */
	/*   of system comparing the use option (config->system).        */
	/*   10/8/99 -- isn't that what the following line does?  Perhaps */
	/*   this TODO comment outlasted its need.  I need to find a way  */
	/*   to show what RCS revision each line of the code comes from.  */
	systemToUse = vrObjectSearch(config->context, VROBJECT_SYSTEM, config->system_name);
	if (systemToUse == NULL) {
		/* if no matching name is found, use the default -- the first on the list */
		vrErrPrintf(RED_TEXT "Warning, could not find system \"%s\", using default (with boring input).\n" NORM_TEXT,
			config->system_name);
		vrSleep(5000000);	/* sleep long enough for user to see the warning */
		systemToUse = (vrSystemInfo *)vrObjectFirst(config->context, VROBJECT_SYSTEM);
		if (systemToUse == NULL) {
			/* if still no system to use, there's a problem with the library */
			vrErrPrintf(RED_TEXT "LIBRARY ERROR: No default system available.\n" NORM_TEXT);
			vrExit();
			exit(-1);
		}
		config->system_name = systemToUse->name;
	}
	config->system = systemToUse;


	/******************************************************/
	/* Initialize all the vrConfigInfo process structures */

	/* add a process to the count for the main (parent) process */
	systemToUse->num_procs++;

	/* add the main process to the system list */
	orig_proc_names = systemToUse->proc_names;

	/* TODO: Filling in the systemToUse structure seems somewhat redundant. */
	systemToUse->proc_names = (char **)vrShmemAlloc0((systemToUse->num_procs) * sizeof(char *));
	systemToUse->procs = (vrProcessInfo **)vrShmemAlloc0((systemToUse->num_procs) * sizeof(vrProcessInfo *));

	systemToUse->proc_names[0] = vrThisProc->name;
	systemToUse->procs[0] = vrThisProc;

	config->procs = systemToUse->procs;
	config->num_procs = 1;			/* the main process is already setup */
	config->num_windows = 0;
	config->num_eyelists = 0;
	config->num_input_devices = 0;

	vrTrace("_SystemConfig", "about to loop though the procs the first time");

	/* process #0 is handled as a special case -- the main(parent) process. */
	for (proccount = 0; proccount < systemToUse->num_procs - 1; proccount++) {
		/*** find the process info in the config list, and copy ***/
		/***   it to the vrConfigInfo & the System structure.   ***/
		this_proc = vrObjectSearch(config->context, VROBJECT_PROCESS, orig_proc_names[proccount]);
		if (this_proc == NULL) {
			vrErrPrintf("_SystemConfig(): "
				RED_TEXT "ERROR: Can't find process \"%s\" -- check config!\n"
					"\t-> System will continue with diminished functionality.\n" NORM_TEXT,
				orig_proc_names[proccount]);

			config->context->startup_error |= VRSTARTUP_BADPROC;

			/* define a dummy ("error") process object */
			this_proc = (vrProcessInfo *)vrShmemAlloc0(sizeof(vrProcessInfo));
			vrProcessClear(this_proc);
			this_proc->type = VRPROC_NOCONFIG;
			this_proc->name = vrShmemStrDup(orig_proc_names[proccount]);
			this_proc->proc_term = -1;	/* this process won't be started, so it's already done */
		}

		/*** if the process has already been initialized, then this is a duplicate  ***/
		/***   execution, so create a copy of the process object and work from that ***/
		if (this_proc->used) {
			vrProcessInfo	*new_proc = vrObjectNew(config->context, VROBJECT_PROCESS, "-");
			vrObjectCopy(config->context, new_proc, this_proc);
			new_proc->used = 0;
			new_proc->name = vrShmemStrCat(this_proc->name, " dup");
			strncpy(new_proc->file_lastmod, "Created on the fly in _SystemConfig()", sizeof(new_proc->file_lastmod));
			vrDbgPrintfN(CONFIG_WARN_DBGLVL, "FreeVR Config: duplicate process given, creating copy: '%s'\n", new_proc->name);
			this_proc->used++;		/* the proc with the original name will have a count of how many copies were begun from it */

			this_proc = new_proc;
		}

		this_proc->used = 1;
		systemToUse->procs[proccount+1] = this_proc;
		systemToUse->proc_names[proccount+1] = this_proc->name;


		/*** add up the number of windows, input_devices, ... ***/
		switch (this_proc->type) {
		case VRPROC_VISREN:
			config->num_windows += this_proc->num_thing_names;
			break;
		case VRPROC_INPUT:
			config->num_input_devices += this_proc->num_thing_names;
			break;
		default: /* VRPROC_PARENT, VRPROC_MAIN, VRPROC_NONE, ... */
			break;
		}

		/* increment the count of the number of configured processes */
		config->num_procs++;
	}

	/* verify that we've added to the configuration the correct number of processes */
	if (config->num_procs != systemToUse->num_procs) {
		/* something is wrong -- I think */
		vrDbgPrintfN(CONFIG_WARN_DBGLVL, "FreeVR Config: The number of configured processes (%d) does not match the number in the system specification (%d)\n", config->num_procs, systemToUse->num_procs);
	}

	/* Initialize the Input Mapping */
	config->input_map_name = systemToUse->input_map_name;

	vrTrace("_SystemConfig", "EXITING");
}


/*************************************************************************/
/* _ProcsConfig(): after the configuration data is read, and the chosen  */
/*   system is being initialized, this function fills in these fields of */
/*   vrConfigInfo: windows[...], input_devices[...], etc.                */
/* It is called by vrConfigure().                                        */
static void _ProcsConfig(vrConfigInfo *config)
{
	int		proccount;
	int		maincount = 0;			/* number of main processes encountered */
	int		thingcount;			/* for looping through thing names (some may be invalid) */
	int		next_thing;			/* indicates the next available thing for the process */
	int		window_num = 0;
	int		indev_num = 0;
	vrProcessInfo	*this_proc = NULL;
	vrWindowInfo	*this_window = NULL;
	vrInputDevice	*this_indev = NULL;

	vrTrace("_ProcsConfig", "ENTERING");

	if (config->num_procs == 0) {
		vrErrPrintf("_ProcsConfig(): " RED_TEXT "Hmmm, no processes in the configuration.\n" NORM_TEXT);
		config->context->startup_error |= VRSTARTUP_NOPROCS;
		return;
	}

	config->windows = (vrWindowInfo **)vrShmemAlloc0((config->num_windows) * sizeof(vrWindowInfo *));
	config->input_devices = (vrInputDevice **)vrShmemAlloc0(config->num_input_devices * sizeof(vrInputDevice *));

	for (proccount = 1; proccount < config->num_procs; proccount++) {
		this_proc = config->procs[proccount];

		if (this_proc == NULL)  /* this happens when the wrong process name is given in the config */
			continue;	/* skip to the next process in the list */

		/* Assign an id number to this process.  NOTE: one is added to the */
		/*   process' position in the array in order to make it a 1-based  */
		/*   number, and have the value of 0 indicate the process is not   */
		/*   part of the system in use.                                    */
		/* Need to subtract one for each main-process type defined because */
		/*   they all get copied to be process[0] anyway.                  */
		this_proc->num = proccount + 1 - maincount;

		/*************************************************************/
		/* Append the things of this process to the appropriate list */
		if (this_proc->num_thing_names > 0)
			this_proc->things = (void *)vrShmemAlloc0(this_proc->num_thing_names * sizeof(void *));

		if (vrDbgDo(CONFIG_DBGLVL)) {
			vrMsgPrintf("Doing process config for ");
			vrFprintProcessInfo(stdout, this_proc, verbose);
		}

		switch (this_proc->type) {

		    /***********/
		case VRPROC_MAIN:
			/* If there is a "main" process defined, then it's values */
			/*   should override the builtin default values.          */

			this_proc->pid = config->procs[0]->pid;
			vrObjectCopy(config->context, config->procs[0], this_proc);

			/* Need to copy some additional details */
			config->procs[0]->name = this_proc->name;
			config->procs[0]->id = this_proc->id;
			strcpy(config->procs[0]->file_created, this_proc->file_created);
			config->procs[0]->line_created = this_proc->line_created;
			strcpy(config->procs[0]->file_lastmod, this_proc->file_lastmod);
			config->procs[0]->line_lastmod = this_proc->line_lastmod;

			config->procs[0]->num = 0;
			maincount++;
			if (maincount > 1) {
				/* It doesn't make sense to have more than one main process defined */
				vrDbgPrintfN(CONFIG_WARN_DBGLVL, RED_TEXT "FreeVR: Warning, more than one \"main\" process type in the configuration!\n");
			}

			/* Now shift all the other process objects down one, done */
			/*   to get rid of the duplicate main process object.     */
			config->num_procs--;
			for (thingcount = proccount; thingcount < config->num_procs; thingcount++) {
				config->procs[thingcount] = config->procs[thingcount+1];
			}
			proccount--;
			break;

		    /************/
		case VRPROC_INPUT:
			for (next_thing = 0, thingcount = 0; thingcount < this_proc->num_thing_names; thingcount++) {
				this_indev = vrObjectSearch(config->context, VROBJECT_INDEV, this_proc->thing_names[thingcount]);
				if (this_indev == NULL) {
					vrErrPrintf("_ProcsConfig(): " RED_TEXT "ERROR: Can't find input device '%s'\n"
						"\t-> Perhaps it is misspelled in the configuration file.\n" NORM_TEXT,
						this_proc->thing_names[thingcount]);
					config->context->startup_error |= VRSTARTUP_BADINDEV;
					config->num_input_devices--;
				} else {
					config->input_devices[indev_num] = this_indev;
					this_proc->things[next_thing] = (void *)this_indev;
					if (this_indev->proc != NULL) {
						/* print a warning if this indev has already been assigned */
						vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_ProcsConfig(): "
							RED_TEXT "Warning, input-device '%s' has already been assigned to process '%s', will reassign.\n" NORM_TEXT,
							this_indev->name,
							this_indev->proc->name);
					}
					this_indev->proc = this_proc;
					indev_num++;
					next_thing++;
				}
			}
			this_proc->num_things = next_thing;	/* set num_things to number of valid things found */

			break;

		    /*************/
		case VRPROC_VISREN:
			for (next_thing = 0, thingcount = 0; thingcount < this_proc->num_thing_names; thingcount++) {
				this_window = vrObjectSearch(config->context, VROBJECT_WINDOW, this_proc->thing_names[thingcount]);
				if (this_window == NULL) {
					vrErrPrintf("_ProcsConfig(): " RED_TEXT "ERROR: Can't find window '%s'."
						"\t-> Perhaps it is misspelled in the configuration file.\n" NORM_TEXT,
						this_proc->thing_names[thingcount]);
					config->context->startup_error |= VRSTARTUP_BADWINDOW;
					config->num_windows--;
				} else {
					config->windows[window_num] = this_window;
					this_proc->things[next_thing] = (void *)this_window;
					if (this_window->proc != NULL) {
						/* print a warning if this window has already been assigned */
						vrDbgPrintfN(CONFIG_WARN_DBGLVL, "_ProcsConfig(): "
							RED_TEXT "Warning, window '%s' has already been assigned to process '%s', will reassign.\n" NORM_TEXT,
							this_window->name,
							this_window->proc->name);
					}
					this_window->proc = this_proc;
					window_num++;
					next_thing++;
				}
			}
			this_proc->num_things = next_thing;	/* set num_things to number of valid things found */

			break;
		default: /* VRPROC_COMPUTE, VRPROC_AUDREN, VRPROC_NONE, ... */
			break;
		}
		if (vrDbgDo(CONFIG_DBGLVL)) {
			vrMsgPrintf("Finished process config for ");
			vrFprintProcessInfo(stdout, this_proc, verbose);
		}
	}

	vrTrace("_ProcsConfig", "EXITING");
}


/*****************************************************************************/
/* _WindowsConfig(): fills in fields of vrWindowInfo after the configuration */
/*   data is read, and it's time to start up the selected system.            */
/* It is called by vrConfigure().                                            */
/*   TODO: figure out whether this function is necessary.  */
/*   perhaps it's only needed to fill in user information. */
static void _WindowsConfig(vrConfigInfo *config)
{
static	char		*whitespace = " \t\r\b";
	int		count;
	int		windowcount;
	vrWindowInfo	*this_window;
	vrVisrenModeType visrenmode;
	char		*this_eyelist_name;
	vrEyelistInfo	*this_eyelist;
	vrEyelistInfo	*add_eyelist;
	vrEyeInfo	*this_eye;
	vrEyeInfo	*first_eye;
	vrEyeInfo	*next_eye;
	int		max_num_eyes = 20;	/* TODO: is their a more accurate way to calc? */

	char		*list_dup;		/* duplicate copy of process list for destructive analysis */
	char		*process_name;		/* string with single process name from list of names */
	char		*end_name;		/* pointer to the end of the name string */
	char		*next;			/* pointer to the next part of the list to parse */
	int		stats_count;		/* the number of processes found in the stats list so far */
	vrProcessInfo	*process;		/* the process associated with the given name */

	vrTrace("_WindowsConfig", "ENTERING");

	if (config->num_windows == 0) {
		vrErrPrintf("Hmmm, no windows in the configuration.\n");
		return;
	}

	/* Allocate the memory for all the possible eyelists used in the config.          */
	/*   There cannot be more eyelists than windows (and there probably are less,     */
	/*   but we just allocate the memory for the maximum number possible.  This is    */
	/*   slightly wasteful, but not worth the trouble of being more memory efficient. */
	config->eyelists = (vrEyelistInfo **)vrShmemAlloc0(config->num_windows * sizeof(vrEyelistInfo *));
	config->num_eyelists = 0;

	config->eyes = (vrEyeInfo **)vrShmemAlloc0(max_num_eyes * sizeof(vrEyeInfo *));
	config->num_eyes = 0;

	/* Add each distinct eyelist to the list of eyelists in vrConfigInfo. */
	for (windowcount = 0; windowcount < config->num_windows; windowcount++) {
		this_window = config->windows[windowcount];

		/* Assign an id number to this window.  NOTE: one is added to the */
		/*   window's position in the array in order to make it a 1-based */
		/*   number, and have the value of 0 indicate the window is not   */
		/*   part of the system in use.                                   */
		this_window->num = windowcount + 1;

		/*****************************************/
		/* determine the eyelist for this window */
		this_eyelist_name = this_window->settings.eyelist_name;
		if (this_eyelist_name == NULL)
			this_eyelist_name = config->system->settings.eyelist_name;
		if (this_eyelist_name == NULL)
			this_eyelist_name = config->defaults.eyelist_name;
		if (this_eyelist_name == NULL) {
			vrMsgPrintf("_WindowsConfig(): "
				RED_TEXT "No Eyelist assigned, using builtin default.\n" NORM_TEXT);
			this_eyelist_name = "default";
		}
		this_eyelist = vrObjectSearch(config->context, VROBJECT_EYELIST, this_eyelist_name);
		this_window->eyelist = this_eyelist;

		if (this_eyelist == NULL) {
			/* Without an eyelist, something is seriously amiss, but don't want to crash */
			vrDbgPrintfN(CONFIG_WARN_DBGLVL, RED_TEXT "FreeVR: Warning, there is no eyelist!\n");
			continue;
		}

		/**************************************************************************/
		/* Add eyelist to list of configured eyelists if not on the list already. */
		/*   NOTE: this is different than for Procs & Windows because */
		/*   the same user can (and is likely) to be handled by more  */
		/*   than one window.                                         */
		add_eyelist = this_eyelist;
		for (count = 0; count < config->num_eyelists; count++) {
			if (add_eyelist == config->eyelists[count])
				add_eyelist = NULL;
		}
		if (add_eyelist != NULL) {
			vrDbgPrintfN(CONFIG_DBGLVL, "_WindowsConfig(): "
				BOLD_TEXT "Adding new eyelist [%#p \"%s\"] to config\n" NORM_TEXT,
				add_eyelist, add_eyelist->name);
			config->eyelists[config->num_eyelists] = add_eyelist;
			config->num_eyelists++;

			/* Parse the eyelist strings into Eye objects */
			vrEyelistInitialize(config->context, add_eyelist);
		}

		/* determine the visrenmode for this window */
		visrenmode = this_window->settings.visrenmode;
		if (visrenmode == VRVISREN_NONE_SELECTED)
			visrenmode = config->system->settings.visrenmode;
		if (visrenmode == VRVISREN_NONE_SELECTED)
			visrenmode = config->defaults.visrenmode;
		if (visrenmode == VRVISREN_NONE_SELECTED) {
			visrenmode = VRVISREN_MONO;
			vrDbgPrintfN(AALWAYS_DBGLVL, "_WindowsConfig(): "
				RED_TEXT "No Visual Rendering Mode selected for window \"%s,\" using Mono mode.\n" NORM_TEXT, this_window->name);
		}

		/* set the window's field to the given mode (so later queries will know) */
		this_window->visrenmode = visrenmode;

/* TODO: 05/02/2006 -- this section of code (setting the first_eye) may need to come after the call to _GlxOpenFunc() */
		/* Choose the particular list based on visrenmode options */
		switch (visrenmode) {
		default:  /* TODO: make sure this always works -- might monofb be NULL? */
		case VRVISREN_MONO:
			first_eye = this_eyelist->monofb;
			break;
		case VRVISREN_LEFT:
			first_eye = this_eyelist->leftvp;
			break;
		case VRVISREN_RIGHT:
			first_eye = this_eyelist->rightvp;
			break;
		case VRVISREN_DUALFB:
			first_eye = this_eyelist->leftfb;
			break;
		case VRVISREN_DUALVP:
			first_eye = this_eyelist->leftvp;
			break;
		case VRVISREN_ANAGLYPHIC:
		case VRVISREN_CHECKERBOARD:		/* this is probably temporary, since checkerboard currently overrides the analgyphic mode -- but won't forever (I hope) */
		case VRVISREN_VIBRATE:			/* TODO: I don't know where this goes yet! */
			first_eye = this_eyelist->anaglfb;
			break;
		}

		if (first_eye == NULL) {
			/* Something has gone wrong -- why would this happen? */
			first_eye = this_eyelist->monofb;
			vrDbgPrintfN(AALWAYS_DBGLVL, "_WindowsConfig(): "
				RED_TEXT "Problem encountered attempting to set the stereo mode \"%s\" (%d) in window \"%s,\" will resort to monoscopic rendering.\n" NORM_TEXT,
				vrVisrenmodeTypeName(visrenmode), visrenmode, this_window->name);
		}
		vrDbgPrintfN(CONFIG_DBGLVL, "_WindowsConfig(): "
			BOLD_TEXT "visrenmode for Window \"%s\" is %s, with first eye \"%s\"\n" NORM_TEXT,
			this_window->name, vrVisrenModeName(this_window->visrenmode), first_eye->name);

		/****************************************************************************/
		/* Now add each individual eye to the list of eyes in the window and config */

		this_window->eyes = (vrEyeInfo **)vrShmemAlloc0(max_num_eyes * sizeof(vrEyeInfo *));
		this_window->num_eyes = 0;

		for (next_eye = first_eye; next_eye != NULL; next_eye = next_eye->next) {
			this_eye = next_eye;
			for (count = 0; count < this_window->num_eyes; count++) {
				if (this_eye == this_window->eyes[count])
					this_eye = NULL;
			}
			if (this_eye != NULL) {
				/* this_eye is not yet in the windows list. */
				/*  (only rarely should an eye already be on the window's list) */
				vrDbgPrintfN(CONFIG_DBGLVL, "_WindowsConfig(): "
					BOLD_TEXT "Adding new eye [%#p \"%s\"] to window \"%s\"\n" NORM_TEXT,
					this_eye, this_eye->name, this_window->name);
				this_window->eyes[this_window->num_eyes] = this_eye;
				this_window->num_eyes++;
			}
			this_eye = next_eye;
			for (count = 0; count < config->num_eyes; count++) {
				if (this_eye == config->eyes[count])
					this_eye = NULL;
			}
			if (this_eye != NULL) {
				vrDbgPrintfN(CONFIG_DBGLVL, "_WindowsConfig(): "
					BOLD_TEXT "Adding new eye [%#p \"%s\"] to config\n" NORM_TEXT, this_eye, this_eye->name);
				config->eyes[config->num_eyes] = this_eye;
				config->num_eyes++;
			}
		}

		/* NOTE: I'm sure this could be written a little cleaner, */
		/*   rather than doing the same stuff twice.              */

		/* Choose a second list based on visrenmode options */
		switch (visrenmode) {
		default:
		case VRVISREN_MONO:
		case VRVISREN_LEFT:
		case VRVISREN_RIGHT:
		case VRVISREN_ANAGLYPHIC:
			first_eye = NULL;	/* only one list */
			break;
		case VRVISREN_DUALFB:
			first_eye = this_eyelist->rightfb;
			break;
		case VRVISREN_DUALVP:
			first_eye = this_eyelist->rightvp;
			break;
		}

		/* Now add each individual eye to the list of eyes in the window and configuration */
		for (next_eye = first_eye; next_eye != NULL; next_eye = next_eye->next) {
			this_eye = next_eye;
			for (count = 0; count < this_window->num_eyes; count++) {
				if (this_eye == this_window->eyes[count])
					this_eye = NULL;
			}
			if (this_eye != NULL) {
				/* this_eye is not yet in the windows list. */
				/*  (only rarely should an eye already be on the window's list) */
				vrDbgPrintfN(CONFIG_DBGLVL, "_WindowsConfig(): "
					BOLD_TEXT "Adding new eye [%#p \"%s\"] to window \"%s\"\n" NORM_TEXT,
					this_eye, this_eye->name, this_window->name);
				this_window->eyes[this_window->num_eyes] = this_eye;
				this_window->num_eyes++;
			}
			this_eye = next_eye;
			for (count = 0; count < config->num_eyes; count++) {
				if (this_eye == config->eyes[count])
					this_eye = NULL;
			}
			if (this_eye != NULL) {
				vrDbgPrintfN(CONFIG_DBGLVL, "_WindowsConfig(): "
					BOLD_TEXT "Adding new eye [%#p \"%s\"] to window \"%s\"\n" NORM_TEXT,
					this_eye, this_eye->name, this_window->name);
				config->eyes[config->num_eyes] = this_eye;
				config->num_eyes++;
			}
		}

		if (this_window->num_eyes == 0) {
			vrErrPrintf("_WindowsConfig(): "
				RED_TEXT "Window \"%s\" has no eyes associated with it\n" NORM_TEXT, this_window->name);
		}


		/**********************************************************/
		/* Setup the array of pointers to process statistics data */
		stats_count = 0;
		if (this_window->stats_procs != NULL) {
			list_dup = vrShmemStrDup(this_window->stats_procs);
			process_name = list_dup;
			do {
				process_name += strspn(process_name, whitespace);	/* skip white */
				end_name = strchr(process_name, ',');
				if (end_name != NULL) {
					next = end_name+1;
					end_name[0] = '\0';
				}

				/* If the process_name is an empty string, then move on */
				if (process_name[0] == '\0') {
					process_name = next;
					continue;
				}

				/* Compare name with special-purpose names */
				if (!strcasecmp(process_name, "self")) {
					process = this_window->proc;
				} else if (!strcasecmp(process_name, "main")) {
					process = config->procs[0];
				} else if (!strcasecmp(process_name, "sim")) {
					process = config->procs[0];

				} else if (0) {
					/* TODO: we may want to have the ability to get the Nth process */
					/*   of a certain type -- eg: "input[0], or visren[1]".         */

				} else {
					/* Find the existing process from name and set the stats pointer */
					process = vrObjectArraySearch(VROBJECT_PROCESS, (vrObjectInfo **)(config->procs), config->num_procs, process_name);
				}

				if (process != NULL) {
					this_window->stats[stats_count] = &(process->stats);
					stats_count++;

					/* Make sure the stats are enabled for the process */
					if (process->stats_args == NULL)
						process->stats_args = vrShmemStrDup("");
				} else {
					vrDbgPrintfN(CONFIG_WARN_DBGLVL, "Configuration initialization Warning: "
						RED_TEXT "Window '%s' -- no process '%s' to show statistics of.\n" NORM_TEXT,
						this_window->name, process_name);
				}

				/* get ready to parse the next process name in the list */
				process_name = next;
			} while (end_name != NULL && stats_count < VR_MAXSTATS);
		}

		/* If no process statistics were specifically listed for this window then add "self" */
		if (stats_count == 0) {
			vrDbgPrintf("_WindowsConfig(): Window %s had no associated stats processes, so defaulted to 'self'\n", this_window->name);
			process = this_window->proc;
			this_window->stats[stats_count] = &(process->stats);
			stats_count++;

			/* Make sure the stats are enabled for the process */
			if (process->stats_args == NULL)
				process->stats_args = vrShmemStrDup("");
		}


		/*************************************************************/
		/* Fill in the rest of the windowInfo structure with default */
		/*    values, and create/set graphics handling callbacks.    */
		vrVisrenGetInfo(this_window);
	}

	vrTrace("_WindowsConfig", "Calling PreOpenInit");
	/* NOTE: this assumes that all the windows require the same pre-window */
	/*   opening initialization as window #0.                              */
	vrCallbackInvoke(config->windows[0]->PreOpenInit);

	vrTrace("_WindowsConfig", "EXITING");
}


/****************************************************************************/
/* _EyesConfig(): fills in fields of vrEyeInfo as well as vrUserInfo after  */
/*   the configuration data is read, and it's time to start up the selected */
/*   system.                                                                */
/* It is called by vrConfigure().                                           */
static void _EyesConfig(vrConfigInfo *config)
{
	int		count;
	int		eyecount;
	int		max_num_users = 0;
	vrEyeInfo	*this_eye;
	vrUserInfo	*this_user;

	vrTrace("_EyesConfig", "ENTERING");

	if (config->num_eyes == 0) {
		vrErrPrintf("Hmmm, no eyes in the configuration.\n");
		return;
	}

	/* Unfortunately, the only way to determine the number of users is to */
	/*   go through all the eyes and count the unique ones.  And since    */
	/*   we'd like to know the number of users before we do this, we're   */
	/*   sort of stuck.  But we do know that the number of users cannot   */
	/*   exceed the number of eyes.                                       */
	max_num_users = config->num_eyes;

	/* allocate the memory for the maximum number of users. */
	/*   slightly wasteful, but not worth the trouble of    */
	/*   being more memory efficient.                       */
	config->users = (vrUserInfo **)vrShmemAlloc0(max_num_users * sizeof(vrUserInfo *));
	config->num_users = 0;

	/* ?? TODO: the config file needs to (only?) keep track of the eyes */

	for (eyecount = 0; eyecount < config->num_eyes; eyecount++) {
		this_eye = config->eyes[eyecount];

		/* Assign an id number to this eye.  NOTE: one is added to the eye's */
		/*   position in the array in order to make it a 1-based number, and */
		/*   have the value of 0 indicate the eye is not part of the system  */
		/*   in use.                                                         */
		this_eye->num = eyecount + 1;


		/*************************************************************/
		/* Append this eye's user to the list of users, if necessary */
		this_user = this_eye->user;
		if (this_user == NULL) {
			vrErrPrintf("Eye '%s' doesn't have a user\n", config->eyes[eyecount]);
			/* TODO: what to do here?  exit?  or continue? */
		}

		/* Add user to list of configured users if not on the list already. */
		/*   NOTE: this is different than for Procs & Windows because */
		/*   the same user can (and is likely) to be handled by more  */
		/*   than one window.                                         */
		for (count = 0; count < config->num_users; count++) {
			if (this_user == config->users[count])
				this_user = NULL;
		}
		if (this_user != NULL) {
			vrDbgPrintfN(CONFIG_DBGLVL, "_EyesConfig(): "
				BOLD_TEXT "Adding new user [%#p \"%s\"] to config\n" NORM_TEXT,
				this_user, this_user->name);
			config->users[config->num_users] = this_user;
			config->num_users++;
		}
	}

	vrTrace("_EyesConfig", "EXITING");
}

/*****************************************************************************/
/* _UsersConfig(): fills in the "head" field of vrUserInfo after the         */
/*   configuration data is read, and it's time to start up the selected      */
/*   system.                                                                 */
/* It is called by vrConfigure().                                            */
/*                                                                           */
/* NOTE: most of the vrUserInfo data is actually filled in in _EyesConfig(), */
/*   so all that's left here is to assign a number to each user (1-based),   */
/*   and point the head to a dummy 6-sensor input.                           */
/* TODO: probably use vrInputAddDummyToInputMap() to create the dummy 6-sensor*/
static void _UsersConfig(vrConfigInfo *config)
{
	int		usercount;
	vrUserInfo	*this_user;

	vrTrace("_UsersConfig", "ENTERING");

	if (config->num_users == 0) {
		vrErrPrintf("Hmmm, no users in the configuration.\n");
		return;
	}

	for (usercount = 0; usercount < config->num_users; usercount++) {
		this_user = config->users[usercount];

		/* Assign an id number to this user.  NOTE: one is added to the */
		/*   user's position in the array in order to make it a 1-based */
		/*   number, and have the value of 0 indicate the user is not   */
		/*   part of the system in use.                                 */
		this_user->num = usercount + 1;

		/* Create a dummy sensor that the user's head points to. */
		/*   Once the input devices are initialized, then one of */
		/*   those should replace the head sensor for each user. */
		/* TODO: use the same dummy sensor for each user (ie. create the dummy before this loop) */
		this_user->head = vrInputCreateDataContainerArrayOfType(VRINPUT_6SENSOR, 1, NULL);	/* TODO: 05/26/2006 -- I think we can now use vrInputAddDummyToInputMap() to do this! */ /* 09/29/21: I think do this and then -- delete me */
		vrDbgPrintfN(CONFIG_DBGLVL, "_UsersConfig(): "
			BOLD_TEXT "made a dummy position sensor (%#p) for the head of user %d (%#p).\n" NORM_TEXT,
			this_user->head, usercount, this_user);
	}

	vrTrace("_UsersConfig", "EXITING");
}


/***********************************************************************/
/* _PropsConfig(): is a bit hacky, because while we can define/declare */
/*   props in the config file, there is currently no way to associate  */
/*   props to anything else, so for now we just setup a single dummy   */
/*   prop  in vrConfigInfo.                                            */
/* It is called by vrConfigure().                                      */
static void _PropsConfig(vrConfigInfo *config)
{
	vrTrace("_PropsConfig", "ENTERING");

	config->num_props = 1;
	config->props = (vrPropInfo **)vrShmemAlloc0(config->num_props * sizeof(vrPropInfo *));

	config->props[0] = vrObjectSearch(config->context, VROBJECT_PROP, DEFAULT_PROP_NAME);

	vrTrace("_PropsConfig", "EXITING");
}


/*****************************************************************************/
/* _InputsConfig(): fills in fields of vrInputDevice after the configuration */
/*   data is read, and it's time to start up the selected system.  In this   */
/*   case (input devices), the work is mostly done by vrGetInputDeviceInfo() */
/*   in vr_inputs.c.                                                         */
/* It is called by vrConfigure().                                            */
static void _InputsConfig(vrConfigInfo *config)
{
	int		count;
	vrInputDevice	*this_indev;

	vrTrace("_InputsConfig", "ENTERING");

	for (count = 0; count < config->num_input_devices; count++) {
		this_indev = config->input_devices[count];
		vrGetInputDeviceInfo(this_indev);
	}

	vrTrace("_InputsConfig", "EXITING");
}


/****************************************************************************/
/* vrConfigInitialize(): This is just to initialize an empty configuration  */
/*   structure.  This empty structure will then be built upon to create the */
/*   final configuration, starting with the call to _DefaultConfig().       */
void vrConfigInitialize(vrConfigInfo *config)
{
	char		*env_string;		/* for reading environment variables */

	config->object_type = VROBJECT_CONFIGINFO;
	config->configured = 0;

	/********************************************/
	/*** set some default vrConfigInfo values ***/
	/********************************************/
	vrTrace("vrConfigInitialize", "before default settings");

	vrSettingsClear(&config->defaults);

	/** set some debug printing values **/
	config->defaults.debug_level = DEFAULT_DEBUG_LEVEL;
	env_string = getenv("FREEVR_DEBUGLEVEL");
	if (env_string != NULL)
		config->defaults.debug_level = vrAtoI(env_string);
	env_string = getenv("FREEVR_DEBUGLEVEL_NO");
	if (env_string != NULL)
		config->defaults.debug_level = vrAtoI(env_string);

	config->defaults.debug_exact = 0;
	env_string = getenv("FREEVR_DEBUGEXACT");
	if (env_string != NULL)
		config->defaults.debug_exact = vrAtoI(env_string);
	env_string = getenv("FREEVR_DEBUGEXACT_NO");
	if (env_string != NULL)
		config->defaults.debug_exact = vrAtoI(env_string);

	/** set the default configuration info printing styles **/
	config->defaults.pre_context_print = none;
	config->defaults.pre_config_print = none;
	config->defaults.pre_input_print = none;
	config->defaults.post_context_print = none;
	config->defaults.post_config_print = none;
	config->defaults.post_input_print = none;

	/** set the default value of exit-uponerror **/
	config->defaults.exit_uponerror = 0;

	/** set the default rendering parameters **/
	/* NOTE: config->defaults->visrenmode is left as default and handled in _WindowsConfig() */
	config->defaults.near_clip =   0.1;
	config->defaults.far_clip = 1000.0;

	/***********************************/
	/*** clear some storage pointers ***/
	/***********************************/
	config->eyes = NULL;
}


/* ======================================================================== */
/* ======================================================================== */
/* ======================================================================== */
/* ======================================================================== */
/* ======================================================================== */
/* ======================================================================== */
/* Functions below this point are those that get called by the application. */


/****************************************************************************/
/* vrConfigure(): the very top level configuration function that is called  */
/*   by the user (or if not, then indirectly called by the user via the     */
/*   vrStart() call.                                                        */
/* This call makes sure the vrContext is initialized, and then calls the    */
/*   _DefaultConfig() function to create the basic simulator configuration, */
/*   and then reads (if available) the VRCONFIG_RCFILENAME (".freevrrc")    */
/*   file is read into the configuration.  And then, when provided it may   */
/*   additionally augment the configuration structure with data from the    */
/*   VRCONFIG_RCFILENAME ("FREEVR") environment variable, and then from     */
/*   command line, and even from the application itself via the "argc" and  */
/*   "argv" arguments.                                                      */
/*                                                                          */
/* Once all the configuration parsing is complete, the actual system        */
/*   structures are populated by calling the internal object initialization */
/*   functions such as _SystemConfig().                                     */
/*                                                                          */
/* The configuration isn't actually run until vrStart() is called.          */
vrContextInfo *vrConfigure(int *argc, char **argv, char **appargs)
{
	vrContextInfo	*context;
	vrConfigInfo	*config;

	/**********************************************/
	/*** make sure context is initialized first ***/
	/**********************************************/
	if (vrContext == NULL)
		context = vrContextInitialize();	/* Performer version will alloc from a pfSharedArena */
	else	context = vrContext;

	config = context->config;

	if (config->configured)
		return context;

	vrDbgPrintfN(ALWAYS_DBGLVL, "FreeVR: " BOLD_TEXT "**** Configuring the System! ****\n" NORM_TEXT);

	/**********************************************/
	/*** read all the configuration information ***/
	/**********************************************/
	vrTrace("vrConfigure", "before reading config files");
	_DefaultConfig(config);

	_ConfigReadFile(config, context->homedir, VRCONFIG_RCFILENAME);
	_ConfigReadFile(config, getenv("HOME"), VRCONFIG_RCFILENAME);
	/* TODO: "./VRCONFIG_RCFILENAME" should only be read if "." != "~" */
	/*   This is now done in the NCSA extended CAVE library, get code */
	/*   from there.                                                  */
	if (1)
		_ConfigReadFile(config, ".", VRCONFIG_RCFILENAME);

	_ConfigReadEnvVariable(config, VRCONFIG_RC_ENVVAR);
	_ConfigReadStringList(config, argc, argv);
	_ConfigReadStringList(config, NULL, appargs);	/* CAVElib-style list of string args */

	/**************************************************************************************/
	/*** Create/populate the correct vrConfigInfo structures based on the chosen system ***/
	/**************************************************************************************/
	_SystemConfig(config);
	_ProcsConfig(config);
	_WindowsConfig(config);
	_EyesConfig(config);
	_UsersConfig(config);
	_PropsConfig(config);
	_InputsConfig(config);

	/***********************************************************************************/
	/*** Print the configuration in re-readable format if "printconfig" option given ***/
	/***********************************************************************************/
	if (config->config_print != NULL) {
		FILE	*fp_config;

		fp_config = fopen(config->config_print, "w");
		if (fp_config == NULL)
			fp_config = stdout;

		vrDbgPrintf("FreeVR config: printing active configuration to file '%s'\n", (fp_config == NULL ? "<stdout>" : config->config_print));
		vrFprintf(fp_config, "###################\n");
		vrFprintf(fp_config, "## This config file generated from vrConfigInfo structure\n\n");

		vrFprintConfig(fp_config, config, file_format);

		if (fp_config != stdout)
			fclose(fp_config);
	}

	/****************************************************************/
	/*** Print pre-startup configuration information if requested ***/
	/****************************************************************/
#if 0
	/* TODO: have a configuration setting that determines at what debug level these are printed */
	if (vrDbgDo(DEFAULT_DBGLVL)) {
		vrFprintContext(stdout, context, verbose);
		vrFprintConfig(stdout, config, verbose);
	} else {
		vrFprintContext(stdout, context, brief);
		vrFprintConfig(stdout, config, brief);
	}
#else
	vrFprintContext(stdout, context, (config->system->settings.pre_context_print == def) ? config->defaults.pre_context_print : config->system->settings.pre_context_print);
	vrFprintConfig(stdout, config, (config->system->settings.pre_config_print == def) ? config->defaults.pre_config_print : config->system->settings.pre_config_print);
	vrFprintInput(stdout, context->input, (config->system->settings.pre_input_print == def) ? config->defaults.pre_input_print : config->system->settings.pre_input_print);
#endif

	/*********************************************************/
	/*** Indicate that the configuration stage is complete ***/
	/*********************************************************/
	config->configured = 1;
	vrTrace("vrConfigure", "done");

	return context;
}


char *vrProcessTypeName(vrProcessType type);	/* temporary declaration for vrFprintConfig() -- delete me */
/*************************************************************************/
/* vrFprintConfig(): this function outputs the internal values of the    */
/*   vrConfigInfo structure and many of the nodes that it points to (ie. */
/*   other object structures).  The amount of detail is controlled by    */
/*   the value of the "style" argument.                                  */
void vrFprintConfig(FILE *file, vrConfigInfo *config, vrPrintStyle style)
{
	int		count,
			input,
			prop,
			user,
			eye,
			eyelist,
			window,
			proc;

	if (config == NULL) {
		vrErrPrintf("vrFprintConfig(): " RED_TEXT "memory argument is NULL!\n" NORM_TEXT);
		return;
	}

	if (config->object_type != VROBJECT_CONFIGINFO) {
		vrErrPrintf("vrFprintConfig(): " RED_TEXT "memory argument does not point to the vrConfigInfo structure!\n" NORM_TEXT);
		return;
	}


	switch (style) {
	case none:
		break;

	case brief:
		vrFprintf(file, "==================================================\n");
		vrFprintf(file, "FreeVR Version = " BOLD_TEXT "\"%s\"" NORM_TEXT "\n", config->context->version);
		vrFprintf(file, "%s\n", config->context->compile);
		vrFprintf(file, "Using System = " BOLD_TEXT "\"%s\"" NORM_TEXT " (initialized = %d)\n", config->system_name, config->system_init);
		vrFprintf(file, "debug level = %s (%d)\n", vrDbgValueString(config->defaults.debug_level), config->defaults.debug_level);
		vrFprintf(file, "debug exact = %s (%d)\n", vrDbgValueString(config->defaults.debug_exact), config->defaults.debug_exact);
		vrFprintf(file, "visrenmode default = \"%s\" (%d)\n",
			vrVisrenModeName(config->defaults.visrenmode), config->defaults.visrenmode);
#if 0
		vrFprintf(file, "processor lock default = \"%s\" (%d)\n", vrTokenStringRepresentation(config->defaults.lock_proc), config->defaults.lock_proc);
#else
		vrFprintf(file, "processor lock default = \"%s\"\n", vrTokenStringRepresentation(config->defaults.lock_proc));
#endif
		vrFprintf(file, "InputMap = \"%s\"\n", (config->input_map_name == NULL ? RED_TEXT "(null)" NORM_TEXT : config->input_map_name));

		vrFprintf(file, "%d input devs = [ ", config->num_input_devices);
		for (count = 0; count < config->num_input_devices; count++) {
			if (config->input_devices[count]->operating)
				vrFprintf(file, "\"%s\" ", config->input_devices[count]->name);
			else	vrFprintf(file, RED_TEXT "\"%s\"(FAILURE) " NORM_TEXT, config->input_devices[count]->name);
		}
		vrFprintf(file, "] (initialized = %d)\n", config->inputs_init);

#if 0 /* TODO: add this back when we add the "props" feature to FreeVR. */
		vrFprintf(file, "%d props = [ ", config->num_props);
		for (count = 0; count < config->num_props; count++) {
			if (config->props[count] == NULL) {
				vrFprintf(file, " (nil) [this shouldn't happen] ");
			} else {
				vrFprintf(file, "\"%s\" ", config->props[count]->name);
			}
		}
		vrFprintf(file, "] (initialized = NYI)\n" /* , config->props_init */);
#endif

		vrFprintf(file, "%d users = [ ", config->num_users);
		for (count = 0; count < config->num_users; count++) {
			vrFprintf(file, "\"%s\" ", config->users[count]->name);
		}
		vrFprintf(file, "] (initialized = %d)\n", config->users_init);

		vrFprintf(file, "%d eyes = [ ", config->num_eyes);
		for (count = 0; count < config->num_eyes; count++) {
			vrFprintf(file, "\"%s\" ", config->eyes[count]->name);
		}
		vrFprintf(file, "]\n");

		vrFprintf(file, "%d eyelists = [ ", config->num_eyelists);
		for (count = 0; count < config->num_eyelists; count++) {
			vrFprintf(file, "\"%s\" ", config->eyelists[count]->name);
		}
		vrFprintf(file, "]\n");

		vrFprintf(file, "%d windows = [ ", config->num_windows);
		for (count = 0; count < config->num_windows; count++) {
			//vrFprintf(file, "\"%s\" ", config->windows[count]->name); /* delete me */
			if (config->windows[count]->operating)
				vrFprintf(file, "\"%s\" ", config->windows[count]->name);
			else	vrFprintf(file, RED_TEXT "\"%s\"(FAILURE) " NORM_TEXT, config->windows[count]->name);
		}
#if 0
		vrFprintf(file, "] (initialized = NYI)\n" /*, config->windows_init */);
#else
		/* Until we check for window initialization somewhere, leave this off. */
		vrFprintf(file, "]\n");
#endif

		vrFprintf(file, "%d processes = [ ", config->num_procs);
		for (count = 0; count < config->num_procs; count++) {
			vrFprintf(file, "\"%s\" (%d) ",
				config->procs[count]->name,
				config->procs[count]->pid);
		}
		vrFprintf(file, "]\n", config->procs_init);

		vrFprintf(file, "==================================================\n");
		break;

	case everything:
		/* here, this is the same as verbose -- it also means more info */
		/* during the read phase of configuration.                      */
		/* TODO: decide if it should be more.                           */

	case verbose:
		vrFprintf(file, "==================================================\n");
		vrFprintf(file, "Using System = " BOLD_TEXT "\"%s\"\n" NORM_TEXT, config->system_name);
		vrFprintf(file, "vrConfigInfo at %#p\n", config);
		vrFprintSystemSettings(file, &config->defaults, style);

#if 0
		vrFprintf(file, "General callbacks:\n"
				"\tVisrenInit = %#p\n\tVisrenFrame = %#p\n\tVisrenWorld = %#p\n"
				"\tVisrenSim = %#p\n\tVisrenExit = %#p\n",
				config->VisrenInit,
				config->VisrenFrame,
				config->VisrenWorld,
				config->VisrenSim,
				config->VisrenExit);
#endif

		vrFprintf(file, "InputMap = \"%s\"\n", (config->input_map_name == NULL ? RED_TEXT "(null)" NORM_TEXT : config->input_map_name));

		vrFprintf(file, "=== %d input_devices ===\n", config->num_input_devices);
		vrFprintf(file, "Inputs initialized = %d\n", config->inputs_init);
		for (input = 0; input < config->num_input_devices; input++) {
			vrFprintf(file, "input[%d] at %#p (%#p) = ", input, config->input_devices[input], &config->input_devices[input]);
			vrFprintInputDevice(file, config->input_devices[input], style);
		}

		vrFprintf(file, "=== %d props ===\n", config->num_props);
		vrFprintf(file, "Props initialized = NYI\n" /* , config->props_init */);
		for (prop = 0; prop < config->num_props; prop++) {
			vrFprintf(file, "prop[%d] at %#p (%#p) = ", prop, config->props[prop], &config->props[prop]);
			vrFprintPropInfo(file, config->props[prop], style);
		}

		vrFprintf(file, "=== %d users ===\n", config->num_users);
		vrFprintf(file, "Users initialized = %d\n", config->users_init);
		for (user = 0; user < config->num_users; user++) {
			vrFprintf(file, "user[%d] at %#p (%#p) = ", user, config->users[user], &config->users[user]);
			vrFprintUserInfo(file, config->users[user], style);
		}

		vrFprintf(file, "=== %d eyes ===\n", config->num_eyes);
		for (eye = 0; eye < config->num_eyes; eye++) {
			vrFprintf(file, "eye[%d] at %#p (%#p) = ", eye, config->eyes[eye], &config->eyes[eye]);
			vrFprintEyeInfo(file, config->eyes[eye], style);
		}

		vrFprintf(file, "=== %d eyelists ===\n", config->num_eyelists);
		for (eyelist = 0; eyelist < config->num_eyelists; eyelist++) {
			vrFprintf(file, "eyelist[%d] at %#p (%#p) = ", eyelist, config->eyelists[eyelist], &config->eyelists[eyelist]);
			vrFprintEyelistInfo(file, config->eyelists[eyelist], style);
		}

		vrFprintf(file, "=== %d windows ===\n", config->num_windows);
		vrFprintf(file, "Windows initialized = NYI\n" /* , config->windows_init */);
		for (window = 0; window < config->num_windows; window++) {
			vrFprintf(file, "window[%d] at %#p (%#p) = ", window, config->windows[window], &config->windows[window]);
			vrFprintWindowInfo(file, config->windows[window], style);
		}

		vrFprintf(file, "=== %d processes ===\n", config->num_procs);
		vrFprintf(file, "Processes initialized = %d\n", config->procs_init);
		for (proc = 0; proc < config->num_procs; proc++) {
			vrFprintf(file, "process[%d] at %#p (%#p) = ", proc, config->procs[proc], &config->procs[proc]);
			vrFprintProcessInfo(file, config->procs[proc], style);
		}

		vrFprintf(file, "=== system ===\n");
		vrFprintf(file, "System initialized = %d\n", config->system_init);
		vrFprintf(file, "system at %#p = ", config->system);
		vrFprintSystemInfo(file, config->system, style);

		vrFprintf(file, "==================================================\n");
		break;

	case file_format:
		/***************/
		/* Global Info */
		vrFprintf(file, "\n###################\n");
		vrFprintf(file,   "## Global defaults:\n");
		vrFprintSystemSettings(file, &config->defaults, style);
		vrFprintf(file, "\n");

		/***************/
		/* System Info */
		vrFprintf(file, "\n###################\n");
		vrFprintf(file,   "## System settings:\n");
		vrFprintSystemInfo(file, config->system, style);

		/****************/
		/* Process Info */
		vrFprintf(file, "\n####################\n");
		vrFprintf(file,   "## Process settings:\n");
		for (proc = 0; proc < config->num_procs; proc++) {
			vrFprintProcessInfo(file, config->procs[proc], style);
		}

		/***************/
		/* Window Info */
		vrFprintf(file, "\n###################\n");
		vrFprintf(file,   "## Window settings:\n");
		for (window = 0; window < config->num_windows; window++) {
			vrFprintWindowInfo(file, config->windows[window], style);
		}

		/*************/
		/* Eyes Info */
		vrFprintf(file, "\n#################\n");
		vrFprintf(file,   "## Eyes settings: (NOTE: commented out because not part of the configuration file.)\n");
		for (eye = 0; eye < config->num_eyes; eye++) {
			vrFprintEyeInfo(file, config->eyes[eye], style);
		}

		vrFprintf(file, "\n####################\n");
		vrFprintf(file,   "## Eyelist settings:\n");
		for (eyelist = 0; eyelist < config->num_eyelists; eyelist++) {
			vrFprintEyelistInfo(file, config->eyelists[eyelist], style);
		}

		/**************/
		/* Users Info */
		vrFprintf(file, "\n#################\n");
		vrFprintf(file,   "## User settings:\n");
		for (user = 0; user < config->num_users; user++) {
			vrFprintUserInfo(file, config->users[user], style);
		}

		/**************/
		/* Props Info */
		vrFprintf(file, "\n#################\n");
		vrFprintf(file,   "## Prop settings:\n");
		for (prop = 0; prop < config->num_props; prop++) {
			vrFprintPropInfo(file, config->props[prop], style);
		}

		/**********************/
		/* Input Devices Info */
		vrFprintf(file, "\n#########################\n");
		vrFprintf(file,   "## Input Device settings:\n");
		for (input = 0; input < config->num_input_devices; input++) {
			vrFprintInputDevice(file, config->input_devices[input], style);
		}

		/******************/
		/* Input Map Info */
		vrFprintf(file, "\n#################\n");
		vrFprintf(file,   "## Input Mapping:   TODO\n");
		/* TODO: */

		vrFprintf(file,   "\n\n");
		break;

	case diagram:
		/* TODO: the sub-types should be handled by calling the sub-type print functions */
		/* TODO: consider trying HTML commands in the labels to change fonts, size & color */

		/* Begin */
		vrFprintf(file, "# Diagram of the current FreeVR system.\n");
		vrFprintf(file, "# Use the following command to make a PNG image.\n");
		vrFprintf(file, "#   %% dot -Tpng -O <file>.dot\n");
		vrFprintf(file, "digraph FreeVRconfiguration {\n");

		/* The overall system */
		vrFprintf(file, "\t# System (top level)\n", proc);
		vrFprintf(file, "\tsystem [label=\"%s\\nsystem\", tooltip=\"visrenmode=%s\"]\n", config->system->name, "mono");

		/* The processes */
		for (proc = 0; proc < config->num_procs; proc++) {
			vrFprintf(file, "\t# Process #%d\n", proc);
			vrFprintf(file, "\tsystem -> \"%s\"\n", config->procs[proc]->name);
			vrFprintf(file, "\t\"%s\" [label=\"%s\\nprocess\", tooltip=\"type=%s\"]\n", config->procs[proc]->name, config->procs[proc]->name, vrProcessTypeName(config->procs[proc]->type));

			/* The "things" of each processes */
			for (count = 0; count < config->procs[proc]->num_things; count++) {
				vrFprintf(file, "\t\"%s\" -> \"%s\"\n", config->procs[proc]->name, config->procs[proc]->thing_names[count]);
				/* TODO: put something interesting in the 'tooltip' labels */
				vrFprintf(file, "\t\"%s\" [label=\"%s\\n%s\", tooltip=\"xxx\"]\n", config->procs[proc]->thing_names[count], config->procs[proc]->thing_names[count], vrObjectTypeName(((vrObjectInfo*)(config->procs[proc]->things[count]))->object_type));
			}

		}

		/* Finish */
		vrFprintf(file, "}\n");
		break;

	default:
	case one_line:	/* TODO: implement this */
		vrFprintf(file, "TODO: put \"one_line\" version of config info here\n");
		break;
	case machine:	/* TODO: implement this */
		vrFprintf(file, "TODO: put \"machine\" version of config info here\n");
		break;
	}
}


/****************************************************************/
/* Some helper functions that make use of the config structure. */
/****************************************************************/

/****************************************************************************/
/* vrSetNear(): Set the global default value of the near clipping plane for */
/*   the configuration.  Any renderings that don't set their own value will */
/*   inherit this value.                                                    */
/* TODO: consider naming this vrConfigSetNear() -- [04/16/21: I decided against this] */ /* delete me (this comment) */
/* TODO: add the ability to set for specific user, window, or as global default */
/*	NOTE: the above could be written as one function by passing a pointer   */
/*	to whatever structure should be set, and then using the struct-type     */
/*      field to determine how to do the setting -- and in app-prog code, they  */
/*      will generally pass a vrContext handle, so it will set for the system   */
/*	in that case.                                                           */
void vrSetNear(vrContextInfo *context, double newnear)
{
	int		count;

	/* TODO: [04/16/21] consider setting the overall defaults -- can that be done?  Yes, I think it can, but with reconsideration, since there is just one running system, I don't know that it's necessary to go any deeper into the inheritance. */
	context->config->system->settings.near_clip = newnear;
}


/******************************************************************************/
/* vrSetFar(): Set the global default value of the far clipping plane for the */
/*   configuration.  Any renderings that don't set their own value will       */
/*   inherit this value.                                                      */
/* TODO: consider naming this vrConfigSetFar() -- [04/16/21: I decided against this] */ /* delete me (this comment) */
/* TODO: add the ability to set for specific user, window, or as global default */
/*	NOTE: the above could be written as one function by passing a pointer   */
/*	to whatever structure should be set, and then using the struct-type     */
/*      field to determine how to do the setting -- and in app-prog code, they  */
/*      will generally pass a vrContext handle, so it will set for the system   */
/*	in that case.                                                           */
void vrSetFar(vrContextInfo *context, double newfar)
{
	int		count;

	/* TODO: consider setting the overall defaults -- can that be done? */
	context->config->system->settings.far_clip = newfar;
}


/*********************************************/
/*********************************************/
#ifdef TEST_APP /* { */

int main(int argc, char *argv[])
{
	vrConfigure(0, NULL, NULL);
}
#endif /* } TEST_APP */

